/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Win32/DirectX
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		Direct Draw
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	1996/06/27	porting DOS version on WinG
 **		1.20	1997/08/10	porting WinG version on Win32/DirectX
 **		1.30	1998/03/12	
 **		1.43	2007/09/01	XPΉ
 **		1.44	2009/12/15	Vista/7Ή
 **		1.51	2010/01/27	DirectSound OnlyɕύX
 **		1.80	2020/09/25	^PCGΉ(3)
 **------------------------------------------------------------------------*/
#include <crtdbg.h>
#include <stdio.h>
#include "n80_form.h"
#include "n80_directx.h"
#include "n80_assay.h"
#include "../z80/z80_thread.h"
#include "../pc-8001/pc-8001.h"
#include "../lib/icj_memory.h"
#include "../lib/icj_dib.h"
/**--------------------------------------------------------------------------
 **		O[oϐ
 **------------------------------------------------------------------------*/
extern n80_assay	*assay;
extern n80_form		*form;
extern z80_thread	*th;
/**--------------------------------------------------------------------------
 **		RXgN^^fXgN^
 **------------------------------------------------------------------------*/
n80_directx::n80_directx(void) {
	m_dd    = NULL;
	m_dds1  = NULL;
	m_dds2  = NULL;
	m_ddp   = NULL;
	m_ds	= NULL;
	m_dssb1	= NULL;
	m_dssb2	= NULL;
	m_hwnd  = NULL;

	m_palentry   = NULL;
	m_mute = FALSE;
	m_show_sys_info = 0;

	m_wait_vb = FALSE;			// A҂敪
	m_beep_amplitude = 6000;	// Beep WaveU
	m_pcg_amplitude = 6000;		// PCG WaveU
	m_use_palette = true;
	m_fullscreen_mode = false;
	n80win32.screen_size = 0;
	
	memset(&this->m_beep_bank, 0, sizeof(this->m_beep_bank));
	memset(&this->m_sound_bank, 0, sizeof(this->m_sound_bank));
	this->m_beep_phase_adjuster = 0;
	this->m_sound_phase_adjuster = 0;
	this->m_beep_bank.mCurrent.mState = 1;
	this->m_beep_bank.mCurrent.mTicks = ((LONGLONG)n80device.param.sync_ticks) * 40;

	m_font = ::CreateFont(12, 0, 0, 0, FW_REGULAR, FALSE, FALSE, FALSE, SHIFTJIS_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, PROOF_QUALITY, FIXED_PITCH | FF_MODERN, "lr@SVbN");
}

n80_directx::~n80_directx (void) {
	::DeleteObject(m_font);
	dispose_dx();
}
/**--------------------------------------------------------------------------
 **		DirectX^I
 **------------------------------------------------------------------------*/
BOOL
n80_directx::init_dx(HWND hWnd) {
	m_hwnd = hWnd;

	// pbg擾
	m_palentry = (LPPALETTEENTRY) xcalloc(sizeof (PALETTEENTRY), 256);
	if (m_palentry == NULL)
		return  FALSE;
	icj_dib::setDefEntry(8, m_palentry);

	// DirectDraw
	if (!init_dd()) {
		dispose_dx();
		return FALSE;
	}
	// DirectSound
	if (!init_ds()) {
		dispose_dx();
		return FALSE;
	}
	return TRUE;
}

void
n80_directx::dispose_dx() {
	dispose_dd();
	dispose_ds();
	//pbgz
	if (m_palentry != NULL)
	xfree(m_palentry);
	m_palentry = NULL;
}
/**--------------------------------------------------------------------------
 **		DirectDraw ď
 **------------------------------------------------------------------------*/
BOOL
n80_directx::reinit_dd() {
	this->dispose_dd();
	this->m_fullscreen_mode = !this->m_fullscreen_mode;
	return this->init_dd();
}
/**--------------------------------------------------------------------------
 **		DirectDraw 
 **------------------------------------------------------------------------*/
BOOL
n80_directx::init_dd() {
	static const POINT pts[] = {{660, 420},{640, 400},{640, 480},{720, 480},{800, 480},{800, 600}};
	// Surface Discripter
	::ZeroMemory(&m_ddsd, sizeof(m_ddsd));
	m_ddsd.dwSize = sizeof(m_ddsd);

	// DirectDraw
	BOOL success = FALSE;
	if (this->getFullScreenMode()) {
		// ŋ߂CPU͏\̂32bppŌŒ肷
		//for (int bpp = 0; bpp < 2 && !success; ++bpp) {
			int depth = 32;// bpp == 0 ? 8 : 32;
			if (n80win32.screen_size == 0) {
				for (int i = 0; i < 6; ++i) {
					if (create_dd(pts[i].x, pts[i].y, depth)) {
						success = TRUE;
						break;
					}
				}
			} else {
				if (create_dd(720, 480, 32/*bpp == 0 ? 8 : 32*/)) {
					success = TRUE;
					//break;
				}
			}
		//}
	} else {
		// Window Mode
		for (int i = 0; i < 6; ++i) {
			if (create_dd(pts[i].x, pts[i].y, 32)) {
				success = TRUE;
				break;
			}
		}
	}
	if (!success)
		return  FALSE;

	// Surface/Palette
	if (!create_surface() || !create_palette()) {
		dispose_dd();
		return  FALSE;
	}
	//pbg
	if (m_ddp != NULL)
		m_dds1->SetPalette(m_ddp);

	return  TRUE;
}
/**--------------------------------------------------------------------------
 **		DirectDrawC^[tF[X쐬^j
 **------------------------------------------------------------------------*/
BOOL
n80_directx::create_dd(int width, int height, int depth) {
	// DirectDrawIuWFNg
	//if (FAILED(::DirectDrawCreateEx(NULL, (VOID**)&m_dd, IID_IDirectDraw7, NULL))) {
	if (FAILED(::DirectDrawCreate(NULL, &m_dd, NULL))) {
		assay->outputLog(TEXT("DirectX: DirectDrawCreate Failed"));
		return  FALSE;
	}
	if (this->getFullScreenMode()) {
		// ### Full Screen Mode ###
		assay->outputLog(TEXT("DirectX:  Full Screen Mode start"));

		// Window StyleύX
		form->prepareFullScreenMode(width, height);

		// xݒ
		if (FAILED(m_dd->SetCooperativeLevel(m_hwnd,
		//	DDSCL_ALLOWREBOOT | DDSCL_NOWINDOWCHANGES |	DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN)))
			DDSCL_ALLOWREBOOT | DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN))) {
		//	DDSCL_SETFOCUSWINDOW))){
			assay->outputLog(TEXT("DirectX: dd->SetCooperativeLevel Failed"));
			goto exit_function;
		}
		if (FAILED(m_dd->SetDisplayMode(width, height, depth))) {
			assay->outputLog(TEXT("DirectX: dd->SetDisplayMode(Width=%d, Height=%d, Depth=%d) Failed"), width, height, depth);
			goto exit_function;
		}
		assay->outputLog(TEXT("DirectX: dd->SetDisplayMode(Width=%d, Height=%d, Depth=%d) Succeeded"), width, height, depth);

		::ZeroMemory(&m_ddcap, sizeof(m_ddcap));
		m_ddcap.dwSize = sizeof(m_ddcap);
		if (FAILED(m_dd->GetCaps(&m_ddcap, NULL))) {
			assay->outputLog(TEXT("DirectX:  dd->RestoreDisplayMode"));
			m_dd->RestoreDisplayMode();
			goto exit_function;
		}
		assay->outputLog(TEXT("DirectX: OKOKOKOKOK"));
	} else {
		// ### Window Mode ###
		assay->outputLog(TEXT("DirectX:  Window Mode start"));

		// Window StyleύX
		form->prepareWindowMode(width, height);

		// Windows Client Rect(GlobalW)擾
		this->reaclcWindowRect();

		// xݒ
		if (FAILED(m_dd->SetCooperativeLevel(m_hwnd, DDSCL_NORMAL))) {
			assay->outputLog(TEXT("DirectX: dd->SetCooperativeLevel Failed"));
			goto exit_function;
		}
		::ZeroMemory(&m_ddcap, sizeof(m_ddcap));
		m_ddcap.dwSize = sizeof(m_ddcap);
		if (FAILED(m_dd->GetCaps(&m_ddcap, NULL))) {
			assay->outputLog(TEXT("DirectX:  dd->RestoreDisplayMode"));
			m_dd->RestoreDisplayMode();
			goto exit_function;
		}
	}
	m_width  = width;
	m_height = height;
	return  TRUE;
	//	G[I
exit_function:

	dispose_dd();
	return  FALSE;
}

void
n80_directx::dispose_dd() {
	dispose_palette();
	dispose_surface();
	if (m_dd) {
		m_dd->RestoreDisplayMode();
		m_dd->Release();
		m_dd = NULL;
	}
}
/**--------------------------------------------------------------------------
 **		DirectDraw Surface 쐬^j
 **------------------------------------------------------------------------*/
BOOL
n80_directx::create_surface() {
	DDSURFACEDESC	ddsd;
	//DDSCAPS			caps;

	if (!m_dd)
		return  FALSE;

	if (this->getFullScreenMode()) {
		::ZeroMemory (&ddsd, sizeof(ddsd));
		ddsd.dwSize            = sizeof(ddsd);
		ddsd.dwFlags           = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP |
								 DDSCAPS_COMPLEX;
		ddsd.dwBackBufferCount = 1;
		//	Primary surface
		if (FAILED(m_dd->CreateSurface(&ddsd, &m_dds1, NULL))) {
				assay->outputLog(TEXT("DirectX:  dd->CreateSurface(Primary [Full Screen Mode]) Failed"));
				dispose_surface();
				return  FALSE;
		}
		//	Back-buffer surface
		//caps.dwCaps = DDSCAPS_BACKBUFFER;
		//if (FAILED(m_dds1->GetAttachedSurface(&caps, &m_dds2)))
		//{
		//	dispose_surface();
		//	return  FALSE;
		//}
		//	Off-screen surface
		::ZeroMemory (&ddsd, sizeof(ddsd));
		ddsd.dwSize				= sizeof(ddsd);
		ddsd.dwFlags			= DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		//ddsd.ddsCaps.dwCaps		= DDSCAPS_OFFSCREENPLAIN | 
		//		m_use_video_mem ? DDSCAPS_VIDEOMEMORY : DDSCAPS_SYSTEMMEMORY;
		ddsd.ddsCaps.dwCaps		= DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;
		ddsd.dwWidth			= m_width;
		ddsd.dwHeight			= m_height;
		if (FAILED(m_dd->CreateSurface(&ddsd, &m_dds2, NULL))) {
			assay->outputLog(TEXT("DirectX:  dd->CreateSurface(Off-screen [Full Screen Mode]) Failed"));
			dispose_surface();
			return FALSE;
		}
	} else {	// Window Mode
		::ZeroMemory (&ddsd, sizeof(ddsd));
		ddsd.dwSize            = sizeof(ddsd);
		ddsd.dwFlags           = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE;
		//	Primary surface
		if (FAILED(m_dd->CreateSurface(&ddsd, &m_dds1, NULL))) {
			assay->outputLog(TEXT("DirectX:  dd->CreateSurface(Primary) Failed"));
			dispose_surface();
			return  FALSE;
		}
		::ZeroMemory (&ddsd, sizeof(ddsd));
		ddsd.dwSize				= sizeof(ddsd);
		ddsd.dwFlags			= DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		//ddsd.ddsCaps.dwCaps	= DDSCAPS_OFFSCREENPLAIN | 
		//		m_use_video_mem ? DDSCAPS_VIDEOMEMORY : DDSCAPS_SYSTEMMEMORY;
		ddsd.ddsCaps.dwCaps		= DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;
		ddsd.dwWidth			= m_width;
		ddsd.dwHeight			= m_height;
		if (FAILED(m_dd->CreateSurface(&ddsd, &m_dds2, NULL))) {
			ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
			if (FAILED(m_dd->CreateSurface(&ddsd, &m_dds2, NULL))) {
				assay->outputLog(TEXT("DirectX:  dd->CreateSurface(Off-screen) Failed"));
				dispose_surface();
				return FALSE;
			}
		}
		// Clipper
		LPDIRECTDRAWCLIPPER lpDDClip;
		if (FAILED(m_dd->CreateClipper(0, &lpDDClip, 0))) {
			assay->outputLog(TEXT("DirectX: dd->CreateClipper Failed"));
			return FALSE;
		}
		if (FAILED(lpDDClip->SetHWnd(0, m_hwnd))) {
			lpDDClip->Release();
			assay->outputLog(TEXT("DirectX: clip->SetHWnd Failed"));
			return FALSE;
		}
		if (FAILED(m_dds1->SetClipper(lpDDClip))) {
			lpDDClip->Release();
			assay->outputLog(TEXT("DirectX: dds1->SetClipper Failed"));
			return FALSE;
		}
		lpDDClip->Release();
	}
	// Off-screen pixel format擾
	::ZeroMemory (&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	if (FAILED(m_dds2->GetSurfaceDesc(&ddsd))) {
		assay->outputLog(TEXT("DirectX:  dds->GetSurfaceDesc(Off-screen) Failed"));
		dispose_surface();
		return FALSE;
	}
	if ((ddsd.ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED8) == DDPF_PALETTEINDEXED8) {
		assay->outputLog(TEXT("DirectX:  Surface pixcel format = 8-bit color indexed palette"));
		this->setUsePalette(TRUE);
		m_bpp_mode = 0;
	} else {
		this->setUsePalette(FALSE);
		m_bpp_mode = 1;
		if ((ddsd.ddpfPixelFormat.dwFlags & DDPF_RGB) == DDPF_RGB) {
		} else {
			assay->outputLog(TEXT("DirectX:  Surface does not support RGB pixcel format!"));
			dispose_surface();
			return FALSE;
		}
		assay->outputLog(TEXT("DirectX:  Surface pixcel format = RGB"));
	}
	m_width						= ddsd.dwWidth;
	m_height					= ddsd.dwHeight;
	n80win32.width				= m_width;
	n80win32.height				= m_height;
	n80win32.vram_lpitch		= ddsd.lPitch;
	n80win32.vram_pixcel_pitch	= this->getUsePalette() ? n80win32.vram_lpitch : n80win32.vram_lpitch / 4;
	n80win32.vram_top_offset	= (m_height - 400) / 2;
	n80win32.vram_bottom_offset	= n80win32.vram_top_offset + 400;
	n80win32.vram_left_offset	= (m_width - 640) / 2;
	n80win32.vram_right_offset	= n80win32.vram_left_offset + 640;
	reaclcWindowRect();

	for (int line = 0; line < 20; ++line)
		n80win32.vtbl_vram20[line] = n80win32.vram_pixcel_pitch * (n80win32.vram_top_offset + line * 20) + n80win32.vram_left_offset;
	for (int line = 0; line < 25; ++line)
		n80win32.vtbl_vram25[line] = n80win32.vram_pixcel_pitch * (n80win32.vram_top_offset + line * 16) + n80win32.vram_left_offset;
	for (int v = 0; v < 20; ++v)
		n80win32.vram_vpitch[v] = n80win32.vram_pixcel_pitch * v;

	assay->outputLog(TEXT("DirectX: Surface desc lpitch = %d, height = %d, pixcel format = %d"), ddsd.lPitch, ddsd.dwHeight, ddsd.ddpfPixelFormat.dwFlags);
	return  TRUE;
}

void
n80_directx::dispose_surface () {
	if (m_dds2) {
		m_dds2->Unlock(NULL);
		m_dds2->Release();
		m_dds2 = NULL;
	}
	if (m_dds1) {
		m_dds1->Unlock(NULL);
		m_dds1->Release();
		m_dds1 = NULL;
	}
}

void
n80_directx::restore_surface () {
	m_dds1->Restore();
	m_dds2->Restore();
	clear_surface();
}

void
n80_directx::clear_surface () {
	DDBLTFX ddbltfx;
	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor = 0;
	
	RECT rect;
	rect.left = 0;		rect.top = 0;
	rect.right = m_width;	rect.bottom = m_height;
	m_dds1->Blt(&rect, NULL, NULL, DDBLT_COLORFILL, &ddbltfx);
	m_dds2->Blt(&rect, NULL, NULL, DDBLT_COLORFILL, &ddbltfx);
}
/**--------------------------------------------------------------------------
 **		DirectDraw Palette 쐬^j
 **------------------------------------------------------------------------*/
BOOL
n80_directx::create_palette() {
	if (!m_dd)
		return  FALSE;

	if (this->getUsePalette()) {
		if (FAILED(m_dd->CreatePalette(DDPCAPS_8BIT, m_palentry, &m_ddp, NULL))) {
			dispose_palette();
			return  FALSE;
		}
	}
	return  TRUE;
}

void
n80_directx::dispose_palette () {
	if (m_ddp) {
		m_ddp->Release();
		m_ddp = NULL;
	}
}
/**--------------------------------------------------------------------------
 **		DirectDraw PC-8001 `
 **------------------------------------------------------------------------*/
LPBYTE
n80_directx::surface_lock() {
	HRESULT	hr;
	for (;;) {
		hr = m_dds2->Lock(NULL, &m_ddsd, 0, NULL);
		if (hr == DDERR_WASSTILLDRAWING) {
			assay->outputLog(TEXT("DirectX: realizer() still drawing"));
			continue;
		}
		//if (hr == DDERR_SURFACELOST || hr == DDERR_SURFACEBUSY)
		//	return;
		if (hr == DDERR_SURFACELOST) {
			assay->outputLog(TEXT("DirectX: realizer() surface lost"));
			if (m_hwnd == ::GetForegroundWindow()) {
				if (SUCCEEDED(m_dds2->Restore()))
					continue;
				assay->outputLog(TEXT("DirectX: realizer() surface restore Failed"));
			}
		}
		if (hr != DD_OK)
			return NULL;
		else
			break;
	}
	return (LPBYTE)m_ddsd.lpSurface;
}

int
n80_directx::realizer() {
	int work_lines = 0;

	// Surface VRAM pointer
	n80win32.vram_addr = this->surface_lock();
	if (n80win32.vram_addr == NULL)
		return 0;

	// display graphics (GVRAM)
	if ((n80device.io._o31 & 0x08) == 0x08)
		n80device.blt.drawgvram(m_bpp_mode);
	else
		n80device.blt.blank(m_bpp_mode);

	//TXTDS disabled or DMAC ch.2 not active or CRTC stop display
	if (!(n80device.crtc._text_visible & 0x07)) {
		//if (this->getTxtOnDemand()) {
		//	//`
		//	bool change_attr = false;
		//	bool change_prev = false;
		//	for (int i = 0; i < 25; ++i) {
		//		n80device.vram_sync_enter();
		//		if (n80device.attr._attrchg[i] == 1 || change_prev) {
		//			n80device.attr.cache(i);	// VRAMcachփRs[
		//			n80device.attr._attrchg[i] = 0;
		//			change_attr = true;
		//		} else
		//			change_attr = false;
		//		n80device.vram_sync_leave();
		//		// ͂s(펞)
		//		if (change_attr)
		//			change_prev = n80device.attr.eval(i) == 1 ? true : false;
		//		// cach̓eDirectDraw Surfaceɔf
		//		if (change_attr) {
		//			_blt_font(i, m_bpp_mode);
		//			++work_lines;
		//		}
		//	}
		//} else {						//t`
			for (int i = 0; i < 25; ++i) {
				n80device.vram_sync_enter();
				n80device.attr.cache(i);	// VRAMcachփRs[
				n80device.vram_sync_leave();
				n80device.attr.eval(i);
				n80device.blt.font(i, m_bpp_mode);
				++work_lines;
			}
		//}
	}

	if (m_show_sys_info != 0) {
		HDC hdc = NULL;
		if (SUCCEEDED(m_dds2->GetDC(&hdc))) {
			if (m_show_sys_info == 1) {
				TCHAR str[128];
				double minus = (double)n80win32.remain_minus_cnt / (double)n80win32.loop_cnt;
				double topminus = (double)n80win32.remain_top_minus_cnt / (double)n80win32.loop_cnt;
				double ave = n80win32.remain_ticks_sum / (double)n80win32.loop_cnt;
				double topave = n80win32.remain_ticks_top_sum / (double)n80win32.loop_cnt;
				double minave = n80win32.remain_minus_sum / (double)n80win32.remain_minus_cnt;
				HFONT fFntSave = (HFONT)::SelectObject(hdc, m_font);
				sprintf_s(str, sizeof(str), TEXT("TOP=%.5f(msec) AVE=%.5f(msec) MIAVE=%.5f(msec) MINUS=%.3f/%.3f(%%)"), topave * 1000, ave * 1000, minave * 1000, topminus * 100, minus * 100);
				::TextOut(hdc, 0, 0, str, lstrlen(str));
				::SelectObject(hdc, fFntSave);
			} else {
				m_show_sys_info = 0;
				RECT rct = {0, 0, 340, 20};
				HBRUSH hbr = ::CreateSolidBrush(RGB(0,0,0));
				::FillRect(hdc, &rct, hbr);
				::DeleteObject(hbr);
			}
			m_dds2->ReleaseDC(hdc);
		}
	}
	m_dds2->Unlock(NULL);
	return work_lines;
}

void
n80_directx::reaclcWindowRect() {
	POINT pt = {0, 0};
	::ClientToScreen(m_hwnd, (POINT*)&pt);
	this->m_dd_primary_rect.left   = pt.x;
	this->m_dd_primary_rect.top    = pt.y;
	this->m_dd_primary_rect.right  = pt.x + m_width;
	this->m_dd_primary_rect.bottom = pt.y + m_height;
}

int
n80_directx::transfer() {
	int lines = this->realizer();
	return this->flip(lines);
}

int
n80_directx::flip(int lines) {
	if (wait_vertical_blank_begin()) {
		RECT screen = { 0, 0, m_width, m_height };
		for (;;) {
			HRESULT hr;
			if (this->getFullScreenMode())
				hr = m_dds1->BltFast(0, 0, m_dds2, &screen, DDBLTFAST_NOCOLORKEY);
			else {
				hr = m_dds1->Blt(&m_dd_primary_rect, m_dds2, &screen, DDBLT_WAIT, 0);
			}
			if (hr == DD_OK)
				return lines;
			if (hr == DDERR_SURFACELOST) {
				assay->outputLog(TEXT("DirectX: Surface Lost %s"), this->getFullScreenMode() ? TEXT("[FullScreen]") : TEXT("[WindowMode]"));
				if (m_hwnd == ::GetForegroundWindow()) {
					if (SUCCEEDED(m_dds1->Restore()))
						continue;
					assay->outputLog(TEXT("DirectX: dds1->Restore Failed"));
				} else 
					assay->outputLog(TEXT("DirectX: transfer Failed"));
			}
			break;
		}
	}
	return 0;
}

BOOL
n80_directx::wait_vertical_blank_begin() {
	if (this->getSyncVBlank()) {
		if (SUCCEEDED(m_dd->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL)))
			return true;
		else
			return false;
	}
	return true;
}

BOOL
n80_directx::wait_vertical_blank_end() {
	if (SUCCEEDED(m_dd->WaitForVerticalBlank(DDWAITVB_BLOCKEND, NULL)))
		return true;
	else
		return false;
}

void
n80_directx::draw_text(char text[25][80], char attr[25][80]) {
	// Surface VRAM pointer
	LPBYTE surface = this->surface_lock();
	if (surface == NULL)
		return;
	n80win32.vram_addr = surface;
	n80device.blt.blank(m_bpp_mode);
	for (int r = 0; r < 25; ++r) {
		LPBYTE		txts = &n80device.attr._textcach[0];
		LPBYTE		pacs = &n80device.attr._eval_attr[0];
		n80device.vram_sync_enter();
		for (int c = 0; c < 80; ++c) {
			*txts++ = (BYTE)text[r][c];
			*pacs++ = (BYTE)attr[r][c];
		}
		n80device.vram_sync_leave();
		n80device.blt.font(r, m_bpp_mode);
	}
	HDC hdc = NULL;
	if (SUCCEEDED(m_dds2->GetDC(&hdc))) {
		::DrawIconEx(hdc, 12 + n80win32.vram_left_offset, 16 + n80win32.vram_top_offset, form->getIcon(), 32, 32, 0, NULL, DI_IMAGE);
		m_dds2->ReleaseDC(hdc);
	}
	m_dds2->Unlock(NULL);
	this->flip(25);
}
