/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Win32/DirectX
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		Preference
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	1996/06/27	porting DOS version on WinG
 **		1.20	1997/08/10	porting WinG version on Win32/DirectX
 **		1.30	1998/03/12	
 **		1.43	2007/09/01	XPΉ
 **		1.44	2009/12/15	Vista/7Ή
 **		1.51	2010/01/27	DirectSound OnlyɕύX
 **------------------------------------------------------------------------*/
#include "n80_form.h"
#include "n80_directx.h"
#include "n80_assay.h"
#include "../z80/z80_thread.h"
#include "../res/resource.h"
#include <stdio.h>
#include <commctrl.h>
/**--------------------------------------------------------------------------
 **		O[oϐ
 **------------------------------------------------------------------------*/
extern n80_assay	*assay;
extern n80_form		*form;
extern z80_thread	*th;
extern n80_directx	*dx;
/**--------------------------------------------------------------------------
 **	declare structures
 **------------------------------------------------------------------------*/
typedef void* LPVOID;
typedef struct tagITEM {
	int     type;
	char*   name;
	char*   hint;
	int		value;
	int     (*controller)(struct tagITEM*, int, LPVOID);
    LPVOID  arg;
}	ITEM, *LPITEM;
typedef struct tagTEXTIN_ARG {
    TCHAR*  org;
    TCHAR*  tmp;
    LPVOID  aux1;
    LPVOID  aux2;
}   TEXTIN_ARG, *LPTEXTIN_ARG;
typedef struct tagKEYEVT {
	USHORT  key;
	bool    SHIFT;
	bool    ALT;
	bool    CTL;
	char    ascii;
}   KEYEVTG, * LPKEYEVT;
/**--------------------------------------------------------------------------
 **	declare macro
 **------------------------------------------------------------------------*/
#define TYPE_TOGGLE     (0)     // toggle control
#define TYPE_TEXTIN     (1)     // text-in control
#define TYPE_UPDOWN     (2)     // value up-down control
#define TYPE_TERMINATOR (-1)    // terminator dummy control
#define VERB_INIT       (0)     // INIT control
#define VERB_EVENT      (1)     // Key Event handle
#define VERB_PRINT      (2)     // Print control
#define VERB_PRINT_SEL  (3)     // Print control (selected status)
#define VERB_TOGGLE_SET (4)     // on toggled set value
#define VERB_UPDOWN_SET (4)     // on up-down set value
#define VERB_TEXTIN_CH  (4)     // on key char
#define VERB_TEXTIN_SET (5)     // on text-in set value
 /**--------------------------------------------------------------------------
  **	local function declaration
  **------------------------------------------------------------------------*/
static int _controller_crt(LPITEM, int, LPVOID);
static int _controller_bootrom(LPITEM, int, LPVOID);
static int _controller_tenkeyless(LPITEM, int, LPVOID);
static int _controller_interrace(LPITEM, int, LPVOID);
static int _controller_nowait(LPITEM, int, LPVOID);
static int _controller_n80rom(LPITEM, int, LPVOID);
static int _controller_n88rom(LPITEM, int, LPVOID);
static int _controller_4throm(LPITEM, int, LPVOID);
static int _controller_fdurom(LPITEM, int, LPVOID);
static int _controller_kanjirom1(LPITEM, int, LPVOID);
static int _controller_kanjirom2(LPITEM, int, LPVOID);
static int _controller_gvram(LPITEM, int, LPVOID);
static int _controller_db(LPITEM, int, LPVOID);
static int _controller_ap(LPITEM, int, LPVOID);
#if _DEBUG
static int _controller_debug_main(LPITEM, int, LPVOID);
static int _controller_debug_sub(LPITEM, int, LPVOID);
#endif
/**--------------------------------------------------------------------------
 **	local variables
 **------------------------------------------------------------------------*/
static TCHAR tmp_work[256];
static TCHAR tmp_n80_rom[64];
static TCHAR tmp_n88_rom[64];
static TCHAR tmp_4th_rom[64];
static TCHAR tmp_fdu_rom[64];
static TCHAR tmp_kan1_rom[64];
static TCHAR tmp_kan2_rom[64];
static DWORD tmp_save_screen_mode;
static BYTE	 tmp_save_o31;
static BYTE  tmp_save_text_visible;
static TEXTIN_ARG args[] = {
    {NULL, tmp_n80_rom, "8801-n80", ".ROM"},
    {NULL, tmp_n88_rom, "8801-n88", ".ROM"},
    {NULL, tmp_4th_rom, "8801-4th", ".ROM"},
    {NULL, tmp_fdu_rom, "PDISK", ".ROM"},
    {NULL, tmp_kan1_rom, "KANJI1", ".ROM"},
    {NULL, tmp_kan2_rom, "KANJI2", ".ROM"},
};
static ITEM items[] = {
	{TYPE_TOGGLE, "Boot from N-BASIC ROM ................ [ ]", "select boot ROM from either N80 or N88", 0, _controller_bootrom, NULL},
	{TYPE_TOGGLE, "HiRes-CRT (24kHz/400Lines)............ [ ]", "select CRT from 24kHz or 15kHz(Normal/200Lines)", 0, _controller_crt, NULL},
	{TYPE_TOGGLE, "TEN-Keyless .......................... [ ]", "assign Ten-Key (2/4/6/8) to cursor key for Tenkey-less keyboard", 0, _controller_tenkeyless, NULL},
	{TYPE_TOGGLE, "Interrace-like effect ................ [ ]", "display like interrace monitor", 0, _controller_interrace, NULL},
	//{TYPE_TOGGLE, "No-wait .............................. [ ]", "ignore memory wait, M1 wait only", 0, _controller_nowait, NULL},
	//{TYPE_UPDOWN, "GVRAM penarty......................... [   ](T)", "GVRAM access penarty ticks (range:1..40, defalut:10)", 0, _controller_gvram, NULL},
#if _DEBUG
    {TYPE_TOGGLE, "Debug Main CPU ....................... [ ]", "toggle main cpu debugger", 0, _controller_debug_main, NULL},
    {TYPE_TOGGLE, "Debug Sub CPU ........................ [ ]", "toggle sub cpu debugger", 0, _controller_debug_sub, NULL},
#endif
    {TYPE_UPDOWN, "Volume (HARD)  ....................... [    ](dB)", "Sound Hardware Volume", 0, _controller_db, NULL},
    {TYPE_UPDOWN, "Amplitude (SOFT)  .................... [  ]", "Sound Software Volume", 0, _controller_ap, NULL},
    {TYPE_TEXTIN, "N80 rom file name .................... [                        ]", "N-BASIC ROM file name", 20, _controller_n80rom, &args[0]},
    {TYPE_TEXTIN, "N88 rom file name .................... [                        ]", "N88-BASIC ROM file name", 20, _controller_n88rom, &args[1]},
    {TYPE_TEXTIN, "4th rom file name .................... [                        ]", "4TH ROM file name", 20, _controller_4throm, &args[2]},
    {TYPE_TEXTIN, "Disk Unit rom file name .............. [                        ]", "Disk ROM file name", 20, _controller_fdurom, &args[3]},
    {TYPE_TEXTIN, "Kanji #1 rom file name ............... [                        ]", "KANJI #1 ROM file name", 20, _controller_kanjirom1, &args[4]},
    {TYPE_TEXTIN, "Kanji #2 rom file name ............... [                        ]", "KANJI #2 ROM file name", 20, _controller_kanjirom2, &args[5]},
    {TYPE_TERMINATOR}
};
#pragma region Windows Virtual Key<=>ASCII Code
static char to_ascii(LPKEYEVT kv) {
    switch (kv->key) {
    case 0x31:	/* 1 */	return kv->SHIFT ? 0x21 : 0x31;
    case 0x32:	/* 2 */	return kv->SHIFT ? 0x22 : 0x32;
    case 0x33:	/* 3 */	return kv->SHIFT ? 0x23 : 0x33;
    case 0x34:	/* 4 */	return kv->SHIFT ? 0x24 : 0x34;
    case 0x35:	/* 5 */	return kv->SHIFT ? 0x24 : 0x35;
    case 0x36:	/* 6 */	return kv->SHIFT ? 0x26 : 0x36;
    case 0x37:	/* 7 */	return kv->SHIFT ? 0x27 : 0x37;
    case 0x38:	/* 8 */	return kv->SHIFT ? 0x28 : 0x38;
    case 0x39:	/* 9 */	return kv->SHIFT ? 0x29 : 0x39;
    case 0x30:	/* 0 */	return kv->SHIFT ? 0x30 : 0x30;

    case 0x51:	/* Q */	return kv->SHIFT ? 0x51 : 0x71;
    case 0x57:	/* W */	return kv->SHIFT ? 0x57 : 0x77;
    case 0x45:	/* E */	return kv->SHIFT ? 0x45 : 0x65;
    case 0x52:	/* R */	return kv->SHIFT ? 0x52 : 0x72;
    case 0x54:	/* T */	return kv->SHIFT ? 0x54 : 0x74;
    case 0x59:	/* Y */	return kv->SHIFT ? 0x59 : 0x79;
    case 0x55:	/* U */	return kv->SHIFT ? 0x55 : 0x75;
    case 0x49:	/* I */	return kv->SHIFT ? 0x49 : 0x69;
    case 0x4F:	/* O */	return kv->SHIFT ? 0x4F : 0x6F;
    case 0x50:	/* P */	return kv->SHIFT ? 0x50 : 0x70;

    case 0x41:	/* A */	return kv->SHIFT ? 0x41 : 0x61;
    case 0x53:	/* S */	return kv->SHIFT ? 0x53 : 0x73;
    case 0x44:	/* D */	return kv->SHIFT ? 0x44 : 0x64;
    case 0x46:	/* F */	return kv->SHIFT ? 0x46 : 0x66;
    case 0x47:	/* G */	return kv->SHIFT ? 0x47 : 0x67;
    case 0x48:	/* H */	return kv->SHIFT ? 0x48 : 0x68;
    case 0x4A:	/* J */	return kv->SHIFT ? 0x4A : 0x6A;
    case 0x4B:	/* K */	return kv->SHIFT ? 0x4B : 0x6B;
    case 0x4C:	/* L */	return kv->SHIFT ? 0x4C : 0x6C;

    case 0x5A:	/* Z */	return kv->SHIFT ? 0x5A : 0x7A;
    case 0x58:	/* X */	return kv->SHIFT ? 0x58 : 0x78;
    case 0x43:	/* C */	return kv->SHIFT ? 0x43 : 0x63;
    case 0x56:	/* V */	return kv->SHIFT ? 0x56 : 0x76;
    case 0x42:	/* B */	return kv->SHIFT ? 0x42 : 0x62;
    case 0x4E:	/* N */	return kv->SHIFT ? 0x4E : 0x6E;
    case 0x4D:	/* M */	return kv->SHIFT ? 0x4D : 0x6D;

    case VK_ESCAPE:     /*ESC*/     return 0x1B;
    case VK_BACK:       /* BS*/     return 0x08;
    case VK_DELETE:     /*DEL*/     return 0x08;
    case VK_TAB:        /*TAB*/     return 0x09;
    case VK_RETURN:     /*RET*/     return 0x0D;

    case VK_NUMPAD0:	/* 0 */     return 0x30;
    case VK_NUMPAD1:	/* 1 */     return 0x31;
    case VK_NUMPAD2:	/* 2 */     return 0x32;
    case VK_NUMPAD3:	/* 3 */     return 0x33;
    case VK_NUMPAD4:	/* 4 */     return 0x34;
    case VK_NUMPAD5:	/* 5 */     return 0x35;
    case VK_NUMPAD6:	/* 6 */     return 0x36;
    case VK_NUMPAD7:	/* 7 */     return 0x37;
    case VK_NUMPAD8:	/* 8 */     return 0x38;
    case VK_NUMPAD9:	/* 9 */     return 0x39;
    case VK_MULTIPLY:	/* * */     return 0x2A;
    case VK_ADD:    	/* + */     return 0x2B;
    case VK_SEPARATOR:	/* , */     return 0x2C;
    case VK_SUBTRACT:	/* - */     return 0x2D;
    case VK_DECIMAL:	/* . */     return 0x2E;
    case VK_DIVIDE:	    /* / */     return 0x2F;

    case VK_OEM_MINUS:	/* = -   */	return kv->SHIFT ? 0x3D : 0x2D;
    case VK_OEM_7:	    /* ^ ~   */	return kv->SHIFT ? 0x7E : 0x5E;
    case VK_OEM_3:	    /* @ `   */	return kv->SHIFT ? 0x60 : 0x40;
    case VK_OEM_4:	    /* [ {   */	return kv->SHIFT ? 0x7B : 0x5B;
    case VK_OEM_PLUS:	/* + ;   */	return kv->SHIFT ? 0x2B : 0x3B;
    case VK_OEM_1:	    /* * :   */	return kv->SHIFT ? 0x2A : 0x3A;
    case VK_OEM_6:	    /* ] }   */	return kv->SHIFT ? 0x7D : 0x5D;
    case VK_OEM_COMMA:  /* < ,   */	return kv->SHIFT ? 0x3C : 0x2C;
    case VK_OEM_PERIOD: /* > .   */	return kv->SHIFT ? 0x3E : 0x2E;
    case VK_OEM_2:	    /* / ?   */	return kv->SHIFT ? 0x3F : 0x2F;
    case VK_OEM_102:	/* \ _   */	return kv->SHIFT ? 0x5F : 0x5C;
    case VK_OEM_5:      /* \ |   */	return kv->SHIFT ? 0x7C : 0x5C;
    default:    return '\0';
    }
}
#pragma endregion
static int  item_cnt;
static int  cur_idx;
static char save_cmt_name[12];
static char save_cmt_dir[12];
static KEYEVTG keyevt;
/**--------------------------------------------------------------------------
 **	functions
 **------------------------------------------------------------------------*/
static bool is_alpha(LPKEYEVT kv) {
    if (kv->ascii >= 'a' && kv->ascii <= 'z')
        return true;
    if (kv->ascii >= 'A' && kv->ascii <= 'Z')
        return true;
    return false;
}
static bool is_numeric(LPKEYEVT kv) {
    return (kv->ascii >= '0' && kv->ascii <= '9');
}
#pragma region toggle control methods
static int _toggle_event(LPITEM self, LPKEYEVT key) {
    if (key->key == VK_SPACE) { // [SP] Key pressed
        self->value = self->value == 1 ? 0 : 1;
        self->controller(self, VERB_TOGGLE_SET, NULL);
        return 1;
    }
    else
        return 0;
}
static void _toggle_print(LPITEM self, char* buff) {
    memset(buff, 32, 78);
    strcpy(buff, self->name);
    if (self->value == 1)
        buff[40] = 'X';
}
#pragma endregion
#pragma region text input control methods
static void _textin_filename_init(LPITEM self, int maxlen) {
    LPTEXTIN_ARG arg = (LPTEXTIN_ARG)self->arg;
    memset(arg->tmp, 0, maxlen);
    strcpy(arg->tmp, arg->org);
}
static int _textin_event(LPITEM self, LPKEYEVT kv) {
    LPTEXTIN_ARG arg = (LPTEXTIN_ARG)self->arg;
    char* buff = arg->tmp;
    int maxlen = self->value - 1;
    size_t len;
    if (kv->key == VK_TAB || kv->key == VK_UP || kv->key == VK_DOWN) // [TAB] | [ESC] Key pressed
        return self->controller(self, VERB_TEXTIN_SET, kv);
    len = strlen(buff);
    if (len > (size_t)maxlen)
        return 0;
    if (kv->key == VK_BACK && len > 0) { // [BS] Key pressed
        buff[len - 1] = '\0';
        return 1;
    }
    else if (len < (size_t)maxlen && self->controller(self, VERB_TEXTIN_CH, kv) == 1) {
        buff[len] = kv->ascii;
        buff[len + 1] = '\0';
    }
    return 1;
}
static void _textin_print(LPITEM self, TCHAR* vram, bool selected) {
    LPTEXTIN_ARG arg = (LPTEXTIN_ARG)self->arg;
    int maxlen = self->value - 1;
    char* src = arg->tmp;
    size_t l = strlen(src);
    memset(vram, 0x20, 78);
    strcpy(vram, self->name);           // item title format
    for (int i = 0; i < maxlen; ++i) {  // item value (string)
        char ch = *src++;
        if (ch == '\0')
            break;
        vram[40 + i] = ch;
    }
    if (selected) { // show guide cursor at eot when selected
        if (l < (size_t)maxlen)
            vram[40 + l] = 0x1D;
    }
}
static int _textin_filename_ch(LPITEM self, LPKEYEVT kv) {
    LPTEXTIN_ARG arg = (LPTEXTIN_ARG)self->arg;
    if (is_alpha(kv) || is_numeric(kv))
        return 1;
    if (kv->ascii == '_' || kv->ascii == '.')
        return 1;
    if (kv->ascii == '-' && strlen(arg->tmp) > 0)
        return 1;
    return 0;
}
static int _textin_filename_set(LPITEM self) {
    LPTEXTIN_ARG arg = (LPTEXTIN_ARG)self->arg;
    memset(arg->org, 0, self->value);
    char* fn = strlen(arg->tmp) > 0 ? arg->tmp : (char*)arg->aux1;
    if (strlen(fn) == 0)
        return 0;
    strcat(arg->org, fn);
    //strcat(arg->org, (char*)arg->aux2);
    return 0;
}
#pragma endregion
#pragma region value control methods
static int _updown_event(LPITEM self, LPKEYEVT key, int minvalue, int maxvalue) {
    if (key->key == VK_LEFT) {
        if (self->value > minvalue) {
            --self->value;
            self->controller(self, VERB_UPDOWN_SET, NULL);
            return 1;
        }
    }
    else if (key->key == VK_RIGHT) {
        if (self->value < maxvalue) {
            ++self->value;
            self->controller(self, VERB_UPDOWN_SET, NULL);
            return 1;
        }
    }
    return 0;
}
static void _updown_print(LPITEM self, char* buff, int raster, int maxlen) {
    char tmp[64];
    memset(buff, 32, 78);
    strcpy(buff, self->name);
    memset(tmp, 0, 64);
    sprintf(tmp, "%d", self->value);
    char* ptr = &buff[raster];
    int len = strlen(tmp);
    if (len < maxlen) {
        for (int i = 0; i < (maxlen - len); ++i)
            *ptr++ = 32;
    }
    strcpy(ptr, tmp);
    buff[raster + maxlen] = ']';
}
#pragma endregion
#pragma region config items
static int _controller_crt(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = n80device.vrtc.crt_hires_mode == _CRT_HIRES ? 1 : 0;
        break;
    case VERB_EVENT:
        return _toggle_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _toggle_print(self, (char*)param);
        break;
    case VERB_TOGGLE_SET:
        n80device.vrtc.crt_hires_mode = self->value == 1 ? _CRT_HIRES : _CRT_NORMAL;
        break;
    }
    return 0;
}
static int _controller_bootrom(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = n80device.param.dipsw1.SW1_N_ROM == _BOOT_ROM_N80 ? 1 : 0;
        break;
    case VERB_EVENT:
        return _toggle_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _toggle_print(self, (char*)param);
        break;
    case VERB_TOGGLE_SET:
        n80device.param.dipsw1.SW1_N_ROM = self->value == 1 ? _BOOT_ROM_N80 : _BOOT_ROM_N88;
        break;
    }
    return 0;
}
static int _controller_tenkeyless(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = n80win32.tenkeyless ? 1 : 0;
        break;
    case VERB_EVENT:
        return _toggle_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _toggle_print(self, (char*)param);
        break;
    case VERB_TOGGLE_SET:
        n80win32.set_tenkeyless(self->value == 1);
        break;
    }
    return 0;
}
static int _controller_interrace(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = n80device.param.interlace ? 1 : 0;
        break;
    case VERB_EVENT:
        return _toggle_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _toggle_print(self, (char*)param);
        break;
    case VERB_TOGGLE_SET:
        n80device.param.interlace = self->value == 1;
        n80device.blt.chg_palette(n80device.param.interlace);
        break;
    }
    return 0;
}
static int _controller_nowait(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = n80win32.m_z80_no_wait ? 1 : 0;
        break;
    case VERB_EVENT:
        return _toggle_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _toggle_print(self, (char*)param);
        break;
    case VERB_TOGGLE_SET:
        n80win32.m_z80_no_wait = self->value == 1;
        break;
    }
    return 0;
}
#if _DEBUG
static int _controller_debug_main(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = n80device.cpu.dbgctx.enable;
        break;
    case VERB_EVENT:
        return _toggle_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _toggle_print(self, (char*)param);
        break;
    case VERB_TOGGLE_SET:
        n80device.cpu.dbgctx.debug_enable(&n80device.cpu.dbgctx, (BYTE)self->value);
        break;
    }
    return 0;
}
static int _controller_debug_sub(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = n80device.fdu.cpu.dbgctx.enable ? 1 : 0;
        break;
    case VERB_EVENT:
        return _toggle_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _toggle_print(self, (char*)param);
        break;
    case VERB_TOGGLE_SET:
        n80device.fdu.cpu.dbgctx.debug_enable(&n80device.fdu.cpu.dbgctx, (BYTE)self->value);
        break;
    }
    return 0;
}
#endif
static int _controller_gvram(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = n80device.param.gv_wait_ticks;
        break;
    case VERB_EVENT:
        return _updown_event(self, (LPKEYEVT)param, 0, 40);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _updown_print(self, (char*)param, 40, 3);
        break;
    case VERB_TOGGLE_SET:
        n80device.param.gv_wait_ticks = self->value;
        break;
    }
    return 0;
}
static int _controller_db(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = (short)(n80win32.get_hard_volume() - 1);
        break;
    case VERB_EVENT:
        return _updown_event(self, (LPKEYEVT)param, -100, 0);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _updown_print(self, (char*)param, 40, 4);
        break;
    case VERB_TOGGLE_SET:
        n80win32.set_hard_volume((short)self->value);
        break;
    }
    return 0;
}
static int _controller_ap(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:
        self->value = (short)n80win32.get_soft_volume();
        break;
    case VERB_EVENT:
        return _updown_event(self, (LPKEYEVT)param, 1, 18);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _updown_print(self, (char*)param, 40, 2);
        break;
    case VERB_TOGGLE_SET:
        n80win32.set_soft_volume((short)self->value);
        break;
    }
    return 0;
}
static int _controller_n80rom(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:     //erace file extension
        _textin_filename_init(self, self->value);
        break;
    case VERB_EVENT:
        return _textin_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _textin_print(self, (char*)param, verb == VERB_PRINT_SEL);
        break;
    case VERB_TEXTIN_CH:
        return _textin_filename_ch(self, (LPKEYEVT)param);
    case VERB_TEXTIN_SET:
        return _textin_filename_set(self);
    }
    return 0;
}
static int _controller_n88rom(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:     //erace file extension
        _textin_filename_init(self, self->value);
        break;
    case VERB_EVENT:
        return _textin_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _textin_print(self, (char*)param, verb == VERB_PRINT_SEL);
        break;
    case VERB_TEXTIN_CH:
        return _textin_filename_ch(self, (LPKEYEVT)param);
    case VERB_TEXTIN_SET:
        return _textin_filename_set(self);
    }
    return 0;
}
static int _controller_4throm(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:     //erace file extension
        _textin_filename_init(self, self->value);
        break;
    case VERB_EVENT:
        return _textin_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _textin_print(self, (char*)param, verb == VERB_PRINT_SEL);
        break;
    case VERB_TEXTIN_CH:
        return _textin_filename_ch(self, (LPKEYEVT)param);
    case VERB_TEXTIN_SET:
        return _textin_filename_set(self);
    }
    return 0;
}
static int _controller_fdurom(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:     //erace file extension
        _textin_filename_init(self, self->value);
        break;
    case VERB_EVENT:
        return _textin_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _textin_print(self, (char*)param, verb == VERB_PRINT_SEL);
        break;
    case VERB_TEXTIN_CH:
        return _textin_filename_ch(self, (LPKEYEVT)param);
    case VERB_TEXTIN_SET:
        return _textin_filename_set(self);
    }
    return 0;
}
static int _controller_kanjirom1(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:     //erace file extension
        _textin_filename_init(self, self->value);
        break;
    case VERB_EVENT:
        return _textin_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _textin_print(self, (char*)param, verb == VERB_PRINT_SEL);
        break;
    case VERB_TEXTIN_CH:
        return _textin_filename_ch(self, (LPKEYEVT)param);
    case VERB_TEXTIN_SET:
        return _textin_filename_set(self);
    }
    return 0;
}
static int _controller_kanjirom2(LPITEM self, int verb, LPVOID param) {
    switch (verb) {
    case VERB_INIT:     //erace file extension
        _textin_filename_init(self, self->value);
        break;
    case VERB_EVENT:
        return _textin_event(self, (LPKEYEVT)param);
    case VERB_PRINT:
    case VERB_PRINT_SEL:
        _textin_print(self, (char*)param, verb == VERB_PRINT_SEL);
        break;
    case VERB_TEXTIN_CH:
        return _textin_filename_ch(self, (LPKEYEVT)param);
    case VERB_TEXTIN_SET:
        return _textin_filename_set(self);
    }
    return 0;
}
#pragma endregion
/**--------------------------------------------------------------------------
 **		_CAO֐
 **------------------------------------------------------------------------*/
LRESULT
n80_form::cfgWndproc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp) {
	switch (msg) {
	case WM_QUERYENDSESSION:		// Windows I𓾂悤ƂĂꍇB
	case WM_CLOSE:					// Window
	case WM_ENDSESSION:         	// VXeI̓AvP[VI
	case WM_DESTROY:	        	// jꂽꍇAAvP[VI
		leave_config_mode();
		return 1;
	case  WM_PAINT: {
		PAINTSTRUCT	ps;
		HDC hdc = ::BeginPaint(mWnd, &ps);
		th->getRender()->postMessage(WM_REQ_DRAW_CUI, (WPARAM)this, NULL);
		::EndPaint(mWnd, &ps);
		return 0;
	}
	case WM_SYSKEYDOWN:
	case WM_KEYDOWN:
        keyevt.key = (USHORT)(wp & 0xffff);
        keyevt.ascii = to_ascii(&keyevt);
//#if _DEBUG
//        n80device.logger("VK = %X, ASCII = %X\n", keyevt.key, keyevt.ascii);
//#endif
        if (wp == VK_SHIFT)
            keyevt.SHIFT = TRUE;
        if (wp == VK_ESCAPE && !(lp & 0x40000000)) {
            leave_config_mode();
        }
        else if (items[cur_idx].controller(&items[cur_idx], VERB_EVENT, &keyevt) == 1) {
            drawitem(&items[cur_idx], cur_idx, true);
            th->getRender()->postMessage(WM_REQ_DRAW_CUI, (WPARAM)this, NULL);
        }
        else if ((wp == VK_TAB && !keyevt.SHIFT) || wp == VK_DOWN) { // [TAB] or [DOWN]
            if (item_cnt > 1) {
                items[cur_idx].controller(&items[cur_idx], VERB_INIT, NULL);
                drawitem(&items[cur_idx], cur_idx, false);
                if (++cur_idx >= item_cnt)
                    cur_idx = 0;
                drawitem(&items[cur_idx], cur_idx, true);
                th->getRender()->postMessage(WM_REQ_DRAW_CUI, (WPARAM)this, NULL);
            }
        }
        else if ((wp == VK_TAB && keyevt.SHIFT) || wp == VK_UP) { // [UP]
            if (item_cnt > 1) {
                items[cur_idx].controller(&items[cur_idx], VERB_INIT, NULL);
                drawitem(&items[cur_idx], cur_idx, false);
                if (--cur_idx < 0)
                    cur_idx = item_cnt - 1;
                drawitem(&items[cur_idx], cur_idx, true);
                th->getRender()->postMessage(WM_REQ_DRAW_CUI, (WPARAM)this, NULL);
            }
        }
        break;
	case WM_SYSKEYUP:
	case WM_KEYUP:
        if (wp == VK_SHIFT)
            keyevt.SHIFT = FALSE;
        break;
    case WM_SYSCOMMAND:
        break;
    default:
		return 1;
	}
	return 0;
}
/**--------------------------------------------------------------------------
 **		Config
 **------------------------------------------------------------------------*/
void
n80_form::enter_config_mode() {
	if (!m_config_mode && th->sync()) {
		m_config_mode = TRUE;
		dx->stop();
		ShowCursorFullScreen();
		::SetFocus(mWnd);
        // Z80syncŒ~ꂽRenderĊJ
        th->getRender()->desync();

		m_key_combo = KEYCOMB_CLEAR;

		tmp_save_screen_mode = n80device.crtc._screen_mode;
		tmp_save_o31 = n80device.io._o31;
		tmp_save_text_visible = n80device.crtc._text_visible;

		n80device.blt.chg_crt_mode(3);		// 80x25
		n80device.io._o31 &= ~0x08;			// gvram not visible
		n80device.crtc._text_visible = 0;	// text visible

		for (int r = 0; r < 25; ++r) {
			for (int c = 0; c < 80; ++c) {
				BYTE atr = 0x23;
				if ((r == 4 || r == 5) && (c > 0 && c < 79))
					atr = 0x27;
				else if (r == 7 && (c > 0 && c < 79))
					atr = 0x26;
				else if ((r > 8 && r < 24) && (c > 0 && c < 79))
					atr = 0x07;
				m_cui_attr[r][c] = atr;
				m_cui_text[r][c] = 32;
			}
		}
        item_cnt = 0, cur_idx = 0;
        const char* title1 = "    \xea\xeaN88\xea\xea PC-8801 Emulator for DirectX";
		const char* title2 = "    Copyright 1993-2024 (C) H.Kanemoto, All Right Reserved.";
		const char* title3 = "System Config";
		const char* title4 = "usage:  [ESC] Quit, [\x1e\x1f] Move, [SP] Toggle, [\x1d\x1c] Value";
		const char* title5 = "hint: ";

        args[0].org = n80win32.n80_rom_file_name;
        args[1].org = n80win32.n88_rom_file_name;
        args[2].org = n80win32._4th_rom_file_name;
        args[3].org = n80win32.fdu_rom_file_name;
        args[4].org = n80win32.kanji1_rom_file_name;
        args[5].org = n80win32.kanji2_rom_file_name;
        for (int r = 0; items[r].type != TYPE_TERMINATOR; ++r) {
            ++item_cnt;
            items[r].controller(&items[r], VERB_INIT, NULL);
            drawitem(&items[r], r, r == 0);
        }
        memcpy(&m_cui_text[1][2], title1, strlen(title1));
		memcpy(&m_cui_text[2][2], title2, strlen(title2));
		memcpy(&m_cui_text[4][2], title3, strlen(title3));
		memcpy(&m_cui_text[5][2], title4, strlen(title4));
		memcpy(&m_cui_text[7][2], title5, strlen(title5));

		th->getRender()->postMessage(WM_USER + 3, (WPARAM)this, NULL);
	}
}
void
n80_form::leave_config_mode() {
	if (m_config_mode) {
		m_config_mode = FALSE;

		n80device.crtc._screen_mode = tmp_save_screen_mode;
		n80device.io._o31 = tmp_save_o31;
		n80device.crtc._text_visible = tmp_save_text_visible;
		n80device.blt.chg_crt_mode(n80device.crtc._screen_mode);

        n80win32.hard_volume = n80win32.get_hard_volume() - 1;
        n80win32.soft_volume = n80win32.get_soft_volume();
        saveProfile(TEXT("Default"));
        
        HideCursorFullScreen();
		::SetFocus(mWnd);
		n80device.attr.invalid();
		th->desync();
		dx->play();
	}
}
void
n80_form::drawitem(LPVOID _item, int line, bool selected) {
    LPITEM item = (LPITEM)_item;
    int row = line + 9;
    item->controller(item, selected ? VERB_PRINT_SEL : VERB_PRINT, &m_cui_text[row][1]);
    for (int i = 1; i < 79; ++i)
        m_cui_attr[row][i] = selected ? 0x22 : 0x07;
    if (selected) {
        memset(&m_cui_text[7][1], 32, 78);
        sprintf(&m_cui_text[7][1], " hint: %s", item->hint);
    }
}
