/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Win32/DirectX
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		Load N80/CMT Files
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	1996/06/27	porting DOS version on WinG
 **		1.20	1997/08/10	porting WinG version on Win32/DirectX
 **		1.30	1998/03/12	
 **		1.43	2007/09/01	XPΉ
 **		1.44	2009/12/15	Vista/7Ή
 **		1.51	2010/01/27	DirectSound OnlyɕύX
 **------------------------------------------------------------------------*/
#include "n80_form.h"
#include "n80_directx.h"
#include "../pc-8001/pc-8001.h"
#include "../z80/z80_thread.h"
#include "../lib/icj_stream.h"
#include "../lib/icj_memory.h"
#include <string.h>
#include <crtdbg.h>
/**--------------------------------------------------------------------------
 **		O[oϐ
 **------------------------------------------------------------------------*/
extern z80_thread	*th;
extern n80_directx	*dx;
/**--------------------------------------------------------------------------
 **		萔`
 **------------------------------------------------------------------------*/
#define  SP_VAL   0x7f3e
#define  PC_VAL   0xff3d
#define  STDSIZE  32576
#define  MAXSIZE  32768
/**--------------------------------------------------------------------------
 **		N80/CMT/D88 t@CI
 **------------------------------------------------------------------------*/
void
n80_form::chooser(BYTE mode, int unit)
{
	char	path_name[512];

	if (th->sync()) {
		dx->stop();
		ShowCursorFullScreen();
		::InvalidateRect(mWnd, NULL, TRUE);

		BOOL result = false;
		if (open_file(mode, path_name)) {
			if (strstr(path_name, ".N80") || strstr(path_name, ".n80"))
				read_file(path_name);
			else if (strstr(path_name, ".D88") || strstr(path_name, ".d88"))
				read_d88_file(path_name, unit);
			else if (strstr(path_name, ".CMT") || strstr(path_name, ".cmt"))
				result = cach_file(path_name);
			else if (strstr(path_name, ".T88") || strstr(path_name, ".t88"))
				result = conv_file(path_name);
		}
		n80device.cmt.notify(result ? 1 : 0);

		HideCursorFullScreen();
		::SetFocus(mWnd);
		n80device.attr.invalid();
		n80device.io.clear_keys();
		th->desync();
		dx->play();
	}
}
void
n80_form::drop_file(LPCTSTR path_name, int unit) {
	if (th->sync()) {
		dx->stop();
		ShowCursorFullScreen();
		::InvalidateRect(mWnd, NULL, TRUE);

		BOOL result = false;
		if (strstr(path_name, ".N80") || strstr(path_name, ".n80"))
			read_file(path_name);
		else if (strstr(path_name, ".D88") || strstr(path_name, ".d88"))
			read_d88_file(path_name, unit);
		else if (strstr(path_name, ".CMT") || strstr(path_name, ".cmt"))
			result = cach_file(path_name);
		else if (strstr(path_name, ".T88") || strstr(path_name, ".t88"))
			result = conv_file(path_name);
		n80device.cmt.notify(result ? 1 : 0);

		HideCursorFullScreen();
		::SetFocus(mWnd);
		n80device.attr.invalid();
		n80device.io.clear_keys();
		th->desync();
		dx->play();
	}
}
/**--------------------------------------------------------------------------
 **		D88 CWFNg
 **------------------------------------------------------------------------*/
void
n80_form::fd_eject(int unit)
{
	LPFDUDRIVE drv = &n80device.fdu.fdd.drives[unit];

	if (th->sync()) {
		dx->stop();
		drv->eject(drv);
		th->desync();
		dx->play();
	}
}
/**--------------------------------------------------------------------------
 **		N80/CMT t@CI
 **------------------------------------------------------------------------*/
BOOL
n80_form::open_file(BYTE mode, LPSTR szPath)
{
	OPENFILENAME	ofn;
	char	szName[16];
	char	szCurDir[512];

	::ZeroMemory(&ofn, sizeof(OPENFILENAME));
	::GetCurrentDirectory(sizeof (szCurDir), szCurDir);
	szName[0] = 0x00;
	szPath[0] = 0x00;

	ofn.lStructSize			= sizeof(OPENFILENAME); 
	ofn.hwndOwner			= mWnd;
	ofn.hInstance			= NULL;
	ofn.lpstrFilter			= mode == CHOOSER_MODE_TAPE ? TEXT("N80 ۸̧\0*.N80;*.CMT;*.T88;*.BIN\0") : TEXT("D88 ۸̧\0*.D88;\0\0");
	ofn.lpstrCustomFilter	= NULL; 
	ofn.nMaxCustFilter		= 0;
	ofn.nFilterIndex		= 0; 
	ofn.lpstrFile			= szPath;
	ofn.nMaxFile			= 512;
	ofn.lpstrFileTitle		= szName; 
	ofn.nMaxFileTitle		= sizeof(szName); 
	ofn.lpstrInitialDir		= szCurDir;
	ofn.lpstrTitle			= mode == CHOOSER_MODE_TAPE ? TEXT("N80 ̧ق̑I") : TEXT("D88 ̧ق̑I");
	ofn.Flags				= OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST; 
	ofn.nFileOffset			= 0; 
	ofn.nFileExtension		= 0; 
	ofn.lpstrDefExt			= mode == CHOOSER_MODE_TAPE ? TEXT("N80") : TEXT("D88");
	ofn.lCustData			= 0; 
	ofn.lpfnHook			= NULL; 
	ofn.lpTemplateName		= NULL;

	return  ::GetOpenFileName(&ofn);
}
/**--------------------------------------------------------------------------
 **		N80/CMT t@Cǂݍ
 **------------------------------------------------------------------------*/
BOOL
n80_form::read_file(LPCSTR path_name)
{
	icj_fstream	fs;
	DWORD		fsiz, i, W;
	USHORT		spval, addr;
	BOOL		result = false;

	if (fs.open(path_name, TRUE)) {
		fsiz = fs.getSize();
		if (fsiz >= STDSIZE) {
			fsiz = (fsiz > MAXSIZE) ? MAXSIZE : fsiz;
			for (i = 0, addr = 0x8000; i < fsiz; ++i) {
				BYTE b;
				fs.read(&b, 1, 1);
				n80device.mem.write(&n80device.cpu, addr++, b, &W);
			}
			//fs.read(&n80device.mem._ram[0x8000], fsiz, 1);
			fs.seek(FILE_BEGIN, SP_VAL);
			fs.read(&spval, 2, 1);
			th->restart(PC_VAL, spval);
			result = true;
		}
		fs.close();
	}
	return result;
}
/**--------------------------------------------------------------------------
 **		cacht@C
 **------------------------------------------------------------------------*/
BOOL
n80_form::cach_file(LPCSTR path_name)
{
	icj_fstream	fs;
	BOOL		result = false;

	if (fs.open(path_name, TRUE)) {
		if (n80device.cmt._buff) {
			xfree(n80device.cmt._buff);
			n80device.cmt._buff = NULL;
		}
		n80device.cmt._size = fs.getSize();
		n80device.cmt._buff = (LPBYTE) xmalloc(n80device.cmt._size);
		if (n80device.cmt._buff) {
			fs.read(n80device.cmt._buff, n80device.cmt._size, 1);
			n80device.cmt._index = 0;
			result = true;
		}
		fs.close();
	}
	return result;
}
/**--------------------------------------------------------------------------
 **		BIN t@Cǂݍ
 **------------------------------------------------------------------------*/
void
n80_form::membin_file(LPCSTR path_name)
{
	icj_fstream	fs;
	DWORD		fsiz, W;
	BYTE		buff[32 * 1024];

	if (fs.open(path_name, TRUE)) {
		fsiz = fs.getSize();
		if (fsiz <= 32*1024) {
			//fs.read(&n80device.mem._ram[0x8000], fsiz, 1);
			if (fs.read(buff, fsiz, 1) == fsiz) {
				USHORT addr = 0x8000;
				for (DWORD i = 0; i < fsiz; ++i)
					n80device.mem.write(&n80device.cpu, addr++, buff[i], &W);
			}
		}
		fs.close();
	}
}
/**--------------------------------------------------------------------------
 **		D88 t@Cǂݍ
 **------------------------------------------------------------------------*/
static void d88_image_release(LPDISKIMAGE image) {
	if (--image->ref_count <= 0) {
		image->ref_count = 0;
		if (image->image_raw_data != NULL)
			xfree(image->image_raw_data);
		for (int i = 0; i < image->disk_count; ++i) {
			if (image->disks[i] != NULL) {
				xfree(image->disks[i]);
				image->disks[i] = NULL;
			}
		}
		xfree(image);
	}
}
static LPFLOPPYDISK d88_image_alloc_disk() {
	return (LPFLOPPYDISK)xmalloc(sizeof(FLOPPYDISK));
}
BOOL n80_form::read_d88_file(LPCTSTR path, int unit) {
	BOOL result		= FALSE;
	LPFDU fdu		= &n80device.fdu;
	icj_fstream	rom;

	if (!rom.open(path, TRUE))
		return  FALSE;

	LPDISKIMAGE image = (LPDISKIMAGE)xmalloc(sizeof(DISKIMAGE));
	if (image != NULL) {
		DWORD fsiz = rom.getSize();
		image->image_raw_data = (LPBYTE)xmalloc(fsiz);
		if (image->image_raw_data != NULL) {
			DWORD rsiz = rom.read(image->image_raw_data, fsiz, 1);
			if (fsiz == rsiz) {
				image->image_length = rsiz;
				image->dispose		= d88_image_release;
				image->alloc		= d88_image_alloc_disk;
				if (fdu->media_init(fdu, image)) {
					LPFDUDRIVE drv = &fdu->fdd.drives[unit];
					if (image->disk_count > 1) {
						drv = &fdu->fdd.drives[0];
						drv->eject(drv);
						result = drv->insert(drv, image->disks[0]);
						if (result) {
							drv = &fdu->fdd.drives[1];
							drv->eject(drv);
							result = drv->insert(drv, image->disks[1]);
						}
					} else {
						drv->eject(drv);
						result = drv->insert(drv, image->disks[0]);
					}
				}
			}
		}
		if (!result)
			d88_image_release(image);
	}
	rom.close();
	return result;
}
/**--------------------------------------------------------------------------
 **		T88 t@Cǂݍ
 **------------------------------------------------------------------------*/
BOOL
n80_form::conv_file(LPCSTR path_name)
{
	icj_fstream	fs;
	const char* header = "PC-8801 Tape Image(T88)";
	char buff[23 + 1];
	int  raster = 0;
	BOOL result = false;

	if (fs.open(path_name, TRUE)) {
		if (n80device.cmt._buff) {
			xfree(n80device.cmt._buff);
			n80device.cmt._buff = NULL;
		}
		n80device.cmt._size = fs.getSize();
		n80device.cmt._buff = (LPBYTE)xmalloc(n80device.cmt._size);
		for (; n80device.cmt._buff;) {
			// file header check
			if (fs.read(buff, 24, 1) != 24)
				break;
			if (strcmp(header, buff) != 0)
				break;
			// skip version tag
			if (fs.read(buff, 6, 1) != 6)
				break;
			while (true) {
				// tag + siz read
				if (fs.read(buff, 4, 1) != 4)
					break;
				int tag_siz = 0;
				tag_siz += ((BYTE)buff[3]);
				tag_siz <<= 8;
				tag_siz += ((BYTE)buff[2]);
				// blank/space/mark tags
				if ((buff[0] == 0x00 && buff[1] == 0x01) ||
					(buff[0] == 0x02 && buff[1] == 0x01) ||
					(buff[0] == 0x03 && buff[1] == 0x01)) {
					if (fs.read(buff, tag_siz, 1) != tag_siz)
						break;
				// data tag
				} else if (buff[0] == 0x01 && buff[1] == 0x01) {
					// area time/area length/data length/data type
					if (fs.read(buff, 12, 1) != 12)
						break;
					int data_siz = 0;
					data_siz += ((BYTE)buff[9]);
					data_siz <<= 8;
					data_siz += ((BYTE)buff[8]);
					// check buffer size
					if ((raster + data_siz) > (int)n80device.cmt._size)
						break;
					// read data
					if (fs.read(&n80device.cmt._buff[raster], data_siz, 1) != data_siz)
						break;
					raster += data_siz;
				// end tag
				} else if (buff[0] == 0x00 && buff[1] == 0x00) {
					n80device.cmt._size = raster;
					n80device.cmt._index = 0;
					result = TRUE;
					break;
				// invalid tag
				} else
					break;
			}
			break;
		}
		fs.close();
	}
	return result;
}
/**--------------------------------------------------------------------------
 **		CMT ۑt@CI
 **------------------------------------------------------------------------*/
void
n80_form::cmtsave()
{
	if (th->sync()) {
		dx->stop();
		ShowCursorFullScreen();
		::InvalidateRect(mWnd, NULL, TRUE);

		if (save_cmt_file(n80device.save_cmt_path)) {
		}
		HideCursorFullScreen();
		::SetFocus(mWnd);
		n80device.attr.invalid();
		th->desync();
		dx->play();
	}
}
/**--------------------------------------------------------------------------
 **		ۑt@C_CAO
 **------------------------------------------------------------------------*/
BOOL
n80_form::save_cmt_file(LPSTR szPath)
{
	OPENFILENAME	ofn;
	char	szName[16];
	char	szCurDir[512];

	::ZeroMemory (&ofn, sizeof(OPENFILENAME));
	::GetCurrentDirectory (sizeof (szCurDir), szCurDir);
	szName[0] = 0x00;
	lstrcpy (szPath, TEXT("cassette"));

	ofn.lStructSize			= sizeof (OPENFILENAME); 
	ofn.hwndOwner			= mWnd;
	ofn.hInstance			= NULL;
	ofn.lpstrFilter			= TEXT("CMT o̧\0*.CMT\0\0");
	ofn.lpstrCustomFilter	= NULL; 
	ofn.nMaxCustFilter		= 0;
	ofn.nFilterIndex		= 0; 
	ofn.lpstrFile			= szPath;
	ofn.nMaxFile			= 512;
	ofn.lpstrFileTitle		= szName; 
	ofn.nMaxFileTitle		= sizeof (szName); 
	ofn.lpstrInitialDir		= szCurDir;
	ofn.lpstrTitle			= TEXT("CMT ̧ق̕ۑI"); 
	ofn.Flags				= OFN_OVERWRITEPROMPT | OFN_EXPLORER;
	ofn.nFileOffset			= 0; 
	ofn.nFileExtension		= 0; 
	ofn.lpstrDefExt			= TEXT("CMT"); 
	ofn.lCustData			= 0; 
	ofn.lpfnHook			= NULL; 
	ofn.lpTemplateName		= NULL;

	return  ::GetSaveFileName (&ofn);
}
