/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Win32/DirectX
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		Load/Save Profile
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	1996/06/27	porting DOS version on WinG
 **		1.20	1997/08/10	porting WinG version on Win32/DirectX
 **		1.30	1998/03/12	
 **		1.43	2007/09/01	XPΉ
 **		1.44	2009/12/15	Vista/7Ή
 **		1.51	2010/01/27	DirectSound OnlyɕύX
 **------------------------------------------------------------------------*/
#include "n80_form.h"
#include "n80_directx.h"
#include "../z80/z80_thread.h"
#include "../lib/icj_stream.h"
#include <stdio.h>
#include <string.h>
#include <crtdbg.h>
/**--------------------------------------------------------------------------
 **		O[oϐ
 **------------------------------------------------------------------------*/
extern n80_directx	*dx;
extern z80_thread	*th;
/**--------------------------------------------------------------------------
 **		W[ϐ
 **------------------------------------------------------------------------*/
static LPCTSTR	l_section = {TEXT("Default")};
static LPCTSTR	l_key1    = {TEXT("z80_no_wait")};
static LPCTSTR	l_key2    = {TEXT("gvram_wait")};
static LPCTSTR	l_key3    = {TEXT("hard_volume")};
static LPCTSTR	l_key4    = {TEXT("soft_volume")};
static LPCTSTR	l_key5    = {TEXT("4th_rom_file")};
static LPCTSTR	l_key6    = {TEXT("disk_rom_file")};
static LPCTSTR	l_key8    = {TEXT("n88_rom_file")};
static LPCTSTR	l_key9    = {TEXT("sync_vblank")};
static LPCTSTR	l_key10   = {TEXT("tenkeyless")};
static LPCTSTR	l_key11   = {TEXT("screensize")};
static LPCTSTR	l_key12   = {TEXT("n80_rom_file")};
static LPCTSTR	l_key13   = {TEXT("interlace")};
static LPCTSTR	l_key14   = {TEXT("boot_rom")};
static LPCTSTR	l_key15   = {TEXT("hires_crt")};
static LPCTSTR	l_key16 = { TEXT("kanji1_rom_file") };
static LPCTSTR	l_key17 = { TEXT("kanji2_rom_file") };

#define N80_DEF_BOOT	1		// 1:N88-BASIC ROM
#define N80_DEF_LOAD	100		// DMA load
#define N80_DEF_GAIN	10		// GVRAM wait (busreq interval)
#define N80_DEF_HWVOL	(-4)	// hardware volume
#define N80_DEF_SWVOL	(1)		// software volume
#define N80_DEF_TKLESS	1		// ten key less
#define N80_DEF_NOWAIT	0		// z80 no-wait false
#define N80_DEF_SYNCVB	0		// wait vertical blank
#define N80_DEF_SCRNSIZ	0		// screen size (0:640x400, 1:720x480)
#define N80_DEF_VIDEO	1		// use video ram
#define N80_DEF_DEMAND	0		// draw text on demand
#define N80_DEF_INTLCE	0		// interlace like
#define N80_DEF_CRT		1		// use hi-res CRT (24kHz)
/**--------------------------------------------------------------------------
 **		INIt@C
 **------------------------------------------------------------------------*/
BOOL
n80_form::initProfile()
{
	icj_fstream	fs;

	::GetCurrentDirectory(sizeof (mIniPath), mIniPath);
	lstrcat(mIniPath, TEXT("\\N88.INI"));

	if (!fs.open(mIniPath)) {
		if (!fs.create(mIniPath))
			return  FALSE;
		fs.close();
		loadProfile(NULL);
		saveProfile(NULL);
	} else {
		fs.close();
		loadProfile(NULL);
	}
	return  TRUE;
}
/**--------------------------------------------------------------------------
 **		INIt@Cǂݍ
 **------------------------------------------------------------------------*/
void
n80_form::loadProfile(LPCTSTR sect)
{
	INT  val;

	if (sect == NULL)
		sect = l_section;

	val = ::GetPrivateProfileInt(sect, l_key1, N80_DEF_NOWAIT, mIniPath);
	n80win32.m_z80_no_wait = val == 1 ? TRUE : FALSE;

	val = ::GetPrivateProfileInt(sect, l_key2, N80_DEF_GAIN, mIniPath);
	n80device.param.gv_wait_ticks = val;

	val = ::GetPrivateProfileInt(sect, l_key3, N80_DEF_HWVOL, mIniPath);
	n80win32.hard_volume = val;

	val = ::GetPrivateProfileInt(sect, l_key4, N80_DEF_SWVOL, mIniPath);
	n80win32.soft_volume = val;

	val = ::GetPrivateProfileInt(sect, l_key9, N80_DEF_SYNCVB, mIniPath);
	dx->setSyncVBlank(val == 1 ? TRUE : FALSE);

	val = ::GetPrivateProfileInt(sect, l_key10, N80_DEF_TKLESS, mIniPath);
	n80win32.tenkeyless = val == 1 ? TRUE : FALSE;

	val = ::GetPrivateProfileInt(sect, l_key11, N80_DEF_SCRNSIZ, mIniPath);
	n80win32.screen_size = val == 0 ? 0 : 1;

	val = ::GetPrivateProfileInt(sect, l_key13, N80_DEF_INTLCE, mIniPath);
	n80device.param.interlace = val == 1 ? TRUE : FALSE;

	val = ::GetPrivateProfileInt(sect, l_key14, N80_DEF_BOOT, mIniPath);
	n80device.param.dipsw1.SW1_N_ROM = !val ? 0 : N80_DEF_BOOT;

	val = ::GetPrivateProfileInt(sect, l_key15, N80_DEF_CRT, mIniPath);
	n80device.vrtc.crt_hires_mode = val == 1 ? _CRT_HIRES : _CRT_NORMAL;
	n80device.vrtc.reset();

	val = ::GetPrivateProfileString(sect, l_key12, TEXT("8801-N80.ROM"), n80win32.n80_rom_file_name, 64, mIniPath);
	val = ::GetPrivateProfileString(sect, l_key8, TEXT("8801-N88.ROM"), n80win32.n88_rom_file_name, 64, mIniPath);
	val = ::GetPrivateProfileString(sect, l_key5, TEXT("8801-4th.ROM"), n80win32._4th_rom_file_name, 64, mIniPath);
	val = ::GetPrivateProfileString(sect, l_key6, TEXT("pdidk.rom"), n80win32.fdu_rom_file_name, 64, mIniPath);
	val = ::GetPrivateProfileString(sect, l_key16, TEXT("KANJI1.ROM"), n80win32.kanji1_rom_file_name, 64, mIniPath);
	val = ::GetPrivateProfileString(sect, l_key17, TEXT("KANJI2.ROM"), n80win32.kanji2_rom_file_name, 64, mIniPath);
}
/**--------------------------------------------------------------------------
 **		INIt@Co
 **------------------------------------------------------------------------*/
void
n80_form::saveProfile(LPCTSTR sect)
{
	char  num[64];

	if (sect == NULL)
		sect = l_section;

	sprintf(num, "%d", n80win32.m_z80_no_wait ? 1 : 0);
	::WritePrivateProfileString(sect, l_key1, num, mIniPath);

	sprintf(num, "%u", n80device.param.gv_wait_ticks);
	::WritePrivateProfileString(sect, l_key2, num, mIniPath);

	sprintf(num, "%d", n80win32.hard_volume);
	::WritePrivateProfileString(sect, l_key3, num, mIniPath);

	sprintf(num, "%d", n80win32.soft_volume);
	::WritePrivateProfileString(sect, l_key4, num, mIniPath);

	sprintf(num, "%u", dx->getSyncVBlank() ? 1 : 0);
	::WritePrivateProfileString(sect, l_key9, num, mIniPath);

	sprintf(num, "%u", n80win32.tenkeyless ? 1 : 0);
	::WritePrivateProfileString(sect, l_key10, num, mIniPath);

	sprintf(num, "%d", n80win32.screen_size == 0 ? 0 : 1);
	::WritePrivateProfileString(sect, l_key11, num, mIniPath);

	sprintf(num, "%u", n80device.param.interlace ? 1 : 0);
	::WritePrivateProfileString(sect, l_key13, num, mIniPath);

	sprintf(num, "%u", n80device.param.dipsw1.SW1_N_ROM);
	::WritePrivateProfileString(sect, l_key14, num, mIniPath);

	sprintf(num, "%u", n80device.vrtc.crt_hires_mode == _CRT_HIRES ? 1 : 0);
	::WritePrivateProfileString(sect, l_key15, num, mIniPath);

	::WritePrivateProfileString(sect, l_key12, n80win32.n80_rom_file_name, mIniPath);
	::WritePrivateProfileString(sect, l_key8, n80win32.n88_rom_file_name, mIniPath);
	::WritePrivateProfileString(sect, l_key5, n80win32._4th_rom_file_name, mIniPath);
	::WritePrivateProfileString(sect, l_key6, n80win32.fdu_rom_file_name, mIniPath);
	::WritePrivateProfileString(sect, l_key16, n80win32.kanji1_rom_file_name, mIniPath);
	::WritePrivateProfileString(sect, l_key17, n80win32.kanji2_rom_file_name, mIniPath);
}
