/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Win32/DirectX
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		System Control
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	1996/06/27	porting DOS version on WinG
 **		1.20	1997/08/10	porting WinG version on Win32/DirectX
 **		1.30	1998/03/12	
 **		1.43	2007/09/01	XPΉ
 **		1.44	2009/12/15	Vista/7Ή
 **		1.51	2010/01/27	DirectSound OnlyɕύX
 **------------------------------------------------------------------------*/
#include "n80_form.h"
#include "n80_directx.h"
#include "../z80/z80_thread.h"
#include <stdio.h>
/**--------------------------------------------------------------------------
 **		O[oϐ
 **------------------------------------------------------------------------*/
extern n80_assay	*assay;
extern z80_thread	*th;
extern n80_directx	*dx;
/**--------------------------------------------------------------------------
 **		AvP[Vnh
 **------------------------------------------------------------------------*/
void
n80_form::OnAppActivate (BOOL factive)
{
	if (th->isRunning()) {
		if (factive) {
			if (mLostDDS) {
				restart();
				//assay->outputLog(TEXT("OnAppActivate: Activate without Surface"));
			} else
				;//assay->outputLog(TEXT("OnAppActivate: Activate"));
		} else {
			//assay->outputLog(TEXT("OnAppActivate: Deactivate"));
			// ٽذݒDialogȊO̕sӂDe-Activateꍇ
			if (dx->getFullScreenMode()) {
				dx->stop();
				th->sync();
				mLostDDS = TRUE;
				ShowCursorFullScreen();
			}
		}
	}
}
/**--------------------------------------------------------------------------
 **		X^[g
 **------------------------------------------------------------------------*/
void
n80_form::start()
{
	if (th && dx) {
		mLostDDS = FALSE;		// DirectDrawSurface敪iLj
		mReqRestart = FALSE;	// Z80ċNv敪ivj

		dx->restore_surface();
		n80device.attr.invalid();
		th->start(TRUE);		// start Z80 thread
		th->resume();

		HideCursorFullScreen();
		::SetFocus(mWnd);
		return;
	}
	dx->dispose_dx();
	::DestroyWindow(mWnd);
}
/**--------------------------------------------------------------------------
 **		ăX^[g
 **------------------------------------------------------------------------*/
void
n80_form::restart ()
{
	HideCursorFullScreen();
	::SetFocus(mWnd);
	n80device.attr.invalid();
	initKeyb();						// keyboard reset
	m_key_combo = KEYCOMB_CLEAR;	// drive combination key clear
	mLostDDS = FALSE;
	th->desync();
	dx->stop();
}
/**--------------------------------------------------------------------------
 **		Zbg
 **------------------------------------------------------------------------*/
void
n80_form::reset()
{
	if (th && !mLostDDS && th->sync())
	{
		dx->stop();
		dx->clear_Ts_bank();
		dx->setMute(FALSE);				// awake sound
		initKeyb();						// keyboard reset
		m_key_combo = KEYCOMB_CLEAR;	// drive combination key clear
		n80device.reset(&n80device);
		n80device.pcg_mode = 0;			// pcg off
		n80device.cmt._index = 0;		// cmt rewind
		th->reset();					// reset Z80
		th->desync();					// restart Z80
		dx->play();
	}
}
/**--------------------------------------------------------------------------
 **		I
 **------------------------------------------------------------------------*/
void
n80_form::exit()
{
	assay->outputLog(TEXT("exit: try stop z80 thread"));
	if (th && th->sync(WM_QUIT)) {		// suspend and quit Z80/Render thread
		assay->outputLog(TEXT("exit: terminating N80"));
		dx->stop();
		th->exit();						// request terminate Z80
		th->desync();					// resume Z80
		th->kill();						// terminate Z80
		dx->dispose_dx();
		n80device.dispose(&n80device);
	}
}

void
n80_form::change_screen() {
	assay->outputLog(TEXT("change_screen: try stop z80 thread"));
	if (th && th->sync()) {	// suspend Z80/Render thread
		assay->outputLog(TEXT("change_screen: try stop direct sound"));
		dx->stop();
		assay->outputLog(TEXT("change_screen: try reinit"));
		::ShowCursor(dx->getFullScreenMode());
		if (!dx->reinit_dd()) {
			assay->outputLog(TEXT("change_screen: exit by failed"));
			th->exit();
			th->desync();
			th->kill();
			dx->dispose_dx();
			return;
		}
		assay->outputLog(TEXT("change_screen: restart"));
		::ShowCursor(!dx->getFullScreenMode());
		::SetFocus(mWnd);
		n80device.attr.invalid();
		m_key_combo = KEYCOMB_CLEAR;	// drive combination key clear
		th->desync();
		dx->play();
	}
}
/**--------------------------------------------------------------------------
 **		g`ksI
 **------------------------------------------------------------------------*/
void
n80_form::halt(LPZ80CPU cpu)
{
	TCHAR	mes[128];
	TCHAR*	halt;

	dx->stop();
	switch (cpu->r.x.HALT) {
	case _Z80HALT_76:	halt = "HALT";		break;
	case _Z80HALT_IL:	halt = "`";		break;
	case _Z80HALT_UI:	halt = "";		break;
	case _Z80HALT_BUS:	halt = "IM0oXO";	break;
	case _Z80HALT_INT:	halt = "INT[hO";	break;
	default:
		halt = "\HALT";	break;
	}
	sprintf (mes, "$%4.4X ɂ%soɂ~B", cpu->r.x.head, halt);	
	assay->outputLog(mes);
	::ShowCursor(TRUE);
	::InvalidateRect (mWnd, NULL, FALSE);
	::MessageBox (mWnd, mes, "N88DX", MB_OK | MB_ICONSTOP);
	dx->play();
}
