/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#include "pc-8001.h"
#ifdef _MSC_VER
#include <windows.h>
#include "../N80/N80.h"
#endif
#ifdef __cplusplus
extern "C"
{
#endif
#define export
/**--------------------------------------------------------------------------
 **		{[_[NA
 **------------------------------------------------------------------------*/
export void bltborder() {
	int		bd_grb = n80device.io._o52 & 0x07;	// Boarder Color (GRB)
	BYTE	dot = n80win32.prime_palettes[n80device.attr.grb_rgb[n80device.palette[bd_grb]]];
	BYTE	cls = n80win32.second_palettes[n80device.attr.grb_rgb[n80device.palette[bd_grb]]];
	LPBYTE	vaddr = n80win32.vram_addr;

	// border clear
	if (n80win32.vram_top_offset > 0) {
		for (int v = 0; v < (int)n80win32.vram_top_offset; ++v) {
			BYTE color = v & 0x01 ? cls : dot;
			LPBYTE video = &vaddr[v * n80win32.vram_pixcel_pitch];
			for (int h = 0; h < (int)n80win32.width; ++h)
				*video++ = color;
		}
		for (int v = n80win32.vram_bottom_offset; v < (int)n80win32.height; ++v) {
			BYTE color = v & 0x01 ? cls : dot;
			LPBYTE video = &vaddr[v * n80win32.vram_pixcel_pitch];
			for (int h = 0; h < (int)n80win32.width; ++h)
				*video++ = color;
		}
	}
	if (n80win32.vram_left_offset > 0) {
		for (int v = n80win32.vram_top_offset; v < (int)n80win32.vram_bottom_offset; ++v) {
			BYTE color = v & 0x01 ? cls : dot;
			LPBYTE video1 = &vaddr[v * n80win32.vram_pixcel_pitch];
			LPBYTE video2 = &vaddr[v * n80win32.vram_pixcel_pitch + n80win32.vram_right_offset];
			for (int h = 0; h < (int)n80win32.vram_left_offset; ++h) {
				*video1++ = color;
				*video2++ = color;
			}
		}
	}
}
/**--------------------------------------------------------------------------
 **		GVRAM`
 **------------------------------------------------------------------------*/
static inline int gvram2color(CBYTE vram0, CBYTE vram1, CBYTE vram2, CBYTE bit) {
	BYTE grb = 0;
	if (vram0 & bit)
		grb |= 1;
	if (vram1 & bit)
		grb |= 2;
	if (vram2 & bit)
		grb |= 4;
	grb = n80device.palette[grb];
	return n80device.attr.grb_rgb[grb];
}
export void drawgvram() {
	LPBYTE	gvram0 = &n80device.mem._gvram0[0];		// PC-8801 GVRAM #0
	LPBYTE	gvram1 = &n80device.mem._gvram1[0];		// PC-8801 GVRAM #1
	LPBYTE	gvram2 = &n80device.mem._gvram2[0];		// PC-8801 GVRAM #2
	int		color;

	bltborder();
	for (int y = 0; y < 200; ++y) {
		int idx1 = ((y * 2) + n80win32.vram_top_offset) * n80win32.vram_pixcel_pitch + n80win32.vram_left_offset;
		int idx2 = idx1 + n80win32.vram_pixcel_pitch;
		LPBYTE video1 = &n80win32.vram_addr[idx1];	// DirectX Surface GVRAM even raster
		LPBYTE video2 = &n80win32.vram_addr[idx2];	// DirectX Surface GVRAM odd  raster
		for (int x = 0; x < 80; ++x) {
			BYTE vram0 = *gvram0++;
			BYTE vram1 = *gvram1++;
			BYTE vram2 = *gvram2++;
			color = gvram2color(vram0, vram1, vram2, 0x80);
			*video1++ = n80win32.prime_palettes[color];
			*video2++ = n80win32.second_palettes[color];
			color = gvram2color(vram0, vram1, vram2, 0x40);
			*video1++ = n80win32.prime_palettes[color];
			*video2++ = n80win32.second_palettes[color];
			color = gvram2color(vram0, vram1, vram2, 0x20);
			*video1++ = n80win32.prime_palettes[color];
			*video2++ = n80win32.second_palettes[color];
			color = gvram2color(vram0, vram1, vram2, 0x10);
			*video1++ = n80win32.prime_palettes[color];
			*video2++ = n80win32.second_palettes[color];
			color = gvram2color(vram0, vram1, vram2, 0x08);
			*video1++ = n80win32.prime_palettes[color];
			*video2++ = n80win32.second_palettes[color];
			color = gvram2color(vram0, vram1, vram2, 0x04);
			*video1++ = n80win32.prime_palettes[color];
			*video2++ = n80win32.second_palettes[color];
			color = gvram2color(vram0, vram1, vram2, 0x02);
			*video1++ = n80win32.prime_palettes[color];
			*video2++ = n80win32.second_palettes[color];
			color = gvram2color(vram0, vram1, vram2, 0x01);
			*video1++ = n80win32.prime_palettes[color];
			*video2++ = n80win32.second_palettes[color];
		}
	}
}
static __inline LPBYTE gvram_raster(LPBYTE gvram, LPBYTE video, BYTE dot, BYTE cls) {
	for (int x = 0; x < 80; ++x) {
		BYTE vram = *gvram++;
		*video++ = vram & 0x80 ? dot : cls;
		*video++ = vram & 0x40 ? dot : cls;
		*video++ = vram & 0x20 ? dot : cls;
		*video++ = vram & 0x10 ? dot : cls;
		*video++ = vram & 0x08 ? dot : cls;
		*video++ = vram & 0x04 ? dot : cls;
		*video++ = vram & 0x02 ? dot : cls;
		*video++ = vram & 0x01 ? dot : cls;
	}
	return gvram;
}
export void drawgvram_hires() {
	LPBYTE	video = n80win32.vram_addr;					// DirectX Surface GVRAM
	LPBYTE	gvram = &n80device.mem._gvram0[0];			// PC-8801 GVRAM
	int		bg_grb = (n80device.io._o52 >> 4) & 0x07;	// BG Color
	BYTE	dot = n80win32.prime_palettes[n80device.attr.grb_rgb[n80device.palette[7]]];
	BYTE	cls = n80win32.prime_palettes[n80device.attr.grb_rgb[n80device.palette[bg_grb]]];
	BYTE	dot2 = n80win32.second_palettes[n80device.attr.grb_rgb[n80device.palette[7]]];
	BYTE	cls2 = n80win32.second_palettes[n80device.attr.grb_rgb[n80device.palette[bg_grb]]];

	bltborder();
	for (int y = 0; y < 200; ++y) {
		int idx = (y + n80win32.vram_top_offset) * n80win32.vram_pixcel_pitch + n80win32.vram_left_offset;
		if (y & 0x01)
			gvram = gvram_raster(gvram, &video[idx], dot, cls);
		else
			gvram = gvram_raster(gvram, &video[idx], dot2, cls2);
	}
	gvram = &n80device.mem._gvram1[0];
	for (int y = 200; y < 400; ++y) {
		int idx = (y + n80win32.vram_top_offset) * n80win32.vram_pixcel_pitch + n80win32.vram_left_offset;
		if (y & 0x01)
			gvram = gvram_raster(gvram, &video[idx], dot, cls);
		else
			gvram = gvram_raster(gvram, &video[idx], dot2, cls2);
	}
}
/**--------------------------------------------------------------------------
 **		DirectDraw Surface VRAM 
 **------------------------------------------------------------------------*/
static __inline void set_pixel_80(LPBYTE vram) {
	*vram                              = n80win32.cur_pix_rgb_pal;
	vram[n80win32.vram_pixcel_pitch]   = n80win32.cur_pix_rgb_pal2;
}
static __inline void set_pixel_40(LPBYTE vram) {
	*vram                              = n80win32.cur_pix_rgb_pal;
	vram[n80win32.vram_pixcel_pitch]   = n80win32.cur_pix_rgb_pal2;
	vram[1]                            = n80win32.cur_pix_rgb_pal;
	vram[n80win32.vram_pixcel_pitch+1] = n80win32.cur_pix_rgb_pal2;
}
static __inline void reset_pixel_80(LPBYTE vram) {
//	*vram                              = 0;
//	vram[n80win32.vram_pixcel_pitch]   = 0;
}
static __inline void reset_pixel_40(LPBYTE vram) {
//	*vram                              = 0;
//	vram[n80win32.vram_pixcel_pitch]   = 0;
//	vram[1]                            = 0;
//	vram[n80win32.vram_pixcel_pitch+1] = 0;
}
/**--------------------------------------------------------------------------
 **		Font Pattern H
 **------------------------------------------------------------------------*/
static __inline BYTE proc_video_pixcel25(BYTE attr, BYTE bit, int r) {
	if (attr & _PAC_ATTR_UND_LINE && r == 7)	// under line
		bit = 0xff;
	if (attr & _PAC_ATTR_UPR_LINE && r == 0)	// upper line
		bit = 0xff;
	if (attr & _PAC_ATTR_ERASE)					// secret
		bit = 0;
	if (attr & _PAC_ATTR_REVERSE)				// reverse
		bit = ~bit;
	return bit;
}
static __inline BYTE proc_video_pixcel20(BYTE attr, BYTE bit, int r) {
	if (attr & _PAC_ATTR_UND_LINE && r == 9)	// under line
		bit = 0xff;
	if (attr & _PAC_ATTR_UPR_LINE && r == 0)	// upper line
		bit = 0xff;
	if (attr & _PAC_ATTR_ERASE)					// secret
		bit = 0;
	if (attr & _PAC_ATTR_REVERSE)				// reverse
		bit = ~bit;
	return bit;
}
/**--------------------------------------------------------------------------
 **		Text(80x25)`
 **------------------------------------------------------------------------*/
static __inline LPBYTE draw_font_raster8025(LPBYTE vram, int r, BYTE attr, BYTE fnt) {
	register LPBYTE _vram = (vram + (n80win32.vram_pixcel_pitch * 2));
	register BYTE pat = proc_video_pixcel25(attr, fnt, r);

	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram) : reset_pixel_80(vram);

	return _vram;
}
export void blt8025(int line) {
	if (line < 25) {
		LPBYTE		txts   = &n80device.attr._textcach[0];
		LPBYTE		pacs   = &n80device.attr._eval_attr[0];
		int			offset = n80win32.vtbl_vram25[line], col;

		for (col = 0; col < 80; ++col) {
			BYTE	ascii = *txts++;
			BYTE	attr  = *pacs++;
			LPBYTE	fnt   = &n80device.font[ascii*8];
			LPBYTE	vram  = &n80win32.vram_addr[offset + (col * 8)];
			int		color = attr & _PAC_ATTR_RGB;

			n80win32.cur_pix_rgb_pal  = n80win32.prime_palettes[color];
			n80win32.cur_pix_rgb_pal2 = n80win32.second_palettes[color];
			
			if (attr & _PAC_ATTR_GRAPH)		// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;

			vram = draw_font_raster8025(vram, 0, attr, *fnt++);
			vram = draw_font_raster8025(vram, 1, attr, *fnt++);
			vram = draw_font_raster8025(vram, 2, attr, *fnt++);
			vram = draw_font_raster8025(vram, 3, attr, *fnt++);
			vram = draw_font_raster8025(vram, 4, attr, *fnt++);
			vram = draw_font_raster8025(vram, 5, attr, *fnt++);
			vram = draw_font_raster8025(vram, 6, attr, *fnt++);
			vram = draw_font_raster8025(vram, 7, attr, *fnt++);
		}
	}
}
/**--------------------------------------------------------------------------
 **		Text(80x20)`
 **------------------------------------------------------------------------*/
static __inline LPBYTE draw_font_raster8020(LPBYTE vram, int r, BYTE attr, BYTE fnt) {
	register LPBYTE _vram = (vram + (n80win32.vram_pixcel_pitch * 2));
	register BYTE pat = proc_video_pixcel20(attr, fnt, r);

	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
	pat <<= 1;
	pat & 0x80 ? set_pixel_80(vram) : reset_pixel_80(vram);

	return _vram;
}
export void blt8020(int line) {
	if (line < 20) {
		LPBYTE		txts   = &n80device.attr._textcach[0];
		LPBYTE		pacs   = &n80device.attr._eval_attr[0];
		int			offset = n80win32.vtbl_vram20[line], col;
		
		for (col = 0; col < 80; ++col) {
			BYTE	ascii = *txts++;
			BYTE	attr  = *pacs++;
			LPBYTE	fnt   = &n80device.font[ascii*8];
			LPBYTE	vram  = &n80win32.vram_addr[offset + (col * 8)];
			int		color = attr & _PAC_ATTR_RGB;

			n80win32.cur_pix_rgb_pal  = n80win32.prime_palettes[color];
			n80win32.cur_pix_rgb_pal2 = n80win32.second_palettes[color];
			
			if (attr & _PAC_ATTR_GRAPH)		// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;
			
			vram = draw_font_raster8020(vram, 0, attr, 0);
			vram = draw_font_raster8020(vram, 1, attr, *fnt++);
			vram = draw_font_raster8020(vram, 2, attr, *fnt++);
			vram = draw_font_raster8020(vram, 3, attr, *fnt++);
			vram = draw_font_raster8020(vram, 4, attr, *fnt++);
			vram = draw_font_raster8020(vram, 5, attr, *fnt++);
			vram = draw_font_raster8020(vram, 6, attr, *fnt++);
			vram = draw_font_raster8020(vram, 7, attr, *fnt++);
			vram = draw_font_raster8020(vram, 8, attr, *fnt++);
			vram = draw_font_raster8020(vram, 9, attr, 0);
		}
	}
}
/**--------------------------------------------------------------------------
 **		Text(40x25)`
 **------------------------------------------------------------------------*/
static __inline LPBYTE draw_font_raster4025(LPBYTE vram, int r, BYTE attr, BYTE fnt) {
	register LPBYTE _vram = (vram + (n80win32.vram_pixcel_pitch * 2));
	register BYTE pat = proc_video_pixcel25(attr, fnt, r);

	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);

	return _vram;
}
export void blt4025(int line) {
	if (line < 25) {
		LPBYTE		txts   = &n80device.attr._textcach[0];
		LPBYTE		pacs   = &n80device.attr._eval_attr[0];
		int			offset = n80win32.vtbl_vram25[line], col;
		
		for (col = 0; col < 80; col += 2) {
			BYTE	ascii = *txts;
			BYTE	attr  = *pacs;
			LPBYTE	fnt   = &n80device.font[ascii*8];
			LPBYTE	vram  = &n80win32.vram_addr[offset + (col * 8)];
			int		color = attr & _PAC_ATTR_RGB;

			n80win32.cur_pix_rgb_pal  = n80win32.prime_palettes[color];
			n80win32.cur_pix_rgb_pal2 = n80win32.second_palettes[color];
			
			txts += 2;
			pacs += 2;
			
			if (attr & _PAC_ATTR_GRAPH)		// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;
			
			vram = draw_font_raster4025(vram, 0, attr, *fnt++);
			vram = draw_font_raster4025(vram, 1, attr, *fnt++);
			vram = draw_font_raster4025(vram, 2, attr, *fnt++);
			vram = draw_font_raster4025(vram, 3, attr, *fnt++);
			vram = draw_font_raster4025(vram, 4, attr, *fnt++);
			vram = draw_font_raster4025(vram, 5, attr, *fnt++);
			vram = draw_font_raster4025(vram, 6, attr, *fnt++);
			vram = draw_font_raster4025(vram, 7, attr, *fnt++);
		}
	}
}
/**--------------------------------------------------------------------------
 **		Text(40x20)`
 **------------------------------------------------------------------------*/
static __inline LPBYTE draw_font_raster4020(LPBYTE vram, int r, BYTE attr, BYTE fnt) {
	register LPBYTE _vram = (vram + (n80win32.vram_pixcel_pitch * 2));
	register BYTE pat = proc_video_pixcel20(attr, fnt, r);

	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);	vram += 2;
	pat <<= 1;
	pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);

	return _vram;
}
export void blt4020(int line) {
	if (line < 20) {
		LPBYTE		txts   = &n80device.attr._textcach[0];
		LPBYTE		pacs   = &n80device.attr._eval_attr[0];
		int			offset =n80win32.vtbl_vram20[line], col;
		
		for (col = 0; col < 80; col += 2) {
			BYTE	ascii = *txts;
			BYTE	attr  = *pacs;
			LPBYTE	fnt   = &n80device.font[ascii*8];
			LPBYTE	vram  = &n80win32.vram_addr[offset + (col * 8)];
			int		color = attr & _PAC_ATTR_RGB;

			n80win32.cur_pix_rgb_pal  = n80win32.prime_palettes[color];
			n80win32.cur_pix_rgb_pal2 = n80win32.second_palettes[color];
			
			txts += 2;
			pacs += 2;
			
			if (attr & _PAC_ATTR_GRAPH)		// graph
				fnt += 2048;
			else if (n80device.pcg_mode)	// pcg
				fnt += 4096;
			
			vram = draw_font_raster4020(vram, 0, attr, 0);
			vram = draw_font_raster4020(vram, 1, attr, *fnt++);
			vram = draw_font_raster4020(vram, 2, attr, *fnt++);
			vram = draw_font_raster4020(vram, 3, attr, *fnt++);
			vram = draw_font_raster4020(vram, 4, attr, *fnt++);
			vram = draw_font_raster4020(vram, 5, attr, *fnt++);
			vram = draw_font_raster4020(vram, 6, attr, *fnt++);
			vram = draw_font_raster4020(vram, 7, attr, *fnt++);
			vram = draw_font_raster4020(vram, 8, attr, *fnt++);
			vram = draw_font_raster4020(vram, 9, attr, 0);
		}
	}
}
/**--------------------------------------------------------------------------
 **		NA
 **------------------------------------------------------------------------*/
export void bltblank() {
	bltborder();
	for (int h = 0; h < 20; ++h) {
		int offset = n80win32.vtbl_vram20[h];
		LPBYTE	vram = &n80win32.vram_addr[offset];
		for (int i = 0; i < 640; ++i) {
			vram[i+n80win32.vram_vpitch[0]]  = 0;
			vram[i+n80win32.vram_vpitch[1]]  = 0;
			vram[i+n80win32.vram_vpitch[2]]  = 0;
			vram[i+n80win32.vram_vpitch[3]]  = 0;
			vram[i+n80win32.vram_vpitch[4]]  = 0;
			vram[i+n80win32.vram_vpitch[5]]  = 0;
			vram[i+n80win32.vram_vpitch[6]]  = 0;
			vram[i+n80win32.vram_vpitch[7]]  = 0;
			vram[i+n80win32.vram_vpitch[8]]  = 0;
			vram[i+n80win32.vram_vpitch[9]]  = 0;
			vram[i+n80win32.vram_vpitch[10]] = 0;
			vram[i+n80win32.vram_vpitch[11]] = 0;
			vram[i+n80win32.vram_vpitch[12]] = 0;
			vram[i+n80win32.vram_vpitch[13]] = 0;
			vram[i+n80win32.vram_vpitch[14]] = 0;
			vram[i+n80win32.vram_vpitch[15]] = 0;
			vram[i+n80win32.vram_vpitch[16]] = 0;
			vram[i+n80win32.vram_vpitch[17]] = 0;
			vram[i+n80win32.vram_vpitch[18]] = 0;
			vram[i+n80win32.vram_vpitch[19]] = 0;
		}
	}
}

#ifdef __cplusplus
}	// extern "C"
#endif
