/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#include "pc-8001.h"
#ifdef _MSC_VER
#include <windows.h>
#include "../N80/N80.h"
#endif
#ifdef __cplusplus
extern "C"
{
#endif
#define export
extern void blt8025(int line);
extern void blt8020(int line);
extern void blt4025(int line);
extern void blt4020(int line);
extern void blt8025x(int line);
extern void blt8020x(int line);
extern void blt4025x(int line);
extern void blt4020x(int line);
extern void bltblank(void);
extern void bltblankx(void);
extern void bltborder(void);
extern void bltborderx(void);
extern void drawgvram(void);
extern void drawgvram_hires(void);
extern void drawgvramx(void);
extern void drawgvramx_hires(void);

typedef void (*PBLTFNC)(int);
typedef void (*PBLTBLK)(void);
typedef void (*PDRAWGV)(void);
// GRB to RGB order
static const BYTE	font_color[8]  = {0, 252, 250, 254, 249, 253, 251, 255};
static const BYTE	font_colord[8] = {0,   4,   2,   6,   1,   5,   3, 248};
static const DWORD	font_colorx[8] = {
	0x00000000, 0x000000FF, 0x0000FF00, 0x0000FFFF, 0x00FF0000, 0x00FF00FF, 0x00FFFF00, 0x00FFFFFF
};
static const DWORD	font_colorxd[8] = {
	0x00000000, 0x0000007F, 0x00007F00, 0x00007F7F, 0x007F0000, 0x007F007F, 0x007F7F00, 0x007F7F7F
};

static PBLTFNC	blt_funcs[]	= {blt4020, blt4020x};
static PBLTFNC	fnc_tbl[]   = {blt4020, blt4025, blt8020, blt8025};
static PBLTFNC	fnc_tblx[]  = {blt4020x, blt4025x, blt8020x, blt8025x};
static PBLTBLK	blt_blks[]	= {bltblank, bltblankx};
static PBLTBLK	blt_bdrs[]	= {bltborder, bltborderx};
static PDRAWGV	drw_vram[][2] = {{drawgvram_hires, drawgvramx_hires}, {drawgvram, drawgvramx}};

static void chg_bltfunc(int screen_mode) {
	if (screen_mode >= 0 && screen_mode < 4) {
		blt_funcs[0] = fnc_tbl[screen_mode];
		blt_funcs[1] = fnc_tblx[screen_mode];
	}
}

static void _blt_font(int line, int bpp_mode) {
	blt_funcs[bpp_mode](line);
}

static void _blt_blank(int bpp_mode) {
	blt_blks[bpp_mode]();
}

static void _blt_border(int bpp_mode) {
	blt_bdrs[bpp_mode]();
}

static void _drawgvram(int bpp_mode) {
	int index = n80device.vrtc.crt_hires_mode == _CRT_HIRES && !(n80device.io._o31 & 0x01) ? 0 : 1;
	drw_vram[index][bpp_mode]();
}

static void __cdecl chg_palette(BOOL interrace) {
	n80win32.second_palettes   = interrace ? font_colord : font_color;
	n80win32.second_rgb_colors = interrace ? font_colorxd : font_colorx;
}

export void _n80_init_blt(void) {
	n80device.blt.chg_crt_mode	= chg_bltfunc;
	n80device.blt.font			= _blt_font;
	n80device.blt.blank			= _blt_blank;
	n80device.blt.border		= _blt_border;
	n80device.blt.chg_palette	= chg_palette;
	n80device.blt.drawgvram		= _drawgvram;
	n80win32.prime_palettes		= font_color;
	n80win32.prime_rgb_colors	= font_colorx;
}

#ifdef __cplusplus
}	// extern "C"
#endif
