﻿/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#include "pc-8001.h"
#ifdef _MSC_VER
#include <windows.h>
#else
#include "../include/n80types.h"
#endif
#ifndef __APPLE__
#include <stdio.h>
#endif
#ifdef __cplusplus
extern "C"
{
#endif
#define export
#define BAUDRATE	(33280)	//1200baud~=33280T(1byte=8+2bits)
#define bMTON		(0x08)
#define bBS2		(0x20)
/**--------------------------------------------------------------------------
 **		ASM Modules
 **------------------------------------------------------------------------*/
static void _n80_reset_cmt(void) {
	n80device.cmt._baud = BAUDRATE;
	n80device.cmt._reqcas = 0x0d;	// USART:232c, motor:off, tape:eject
	// _reqcas bit 3210
	//             ||x+-- 0:ready, 1:eject
	//             |+---- 0:moto on, 1:moto off
	//             +----- 0:cas, 1:rs-232c
}
	
static void _n80_cmt_control(void) {
	if ((n80device.io._o30 & bMTON) == 0) {	// check motor is on/off
		n80device.cmt._reqcas |= 0x04;		// motor off
		n80device.cmt._baud = BAUDRATE;
	} else {
		n80device.cmt._reqcas &= 0xfb;		// motor on
	}
	if ((n80device.io._o30 & bBS2) != 0)
		n80device.cmt._reqcas |= 0x08;		// usart ch bind to rs232c
	else
		n80device.cmt._reqcas &= 0xf7;		// usart ch bind to cassette tape
}

static void _n80_tape_notify(BYTE flag) {
	if (flag == 1)
		n80device.cmt._reqcas &= 0xfe;		// tape ready
	else
		n80device.cmt._reqcas |= 0x01;		// tape eject (not ready)
}
	
static void _n80_cmt_process(LPZ80CPU cpu) {	// called in a cycle from device time
	BYTE d;
	if ((n80device.io._oE6 & 0x04) == 0)	// USART interrupt mask is on
		return;
	if ((n80device.pic._holding & 1) == 1)	// prev CMT interrupt req is suspended
		return;
	if ((n80device.cmt._reqcas & 0x0f) != 0)// tape not ready
		return;
	n80device.cmt._baud -= cpu->r.x.T;		// 1200baud(~=120bytes/sec)
	if (n80device.cmt._baud > 0)
		return;
	if (n80device.usart._rxrdy == 1)		// buffer full
		return;
	n80device.cmt._baud = BAUDRATE;
	d = n80device.cmt.casdata();			// read data from cmt
	n80device.usart.rxd(cpu, d);			// send data to USART
}

static BYTE _n80_cas_data(void) {
	if (n80device.cmt._buff == NULL)
		return 0xff;
	if (n80device.cmt._index >= n80device.cmt._size)
		n80device.cmt._index = 0;
	return n80device.cmt._buff[n80device.cmt._index++];
}

static void _n80_fast_cas_load(void) {
	int ope = 0, quit = 0;
	USHORT addr = 0;
	BYTE block_size = 0;
	DWORD i, W;
	for (i = n80device.cmt._index; !quit && i < n80device.cmt._size; ++i) {
		BYTE d = n80device.cmt._buff[i];

		switch (ope) {
			case 0:
				if (d == 0x3a)
					ope = 1;
				break;
			case 1:	// address hi
				addr = (USHORT)d;
				addr <<= 8;
				ope = 2;
				break;
			case 2:	// address lo
				addr += d;
				ope = 3;
				break;
			case 3:	// check sum
				ope = 4;
				break;
			case 4:
				if (d != 0x3a)
					ope = 1;//quit = 1;
				else
					ope = 5;
				break;
			case 5:
				block_size = d;
				if (d == 0)
					quit = 1;
				ope = 6;
				break;
			case 6:
				n80device.mem.write(&n80device.cpu, addr++, d, &W);
				//n80device.mem._ram[addr++] = d;
				if (--block_size == 0)
					ope = 7;
				break;
			case 7:	// check sum
				ope = 4;
				break;
			default:
				quit = 1;
				break;
		}
	}
	n80device.cmt._index = i;
}
	
export void _n80_init_cmt(void) {
	n80device.cmt.reset = _n80_reset_cmt;
	n80device.cmt.control = _n80_cmt_control;
	n80device.cmt.notify = _n80_tape_notify;
	n80device.cmt.process = _n80_cmt_process;
	n80device.cmt.fastload = _n80_fast_cas_load;
	n80device.cmt.casdata = _n80_cas_data;
	_n80_reset_cmt();
}

#ifdef __cplusplus
}	// extern "C"
#endif
