/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#include "PC-8001.h"
#include "../Z80/Z80Mac.h"
#ifdef _MSC_VER
#include <windows.h>
#else
#include "../include/n80.h"
#endif
#ifdef __cplusplus
extern "C"
{
#endif
#define export
/**--------------------------------------------------------------------------
 **		ASM Modules
 **------------------------------------------------------------------------*/
static void _n88_sjis_to_jis(const int s1, const int s2, int *j1, int *j2) {
	if (s2 < 0x9F) {
		*j1 = (s1 << 1) - ((s1 >= 0xE0) ? 0x161 : 0xE1);
		*j2 = s2 - ((s2 > 0x7F) ? 0x20 : 0x1F);
	} else {
		*j1 = (s1 << 1) - ((s1 >= 0xE0) ? 0x160 : 0xE0);
		*j2 = s2 - 0x7E;
	}
}

static int _n88_jis_to_offset(BYTE jish, BYTE jisl) {
	int offset = 0;
	USHORT jis = MAKER16(jish, jisl);
	if (jis >= 0x0020 && jis <= 0x00ff) {	// Half-Width char
		offset = jis;
		offset <<= 3;
	} else if (jis >= 0x0100 && jis <= 0x01ff) {
		offset = jis;
		offset <<= 2;
	} else if (jis >= 0x2120 && jis <= 0x277f) {
		offset |= (jis & 0x0700) << 1;
		offset |= (jis & 0x0060) << 7;
		offset |= (jis & 0x001f) << 4;
	} else if (jis >= 0x3020 && jis <= 0x4f5f) {
		offset |= (jis & 0x1f00) << 1;
		offset |= (jis & 0x0060) << 9;
		offset |= (jis & 0x001f) << 4;
	} else if(jis >= 0x5020 && jis <= 0x6f7f) {
		offset |= (jis & 0x0f00) << 1;
		offset |= (jis & 0x0060) << 9;
		offset |= (jis & 0x001f) << 4;
	} else if (jis >= 0x7020 && jis <= 0x705f) {
		offset |= (jis & 0x0700) << 1;
		offset |= (jis & 0x0060) << 7;
		offset |= (jis & 0x001f) << 4;
	} else
		offset = -1;
	return offset;
}

static BYTE _n88_kanji_get(LPN88KANJIROM self, int hl) {
	USHORT jis = MAKER16(self->addr_h, self->addr_l);
	int offset = jis;
	offset <<= 1;
	offset += hl == 1 ? 1 : 0;
	if (offset < self->length)
		return self->rom[offset];
	return 0xff;
}

static void _n88_kanji_reset(void) {
}

static void _n88_kanji_dispose(void) {
	for (int bank = 0; bank < 2; ++bank) {
		if (n80device.kanji.rom[bank].free != NULL) {
			n80device.kanji.rom[bank].free(&n80device.kanji.rom[bank]);
			n80device.kanji.rom[bank].ready = 0;
		}
	}
}

static void _n88_kanji_begin(void) {
	_n88_kanji_reset();
}

static void _n88_kanji_end(void) {
	_n88_kanji_reset();
}

static BYTE _n88_kanji_read(int bank, int hilo) {
	if (n80device.kanji.rom[bank].ready == 1)
		return n80device.kanji.rom[bank].get(&n80device.kanji.rom[bank], hilo);
	else
		return 0xff;
}

static void _n88_kanji_write(int bank, int hilo, BYTE addr) {
	if (hilo == 0)
		n80device.kanji.rom[bank].addr_l = addr;
	else
		n80device.kanji.rom[bank].addr_h = addr;
}

export void _n88_init_kanji(void) {
	n80device.kanji.reset		= _n88_kanji_reset;
	n80device.kanji.dispose		= _n88_kanji_dispose;
	n80device.kanji.begin		= _n88_kanji_begin;
	n80device.kanji.read		= _n88_kanji_read;
	n80device.kanji.write		= _n88_kanji_write;
	n80device.kanji.end			= _n88_kanji_end;
	n80device.kanji.jis2offset	= _n88_jis_to_offset;
	n80device.kanji.rom[0].get	= _n88_kanji_get;
	n80device.kanji.rom[1].get	= _n88_kanji_get;
	n80device.kanji.reset();
}
#ifdef __cplusplus
}	// extern "C"
#endif
