/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#include "PC-8001.h"
#ifdef __cplusplus
extern "C"
{
#endif
static __inline void gvram_wait(DWORD* W, BYTE dispbit) {
	n80device.crtc._gvram_access += 1;
}
/**--------------------------------------------------------------------------
 **		ASM Modules
 **------------------------------------------------------------------------*/
static __inline BYTE _n80_mbread_low(USHORT Adr, DWORD *W) {
	switch ((n80device.io._o31 & 0x06) >> 1) {
		case 0:		//N88-ROM
			return n80device.mem._rom88[Adr];
		case 2:		//N80-ROM
			return n80device.mem._rom80[Adr];
		default:	//RAM
			return n80device.mem._ram[Adr];
	}
}
	
static BYTE __cdecl _n88_mbread(LPZ80CPU cpu, USHORT Adr, DWORD *W) {
	BYTE mb = 0xff;
	//n80device.crtc._gvram_access = 0;
	switch (Adr & 0xf000) {
		case 0x0000:
		case 0x1000:
		case 0x2000:
		case 0x3000:
		case 0x4000:
		case 0x5000:
			mb = _n80_mbread_low(Adr, W);
			break;
		case 0x6000:
		case 0x7000:
			if ((n80device.io._o31 & 0x06) == 0) {	//N88-ROM selected
				BYTE rom4th = ~n80device.io._o71;
				if (rom4th == 0)					//all 4th bank disabled
					mb = n80device.mem._rom88[Adr];
				else if ((rom4th & 0x01) == 0x01)	//N88-4th ROM selected
					mb = n80device.mem._4th[Adr-0x6000U];
				else
					mb = 0xff;
			} else
				mb = _n80_mbread_low(Adr, W);
			break;
		case 0x8000: {
			USHORT adr = Adr;
			if (Adr < 0x8400U && (n80device.io._o31 & 0x06) == 0) {	//Text Window
				adr = (USHORT)n80device.io._o70;
				adr <<= 8;
				adr += (Adr & 0x3ff);
			}
			mb = n80device.mem._ram[adr];
		}	break;
		case 0x9000:
		case 0xa000:
		case 0xb000:
			mb = n80device.mem._ram[Adr];
			break;
		case 0xc000:
		case 0xd000:
		case 0xe000:
		case 0xf000:
			if (n80device.io._i5C & _GVRAM_STS_G0) {
				mb = n80device.mem._gvram0[Adr-0xc000U];
				gvram_wait(W, 2);
			} else if (n80device.io._i5C & _GVRAM_STS_G1) {
				mb = n80device.mem._gvram1[Adr-0xc000U];
				gvram_wait(W, 4);
			} else if (n80device.io._i5C & _GVRAM_STS_G2) {
				mb = n80device.mem._gvram2[Adr-0xc000U];
				gvram_wait(W, 8);
			} else {
				mb = n80device.mem._ram[Adr];
			}
			break;
		default:
			break;
	}
	return mb;
}
    
static void __cdecl _n88_mbwrite(LPZ80CPU cpu, USHORT Adr, BYTE v, DWORD *W) {
	//n80device.crtc._gvram_access = 0;

	USHORT adr = Adr;
	switch (Adr & 0xf000) {
		case 0x0000:
		case 0x1000:
		case 0x2000:
		case 0x3000:
		case 0x4000:
		case 0x5000:
		case 0x6000:
		case 0x7000:
			break;
		case 0x8000:
			if (Adr < 0x8400U && (n80device.io._o31 & 0x06) == 0) {	//Text Window
				adr = (USHORT)n80device.io._o70;
				adr <<= 8;
				adr += (Adr & 0x3ffU);
			}
			break;
		case 0x9000:
		case 0xa000:
		case 0xb000:
			break;
		case 0xc000:
		case 0xd000:
		case 0xe000:
		case 0xf000:
			if (n80device.io._i5C & _GVRAM_STS_G0) {
				n80device.vram_sync_enter();
				n80device.mem._gvram0[Adr-0xc000U] = v;
				n80device.vram_sync_leave();
				gvram_wait(W, 2);
				return;
			} else if (n80device.io._i5C & _GVRAM_STS_G1) {
				n80device.vram_sync_enter();
				n80device.mem._gvram1[Adr-0xc000U] = v;
				n80device.vram_sync_leave();
				gvram_wait(W, 4);
				return;
			} else if (n80device.io._i5C & _GVRAM_STS_G2) {
				n80device.vram_sync_enter();
				n80device.mem._gvram2[Adr-0xc000U] = v;
				n80device.vram_sync_leave();
				gvram_wait(W, 8);
				return;
			}
			break;
		default:
			break;
	}
	if (adr >= n80device.dmac._begin && adr < n80device.dmac._end) {
		// Sync Video RAM Area
		n80device.vram_sync_enter();
		n80device.mem._ram[adr] = v;
		n80device.attr._attrchg[((adr - n80device.dmac._begin) / 120)] = 1;
		n80device.vram_sync_leave();
	} else
		n80device.mem._ram[adr] = v;
}

static void clear_ram_core(LPBYTE mem) {
	int  i;

	for (i = 0; i < 0x1000; ++i) {
		*mem++ = 0x00;
		*mem++ = 0xff;
	}
	for (i = 0; i < 0x1000; ++i) {
		*mem++ = 0xff;
		*mem++ = 0x00;
	}
	for (i = 0; i < 128; ++i) {
		int  j;

		for (j = 0; j < 64; ++j)
			*mem++ = 0xff;
		for (j = 0; j < 64; ++j)
			*mem++ = 0x00;
	}
}

static void _n80_clear_memory(void) {
	LPBYTE	mem = n80device.mem._ram;
	clear_ram_core(mem);
	clear_ram_core(mem + 0x8000);
}

export void _n80_init_memory(void) {
	n80device.mem.read	= _n88_mbread;
	n80device.mem.write	= _n88_mbwrite;
	n80device.mem.clear	= _n80_clear_memory;
	_n80_clear_memory();
}
#ifdef __cplusplus
}	// extern "C"
#endif
