//
//  PB8214.c
//  N88X
//
//  Created by hiroshi on 2016/05/01.
//  Copyright © 2016年 H.Kanemoto. All rights reserved.
//
#include "pc-8001.h"
#ifdef _MSC_VER
#include <windows.h>
#include <crtdbg.h>
#endif
#include <stdio.h>
#ifdef __cplusplus
extern "C"
{
#endif
#define export
#define INT_ACCEPTED		(0)
#define INT_DENY_BY_CPU		(1)		// DI
#define INT_DENY_BY_LEVEL	(2)		// enable Level
#define INT_DENY_BY_MASK	(3)		// masked
#define INT_HOLD_0_USART	(1)
#define INT_HOLD_1_VRTC		(2)
#define INT_HOLD_2_RTC		(4)
//---------------------------------------------------------------------------
//		Debug print
//---------------------------------------------------------------------------
#if _DEBUG
typedef struct req_count_tag {
	int	req_count;
	int	accepted;
	int	deny_cpu;
	int	deny_level;
	int	deny_mask;
	int	overrun;
	int	resend;
}	req_count, *preq_count;
static req_count	reqcnt[3];
static void show_debug_info(int idx, int maxcnt, const char *devint) {
	preq_count r = &reqcnt[idx];
	if (r->req_count >= maxcnt) {
		double percent = ((double)r->accepted) / ((double)r->req_count) * 100;
		double percent2 = ((double)r->deny_level) / ((double)r->req_count) * 100;
		double percent3 = ((double)r->deny_mask) / ((double)r->req_count) * 100;
		double percent4 = ((double)r->overrun) / ((double)r->req_count) * 100;
		double percent5 = ((double)r->resend) / ((double)r->req_count) * 100;
		//n80device.logger("%s interrupt %.1f%% (level:%.1f%%, mask:%.1f%%, over:%.1f%%, re-send:%.1f%%)\n",
		//	devint, percent, percent2, percent3, percent4, percent5);
		r->req_count = 0;
		r->accepted = 0;
		r->deny_level = 0;
		r->deny_mask = 0;
		r->overrun = 0;
		r->resend = 0;
	}
}
#endif
//---------------------------------------------------------------------------
//		Request Interrupt
//---------------------------------------------------------------------------
static int interrupt_request(LPZ80CPU cpu, BYTE mask, BYTE lev) {
	//check interrupt disabled
	if ((n80device.io._oE6 & mask) == 0)
		return INT_DENY_BY_MASK;
	//check interrupt enabled level
	if ((n80device.io._oE4 & 0x08) == 0) {
		BYTE level = n80device.io._oE4 & 0x07;
		if (level > lev)
			;
		else
			return INT_DENY_BY_LEVEL;
	}
	return cpu->int_signal(cpu, lev);
}
//---------------------------------------------------------------------------
//		re-send holding Interrupt
//---------------------------------------------------------------------------
static void intx_resend(LPZ80CPU cpu) {
	if (n80device.pic._holding == 0)
		return;
	if (n80device.pic._holding & INT_HOLD_0_USART) {
		switch (interrupt_request(cpu, 0x04, 0)) {
			case INT_ACCEPTED:		// accepted
				n80device.pic._holding &= ~INT_HOLD_0_USART;
				break;
			case INT_DENY_BY_CPU:	// DI
			case INT_DENY_BY_LEVEL:	// Level
			case INT_DENY_BY_MASK:	// Masked
				break;
		}
		return;
	}
	if ((n80device.pic._holding & 3) == INT_HOLD_1_VRTC) {
#if _DEBUG
		++reqcnt[1].resend;
#endif
		switch (interrupt_request(cpu, 0x02, 1)) {
			case INT_ACCEPTED:		// accepted
#if _DEBUG
				++reqcnt[1].accepted;
#endif
				n80device.pic._holding &= ~INT_HOLD_1_VRTC;
				break;
			case INT_DENY_BY_CPU:	// DI
				break;
			case INT_DENY_BY_LEVEL:	// Level
			case INT_DENY_BY_MASK:	// Masked
				n80device.pic._holding &= ~INT_HOLD_1_VRTC;
				break;
		}
		return;
	}
	if ((n80device.pic._holding & 7) == INT_HOLD_2_RTC) {
#if _DEBUG
		++reqcnt[2].resend;
#endif
		switch (interrupt_request(cpu, 0x01, 2)) {
			case INT_ACCEPTED:		// accepted
#if _DEBUG
				++reqcnt[2].accepted;
#endif
				n80device.pic._holding &= ~INT_HOLD_2_RTC;
				break;
			case INT_DENY_BY_CPU:	// DI
				break;
			case INT_DENY_BY_LEVEL:	// Level
			case INT_DENY_BY_MASK:	// Masked
				n80device.pic._holding &= ~INT_HOLD_2_RTC;
				break;
		}
	}
}
//---------------------------------------------------------------------------
//		Lv0: USART ready(rxd) interrupt
//---------------------------------------------------------------------------
static void int0_request(LPZ80CPU cpu) {
	n80device.pic._irq_0 = 1;
	if (interrupt_request(cpu, 0x04, 0) != 0)
		n80device.pic._holding |= INT_HOLD_0_USART;
}
//---------------------------------------------------------------------------
//		Lv1: VRTC vertical blanking interrupt
//---------------------------------------------------------------------------
static void int1_request(LPZ80CPU cpu) {
#if _DEBUG
	++reqcnt[1].req_count;
#endif
	if (n80device.pic._holding & INT_HOLD_1_VRTC) {
#if _DEBUG
		++reqcnt[1].overrun;
#endif
		return;
	}
	switch (interrupt_request(cpu, 0x02, 1)) {
		case INT_ACCEPTED:
#if _DEBUG
			++reqcnt[1].accepted;
#endif
			break;
		case INT_DENY_BY_CPU:
			n80device.pic._holding |= INT_HOLD_1_VRTC;
			break;
		case INT_DENY_BY_LEVEL:	//level
#if _DEBUG
			++reqcnt[1].deny_level;
#endif
			break;
		case INT_DENY_BY_MASK:	//masked
#if _DEBUG
			++reqcnt[1].deny_mask;
#endif
			break;
	}
#if _DEBUG
	//show_debug_info(1, 600, "VRTC vb");
#endif
}
//---------------------------------------------------------------------------
//		Lv2: RTC(1/600sec) interrupt
//---------------------------------------------------------------------------
static void int2_request(LPZ80CPU cpu) {
#if _DEBUG
	++reqcnt[2].req_count;
#endif
    if (n80device.pic._holding & INT_HOLD_2_RTC) {
#if _DEBUG
		++reqcnt[2].overrun;
#endif
    } else {
        switch (interrupt_request(cpu, 0x01, 2)) {
            case INT_ACCEPTED:
#if _DEBUG
				++reqcnt[2].accepted;
#endif
				break;
            case INT_DENY_BY_CPU:
                n80device.pic._holding |= INT_HOLD_2_RTC;
                break;
            case INT_DENY_BY_LEVEL: //level
#if _DEBUG
				++reqcnt[2].deny_level;
#endif
				break;
            case INT_DENY_BY_MASK:	//masked
#if _DEBUG
				++reqcnt[2].deny_mask;
#endif
				break;
        }
    }
#if _DEBUG
	//show_debug_info(2, 4800, "RTC clock");
#endif
}
	
static void _n80_reset_pic(void) {
	n80device.pic._holding	= 0;
	n80device.pic._result	= 0;
	n80device.pic._irq_0	= 0;
#if _DEBUG
	memset(reqcnt, 0, sizeof(reqcnt));
#endif
}

export void _n80_init_pic(void) {
	n80device.pic.reset		= _n80_reset_pic;
	n80device.pic.resend	= intx_resend;
	n80device.pic.int0		= int0_request;
	n80device.pic.int1		= int1_request;
	n80device.pic.int2		= int2_request;
	_n80_reset_pic();
}
	
#ifdef __cplusplus
}	// extern "C"
#endif
