/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#ifndef _PC_8001_H_
#define _PC_8001_H_
#include <stdarg.h>
#include "../z80/z80.h"
#include "../PC-8031/PC-8031.h"
#ifdef __cplusplus
extern "C"
{
#endif
#define export
/**--------------------------------------------------------------------------
 **		vg^Cv錾
 **------------------------------------------------------------------------*/
struct N80DEVICE_tag;
typedef struct N80DEVICE_tag N80DEVICE, *LPN80DEVICE;
/**--------------------------------------------------------------------------
 **		萔`
 **------------------------------------------------------------------------*/
#define _PO31_GRAPH	(0x08)		// enable graphic
#define _PO40_BEEP	(0x20)		// Beep bit
#define _PO40_CLDS	(0x04)		// Calender data bit
#define _PO40_FLASH	(0x10)		// GVRAM flash mode
#define _PO40_CLK	(0x02)		// Calender clock
#define _PO40_CSTB	(0x01)		// Calender strobe
#define	_VRTC_DISP	(0x00)		// crt display
#define _VRTC_VBBGN	(0x01)		// crt enter vertical blanking
#define	_VRTC_VBEND	(0x02)		// crt leave vertical blanking
#define _TXTVIS_TXTDS		(1)	// text visibility port 53h bit 0
#define _TXTVIS_STOPDISP	(2)	// CRTC stop display
#define _TXTVIS_STOPDMA		(4)	// DMAC ch.2 disabled
#define _TXTVIS_STOPALL		(6)	// CRTC/DMAC both stop display
#define _GVRAM_STS_G0		(1)	// GVRAM status G-VRAM #0 bank selected
#define _GVRAM_STS_G1		(2)	// GVRAM status G-VRAM #1 bank selected
#define _GVRAM_STS_G2		(4)	// GVRAM status G-VRAM #2 bank selected
#define _GVRAM_STS_GVRAM	(7)	// GVRAM status any G-VRAM bank selected
#define _BOOT_ROM_N80       (0)
#define _BOOT_ROM_N88       (1)
#define _CRT_HIRES          (0)
#define _CRT_NORMAL         (1)
#define _CRTC_COLOR_MODE	(0)
#define _CRTC_MONO_MODE		(1)
#define _ATTR_COLOR_GREEN	(0x80)
#define _ATTR_COLOR_RED		(0x40)
#define _ATTR_COLOR_BLUE	(0x20)
#define _ATTR_COLOR_GRB		(0xe0)
#define _ATTR_MONO_GRAPH	(0x80)
#define _ATTR_COLOR_GRAPH	(0x10)
#define _ATTR_UNDER_LINE	(0x20)
#define _ATTR_OVER_LINE		(0x10)
#define _ATTR_FNC_COLOR		(0x08)
#define _ATTR_FNC_REVERSE	(0x04)
#define _ATTR_FNC_BLINK		(0x02)
#define _ATTR_FNC_ERASE		(0x01)
#define _PAC_ATTR_BLUE		(0x01)
#define _PAC_ATTR_GREEN		(0x02)
#define _PAC_ATTR_RED		(0x04)
#define _PAC_ATTR_UPR_LINE	(0x08)
#define _PAC_ATTR_GRAPH		(0x10)
#define _PAC_ATTR_REVERSE	(0x20)
#define _PAC_ATTR_ERASE		(0x40)
#define _PAC_ATTR_UND_LINE	(0x80)
#define _PAC_ATTR_RGB		(0x07)
#ifdef __linux__
#define IVLSEC		(0.01)		// CPU loop interval in sec (default)
#else
#define IVLSEC		(0.02)		// CPU loop interval in sec
#define IVLMSEC		(20)		// CPU loop interval in msec
#endif
/**--------------------------------------------------------------------------
 **		\
 **------------------------------------------------------------------------*/
typedef struct N80MEMORY_tag {	// Memory map
	BYTE	_rom88[32768];		// N88-ROM
	BYTE	_rom80[32768];		// N80-ROM
	BYTE	_4th[8192];			// 4th ROM
	BYTE	_ram[64*1024];		// RAM
	BYTE	_gvram0[16*1024];	// V-RAM bank#0
	BYTE	_gvram1[16*1024];	// V-RAM bank#1
	BYTE	_gvram2[16*1024];	// V-RAM bank#2
	void	(*init)(void);
	void	(*clear)(void);
	LPZ80FNCMBR	read;
	LPZ80FNCMBW write;
}	N80MEM, *LPN80MEM;

typedef struct N88DIP1_tag {	// DIP switch #1 (port 30H)
	BYTE	SW1_N_ROM;			// 0:N-BASIC, 1:N88-BASIC
	BYTE	SW2_TERMINAL;		// 0:Terminal, 1:BASIC
	BYTE	SW3_WIDTH;			// 0:80, 1:40
	BYTE	SW4_LINE;			// 0:25, 1:20
	BYTE	SW5_S_PARAM;		// 0:enabled, 1:disabled
	BYTE	SW6_DEL_CODE;		// 0:process, 1:ignore
	BYTE	UIP0;
	BYTE	UIP1;
}	N88DIPSW1, *LPN88DIPSW1;

typedef struct N88KANJIROM_tag {
	BYTE	ready;				// Kanji ROM ready(1) or not(0)
	LPBYTE	rom;				// Kanji ROM data
	BYTE	addr_h;				// JIS CODE (hi)
	BYTE	addr_l;				// JIS CODE (lo)
	int		length;
	void	(*free)(struct N88KANJIROM_tag* self);
	BYTE	(*get)(struct N88KANJIROM_tag *self, int hl);
}	N88KANJIROM, *LPN88KANJIROM;

typedef struct N88KANJI_tag {
	N88KANJIROM	rom[2];
	void	(*init)(void);
	void	(*reset)(void);
	void	(*dispose)(void);
	void	(*begin)(void);
	BYTE	(*read)(int bank, int hl);
	void	(*write)(int bank, int hl, BYTE);
	void	(*end)(void);
	int		(*jis2offset)(BYTE jish, BYTE jisl);
}	N88KANJI, *LPN88KANJI;

typedef struct N80IOPORT_tag {
	BYTE	_i40;
	BYTE	_i5C;
	BYTE	_i71;				// 4th ROM select (corrected)

	BYTE	_o10;				// PRINTER/CALENDER data
	BYTE	_o20;				// USART data
	BYTE	_o21;				// USART command
	BYTE	_o30;				// CONTROL #1
	BYTE	_o31;				// CONTROL #2
	BYTE	_o40;				// STROBE
	BYTE	_o50;				// CRTC parameter
	BYTE	_o51;				// CRTC command
	BYTE	_o52;				// boarder/background color
	BYTE	_o53;				// TEXT/GVRAM show flag
	BYTE	_o70;				// TEXT WINDOW offset address
	BYTE	_o71;				// 4th ROM select (raw)
	BYTE	_oE4;				// interrupt leveled mask
	BYTE	_oE6;				// interrupt enable/disable mask
	BYTE	keys[12];			// key on/off bit matrix datas
	void	(*init)(void);
	void	(*clear)(void);
	void	(*clear_keys)(void);
	LPZ80FNCIOR read;
	LPZ80FNCIOW write;
}	N80IOPORT, *LPN80IOPORT;

typedef struct N80CMT_tag {
	int		_baud;				// quasy baud rate
	BYTE	_reqcas;			// request device time for CMT
	LPBYTE	_buff;				// cmt data buff ptr
	DWORD	_size;				// buff size
	DWORD	_index;				// buff current offset ptr
	BYTE	_reset;				// reset
	void	(*init)(void);
	void	(*reset)(void);
	void	(*control)(void);
	void	(*notify)(BYTE);
	void	(*process)(LPZ80CPU);
	void	(*fastload)(void);
	void	(*flush)(void);
	BYTE	(*casdata)(void);
}	N80CMT, *LPN80CMT;
	
typedef struct N80USART_tag {	// uPD8251 USART
	BYTE	_ir;				// USART reset (mode set)
	BYTE	_fake;				// dummy data if reset
	BYTE	_rxe;				// recieve enabled
	BYTE	_rxd;				// recieve data
	BYTE	_rxrdy;				// recieve buffer full
	BYTE	_ovr;				// over run error occured
	void	(*init)(void);
	void	(*reset)(void);
	void	(*cmd)(void);
	BYTE	(*status)(void);
	void	(*rxd)(LPZ80CPU,BYTE);
	BYTE	(*read)(void);
	void	(*write)(BYTE);
}	N80USART, *LPN80USART;

typedef struct N80PD1990_tag {	// uPD1990 calender
	int			_sec;			// uPD1990 40bits shift register
	int			_min;
	int			_hour;
	int			_day;
	int			_month;
	ULONGLONG	_shift_reg;		// ڼ޽
	ULONGLONG	_time_counter;	// Ѷ
	ULONGLONG	_clk_counter;	// clk
	BYTE	_cstrb;				// ޺޽۰  2:ON,0:OFF
	BYTE	_cclk;				// ޼ĸۯ      4:ON,0:OFF
	BYTE	_shiften;			// shift enable flag 1:EN,0:DIS
	void	(*init)(void);
	void	(*reset)(void);
	void	(*mode)(void);
	void	(*clk)(int tick);
}	N80PD1990, *LPN80PD1990;

typedef struct N80CRTC_tag {
	BYTE	_burst_mode;		// DMA burst mode 0:off/1:on
	BYTE	_chars_per_line;	// 2..129 chars per line
	BYTE	_lines_per_screen;	// lines count
	BYTE	_raster_per_char;	// raster lines per char
	BYTE	_ver_ret_lines;		// vertical return lines
	BYTE	_hol_ret_chars;		// holizontal return chars
	BYTE	_attribute_en;		// attribute 0:disabled/1:enabled
	BYTE	_attribute_mono;	// attribute 0:color/1:mono
	BYTE	_attribute_tp;		// attribute 0:non-transparent/1:transparent
	BYTE	_special_char;		// special char 0:disabled/1:enabled
	BYTE	_max_attr_cnt;		// max attribute counts
	USHORT	_crtc_line_bytes;	// char+attribute+sc size
	DWORD	_screen_mode;		// 0:40/20, 1:40/25, 2:80/20, 3:80/25
	BYTE	_color_mode;		// color mode        0:color, 1:mono
	BYTE	_width_mode;		// 80/40 colmun mode 0:40, 1:80
	BYTE	_line_mode;			// 20/25 line mode   0:20, 1:25
	BYTE	_attr_reverse;		// attribute color reverse mode flag
	USHORT	_param_count;		// amount of params already accepted.
	BYTE	_prev_cmd;			// current command which is waiting for params
	BYTE	_text_visible;		// text visible: 0:visible, not 0:invisible (b0:TXTDS, b1:crtc stop, b2:dmac stop)
	BYTE	_gvram_access;		// gvram read/write occured
	void	(*init)(void);
	void	(*reset)(void);
	void	(*cmd)(BYTE);		// uPD1330 command
	void	(*data)(BYTE);		// uPD1330 data
	void	(*mode)(BYTE);		// width mode change
}	N80CRTC, *LPN80CRTC;

typedef struct N80CURSOR_tag {
	USHORT	_cur_blink_cycl;	// cursor blinking cycle
	USHORT	_att_blink_cycl;	// attribute blinking cycle
	USHORT	_cur_blink_disp;	// cursor blinking display period
	USHORT	_att_blink_disp;	// attribute blinking display period
	BYTE	_col;				// cursor location x
	BYTE	_row;				// cursor location y
	BYTE	_new_col;			// cursor new location x
	BYTE	_new_row;			// cursor new location y
	BYTE	_enabled;			// cursor enable
	BYTE	_blinking;			// cursor blink status
	BYTE	_attr_blinking;		// attribute blink status
	void	(*init)(void);
	void	(*reset)(void);
	void	(*enable)(void);
	void	(*disable)(void);
	void	(*move)(void);
}	N80CURSOR, *LPN80CURSOR;

typedef struct N80VRTC_tag {
	BYTE	_hsync_start_flg;	// hsync start 1:yes, 0:no
	BYTE	_vrtc_flag;			// vertical blanking flag
	BYTE	_holz_flag;			// holizontal blanking flag
	USHORT	_dma_remain_chars;	// holizontal remain chars via dma transfer
	DWORD	_vrtc_blink;		// blanking interval
	DWORD	vsync_line;			// current line
	DWORD	vsync_state;		// current vsync T-state counter
	DWORD	hsync_state;		// current hsync T-state counter
	DWORD	hsync_dsp_state;	// h-freq disp:44.70us=179T/blank:17.88us=71T
	DWORD	hsync_all_state;	// h-freq 62.58us=250T()
	DWORD	vsync_bck_state;	// v-freq 16.02ms=64080T(disp:12.52ms=50080T/blank:3.504ms=14016T)
    DWORD   vsync_dsp_rasters;  // raster lines (display)
	double	gvram_wait_remain;	// from X88000 source as a reference
	double	gvram_wait_rate;
	double	gvram_wait_rate2;
	int     crt_hires_mode;     // 0:HiRes(24kHz)/1:Normal(15kHz)
	void	(*init)(void);			// initialize
	void	(*reset)(void);			// reset
	int		(*blanking)(LPZ80CPU);	// blanking operation
	int		(*gvram_wait)(LPZ80CPU);// gvram wait
	void	(*edge)(void);			// notify entering vertical blank
	BYTE	(*poll)(BYTE);			// poll vertical blank
}	N80VRTC, *LPN80VRTC;

typedef struct N80DMACH_tag {
	BYTE	_addr_lo;
	BYTE	_addr_hi;
	BYTE	_addr_expect_hl;	// expect addr 0:lo/1:hi
	BYTE	_cnt_lo;
	BYTE	_cnt_hi;
	BYTE	_cnt_expect_hl;		// expect cnt 0:lo/1:hi
	BYTE	_wr_mode;			// 00:verify, 01:write, 10:read
	USHORT	_addr;				// ch address
	USHORT	_cnt;				// terminal count
	USHORT	_remain;
}	N80DMAC_CH, *LPN80DMAC_CH;

typedef struct N80DMAC_tag {
	DWORD	_penarty;			// DMA transfer penarty
	DWORD	_begin;				// DMA begin addr
	DWORD	_end;				// DMA end addr
	BYTE	_autoload;			// DMA autoload mode 0:off/1:on
	BYTE	_dmaact;			// DMA active flag
	struct N80DMACH_tag ch[4];	// DMA 4 channels
	void	(*init)(void);		// initialize
	void	(*reset)(void);		// reset
	void	(*cmd)(BYTE);		// DMA command
	void	(*ch_adr)(int, BYTE);	// ch.N(0..3) addr param
	void	(*ch_cnt)(int, BYTE);	// ch.N(0..3) counter param
}	N80DMAC, *LPN80DMAC;

typedef struct N80ATTR_tag {
	const BYTE* grb_rgb;		// GRB to RGB convert table
	BYTE	_eval_attr[80];		// evaluated attribute
	BYTE	_textcach[80];		// text cach
	BYTE	_attrcach[40];		// attr cach
	BYTE	_attrchg[25];		// last changed attr
	BYTE	_sortstk[22];		// attr area sort stack
	BYTE	_headzero;			// area start zero flag
	BYTE	_nextcolor[25];		// attr effective color
	BYTE	_nextfunc[25];		// attr effective func
	BYTE	_prevcolor;			// previous color
	BYTE	_prevfunc;			// previous func
	void	(*init)(void);
	void	(*reset)(void);
	int		(*eval)(int);		// eval attribute
	void	(*cache)(int);		// copy vram to cache array
	void	(*chg_converter)(int);	// change attribute converter function
	BYTE	(*_convfunc)(BYTE);	// pointer to attribute converter function
	void	(*invalid)(void);	// set all attrchg invalid
	void	(*set)(int, BYTE);	// set attrchg of line
}	N80ATTR, *LPN80ATTR;

typedef struct N80PCG_tag {
	USHORT	_pcg_freq[3];		// PCG i8253 timmer #2 (adjust)
	BYTE	_timmr_hi[3];		// PCG i8253  low byte
	BYTE	_timmr_lo[3];		// PCG i8253  low byte
	BYTE	_timmr_mode[3];		// PCG i8253 mode register
	BYTE	_timmr_cont[3];
	BYTE	_pcg_lo;			// PCG RAM  low address
	BYTE	_pcg_hi;			// high address
	BYTE	_pcg_dat;			// PCG char data
	BYTE	speaker[3];			// PCG speaker on/off flags
	void	(*init)(void);
	void	(*mode)(BYTE, BYTE);
}	N80PCG, *LPN80PCG;

typedef struct N80PIC_tag {		// uPB8214 PIC
	BYTE	_holding;			// waiting INTs flag
	BYTE	_result;			// result of irq
	BYTE	_irq_0;				// INT0 flag
	void	(*init)(void);		// initialize
	void	(*reset)(void);		// reset
	void	(*resend)(LPZ80CPU);// re-request int
	void	(*int0)(LPZ80CPU);	// request int 0
	void	(*int1)(LPZ80CPU);	// request int 1
	void	(*int2)(LPZ80CPU);	// request int 2
}	N80PIC, *LPN80PIC;

typedef struct N80RTC_tag {
	int		count;
	void	(*clk)(LPN80DEVICE);
}	N80RTC, *LPN80RTC;

typedef struct N80BLT_tag {
	void	(*init)(void);
	void	(*chg_crt_mode)(int);
	void	(*chg_palette)(BOOL);
#ifdef __linux__
	void	(*reflesh)(void);
#elif _MSC_VER
	void	(*font)(int, int);
	void	(*blank)(int);
	void	(*border)(int);
	void	(*drawgvram)(int);
#endif
}	N80BLT, *LPN80BLT;

typedef struct N80SOUND_tag {
	void	(*init)(void);
	void	(*beep_on)(void);
	void	(*beep_off)(void);
	void	(*sound_on)(USHORT freq, int ch);
	void	(*sound_off)(void);
	void 	(*append_ticks)(DWORD T);
	void	(*clear_ticks)(void);
#ifdef __linux__
	int 	(*make_pcm)(void);
#endif
}	N80SND, *LPN80SND;

typedef struct N80EMUPARAM_tag {
	DWORD		hz;				// System Clock 4MHz
#ifndef __linux__
	DWORD		sync_interval;	// speed sync interval in msec
#else
	double		interval_sec;	// speed sync interval in sec
#endif
	DWORD		sync_ticks;		// speed sync interval in Z80 state clocks
	BOOL		interlace;		// interlace-like effect
	DWORD		gv_wait_ticks;	// gvram wait ticks (T)
	N88DIPSW1	dipsw1;			// dip-switch #1
}	N80EMUPARAM, *LPN80EMUPARAM;

struct N80DEVICE_tag {
	Z80CPU		cpu;			// PC-8801 main cpu
	N80MEM		mem;			// PC-8801 memory
	N88KANJI	kanji;			// kanji rom
	BYTE		font[4096];		// font datas
	BYTE		pcg[2048];		// pcg datas
	BYTE		pcg_mode;		// PCG ON/OFF mode
	BYTE		palette[8];		// Color Palette
	N80IOPORT	io;				// i/o port cache
	N80PD1990	calender;		// uPD1990 Calender
	N80CMT		cmt;			// uPD8251 USART
	N80USART	usart;			// uPD8251 USART
	N80CRTC		crtc;			// uPD1330 CRTC
	N80CURSOR	cursor;			// uPD1330 Cursor
	N80VRTC		vrtc;			// uPD1330 VRTC
	N80DMAC		dmac;			// DMAC
	N80ATTR		attr;			// uPD1330 Attribute
	N80PCG		hal;			// HAL PCG-8100
	N80PIC		pic;			// uPB8214 PIC
	N80RTC		rtc;			// RTC
	N80BLT		blt;			// draw screen
	N80SND		snd;			// sound driver
	N80EMUPARAM	param;
	FDU			fdu;			// Floppy Disk Drive Unit (PC-8031)

	volatile int	Ttns;		// T state int 10ns unit sec

	void		(*dispatch)(LPN80DEVICE);
	void		(*reset)(LPN80DEVICE);
	void		(*dispose)(LPN80DEVICE);
#if _DEBUG
	void		(*logger)(const char* fmt, ...);
#endif
#ifndef __linux__
	void		(*calc_clk_sync)(LPN80DEVICE dev, DWORD hz, DWORD interval);
#else
	void		(*calc_clk_sync)(struct N80DEVICE_tag * dev, DWORD hz, double interval);
#endif
	double		(*get_tick)(void);
	DWORD		(*msec_to_T)(LPN80DEVICE, double msec);
	void		(*vram_sync_enter)(void);
	void		(*vram_sync_leave)(void);
	char		save_cmt_path[512];	// save cmt path
};

typedef void (*SYNCFNC)(void);
/**--------------------------------------------------------------------------
 **		O[oϐ
 **------------------------------------------------------------------------*/
extern N80DEVICE	n80device;

#ifdef __cplusplus
}	// extern "C"
#endif
#endif//_PC_8001_H_
