/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#include "PC-8001.h"
#ifdef _MSC_VER
#include <windows.h>
#else
#include "../include/n80.h"
#endif
#ifdef __cplusplus
extern "C"
{
#endif
#define export
#define PCG_CTL_REG_BIT_SPK0	(0x08)
#define PCG_CTL_REG_BIT_SPK1	(0x40)
#define PCG_CTL_REG_BIT_SPK2	(0x80)
#define PCG_CTL_REG_BIT_STROBE	(0x10)
#define PCG_CTL_REG_BIT_HIADDR	(0x03)
#define PCG_CTL_REG_BIT_FNTSRC	(0x20)		// copy|move font bit pattern
#define PCG_8253_CTL_REG_SCx	(0xC0)		// affect channel
#define PCG_8253_CTL_REG_RWx	(0x30)		// mode of r/w counter value
/**--------------------------------------------------------------------------
 **		ASM Modules
 **------------------------------------------------------------------------*/
static __inline void set_pcg_freq(int ch)
{
	USHORT	timer;
	DWORD	freq, prev;

	timer = (((USHORT) n80device.hal._timmr_hi[ch]) << 8) + n80device.hal._timmr_lo[ch];
	// freq = 4MHz(CPU clock) / i8253 Counter
	if (timer == 0)
		timer = 1;	// 0Zh~
	freq = (DWORD) ((double) n80device.param.hz / (double) timer + 0.5);
	//if (freq < 24)
	//	freq = 24;
	//else if (freq > 4800)
	//	freq = 0;
	prev = n80device.hal._pcg_freq[ch];
	if (prev != freq) {
		n80device.hal._pcg_freq[ch] = (USHORT) freq;
		if (n80device.hal.speaker[ch] == 1) {
			n80device.snd.sound_on((USHORT)freq, ch);
		}
	}
}

static void _n80_pcg_mode(BYTE port, BYTE dat)
{
	USHORT	offset;
	int		ch;

	switch (port) {
	case 0x00:	// save PCG char data
				n80device.hal._pcg_dat = dat; break;
	case 0x01:	// PCG RAM low address
				n80device.hal._pcg_lo = dat; break;
	case 0x02:	// PCG sound on/off
				if (dat & PCG_CTL_REG_BIT_SPK0) {
					if (n80device.hal.speaker[0] == 0) {
						n80device.hal.speaker[0] = 1;
						n80device.snd.sound_on(n80device.hal._pcg_freq[0], 0);
					}
				} else {
					n80device.hal.speaker[0] = 0;
					n80device.snd.sound_on(0, 0);
					//n80device.sound_off();
				}
				if (dat & PCG_CTL_REG_BIT_SPK1) {
					if (n80device.hal.speaker[1] == 0) {
						n80device.hal.speaker[1] = 1;
						n80device.snd.sound_on(n80device.hal._pcg_freq[1], 1);
					}
				} else {
					n80device.hal.speaker[1] = 0;
					n80device.snd.sound_on(0, 1);
					//n80device.sound_off();
				}
				if (dat & PCG_CTL_REG_BIT_SPK2) {
					if (n80device.hal.speaker[2] == 0) {
						n80device.hal.speaker[2] = 1;
						n80device.snd.sound_on(n80device.hal._pcg_freq[2], 2);
					}
				} else {
					n80device.hal.speaker[2] = 0;
					n80device.snd.sound_on(0, 2);
					//n80device.snd.sound_off();
				}
				// PCG RAM high address
				if (dat & PCG_CTL_REG_BIT_STROBE) {
					n80device.hal._pcg_hi = dat & PCG_CTL_REG_BIT_HIADDR;	// address A8, A9 bits
					offset = (((USHORT) n80device.hal._pcg_hi) << 8) + n80device.hal._pcg_lo;
					if (dat & PCG_CTL_REG_BIT_FNTSRC)	// copy data from font rom
						n80device.pcg[1024 + offset] = n80device.font[1024 + offset];
					else {
					#ifdef _ZAURUS_
						BYTE	bit = 0x01 << (offset % 8);
						BYTE 	ubit = ~bit, dat = n80device.hal._pcg_dat;
						USHORT	i, offtop = offset & 0xfff8;
						for (i = 0; i < 8; ++i) {
							if (dat & 0x80)
								n80device.pcg[1024 + offtop + i] |= bit;
							else
								n80device.pcg[1024 + offtop + i] &= ubit;
							dat <<= 1;
						}
					#else
						n80device.pcg[1024 + offset] = n80device.hal._pcg_dat;
					#endif
					}
				}
				break;
	case 0x03:	// PCG ON/OFF
				n80device.pcg_mode = (dat & 0x08) ? 1 : 0;	break;
	case 0x0c:	// i8253 timmer #0 counter
	case 0x0d:	// i8253 timmer #1 counter
	case 0x0e:	// i8253 timmer #2 counter
				ch = port - 0x0c;
				switch (n80device.hal._timmr_mode[ch]) {
				case 3:	// load/store L/H order
						if (--n80device.hal._timmr_cont[ch] == 1) {
							n80device.hal._timmr_lo[ch] = dat;
						} else {
							n80device.hal._timmr_hi[ch] = dat;
							n80device.hal._timmr_cont[ch] = 2;
							set_pcg_freq(ch);
						}
						break;

				case 2:	// load/store H
						n80device.hal._timmr_hi[ch] = dat;
						set_pcg_freq(ch);
						break;

				case 1:	// load/store L
						n80device.hal._timmr_lo[ch] = dat;
						break;
				}
				break;

	case 0x0f:	// i8253 mode register
				ch = (dat & PCG_8253_CTL_REG_SCx) >> 6;	// get timer ch #0, #1,#2
				if (ch < 3) {	// ignore if read-back command is set
					switch ((dat & PCG_8253_CTL_REG_RWx) >> 4) {
					case 3:	n80device.hal._timmr_cont[ch] = 2;
							n80device.hal._timmr_mode[ch] = 3;
							break;
					case 2: n80device.hal._timmr_cont[ch] = 1;
							n80device.hal._timmr_mode[ch] = 2;
							n80device.hal._timmr_lo[ch] = 0;
							break;
					case 1: n80device.hal._timmr_cont[ch] = 1;
							n80device.hal._timmr_mode[ch] = 1;
							n80device.hal._timmr_lo[ch] = 0;
							break;
					default:
							n80device.hal._timmr_cont[ch] = 0;
							n80device.hal._timmr_mode[ch] = 0;
							break;
					}
				}
				break;
	}
}

export void _n80_init_pcg(void) {
	n80device.hal.mode = _n80_pcg_mode;
}
#ifdef __cplusplus
}	// extern "C"
#endif
