/**==========================================================================

	PC-8031 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2024 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	last modified  Apr,27,2024.

;==========================================================================*/
#include "../PC-8001/PC-8001.h"
#ifdef __cplusplus
extern "C"
{
#endif
#include <stdio.h>
#define export
/**--------------------------------------------------------------------------
**		PROTOTYPES
**------------------------------------------------------------------------*/
extern void _fdc_init_command(LPFDUCOMMAND);				// COMMAND.c
extern void _fdc_init_result(LPFDURESULT);					// RESULT.c
extern void _ses_start_seek_session(LPSESSION, LPFDU765A);	// SESSION.c
extern void _ses_start_specify_session(LPSESSION, LPFDU765A);
extern void _ses_start_recalibrate_session(LPSESSION, LPFDU765A);
extern void _ses_start_sns_int_sts_session(LPSESSION, LPFDU765A);
extern void _ses_start_sns_drv_sts_session(LPSESSION, LPFDU765A);
extern void _ses_start_read_data_session(LPSESSION, LPFDU765A);
extern void _ses_start_read_del_data_session(LPSESSION, LPFDU765A);
extern void _ses_start_read_diag_session(LPSESSION, LPFDU765A);
extern void _ses_start_read_id_session(LPSESSION, LPFDU765A);
extern void _ses_start_invalid_session(LPSESSION, LPFDU765A);
/**--------------------------------------------------------------------------
 **		FDC Phase Control
 **------------------------------------------------------------------------*/
static void _fdc_enter_idol_phase(LPFDU765A self) {
#if _DEBUG
	//n80device.logger("FDU: [IDOL]\n");
#endif
	self->phase   = FDC_PHASE_IDOL;
	self->status |= FDC_ST_RQM;
	self->status &= ~FDC_ST_CB;		// accept command
	self->status &= ~FDC_ST_DIO;	// CPU => FDC (CPU write data to FDC)
	//self->status &= 0xf0;			// clear FDD busy
	self->current = NULL;			// session complete
	for (int u = 0; u < FDC_MAX_UNIT_CNT; u++)
		self->owner->fdd.drives[u].state = '#';
}
static void _fdc_enter_cmd_phase(LPFDU765A self) {
#if _DEBUG
	//n80device.logger("FDU: [CMD]\n");
#endif
	self->phase   = FDC_PHASE_CMD;
	self->status |= FDC_ST_RQM;
	self->status |= FDC_ST_CB;		// reject command by Cmd/Result/Exe(Read/Write) phase
	self->status &= ~FDC_ST_DIO;	// CPU => FDC (CPU write data to FDC)
}
static void _fdc_enter_result_phase(LPFDU765A self) {
#if _DEBUG
	//n80device.logger("FDU: [RESULT]\n");
#endif
	self->phase   = FDC_PHASE_RESULT;
	self->status |= FDC_ST_RQM;
	self->status |= FDC_ST_CB;		// reject command by Cmd/Result/Exe(Read/Write) phase
	self->status |= FDC_ST_DIO;		// FDC => CPU (CPU read data from FDC)
	self->status &= ~FDC_ST_EXM;
}
static void _fdc_enter_exe_phase(LPFDU765A self, int unit) {
#if _DEBUG
	//n80device.logger("FDU: [EXE] \n");
#endif
	self->phase   = FDC_PHASE_EXE;
	//self->status  = (BYTE)(0x01 << unit);
	self->status |= (FDC_ST_CB | FDC_ST_EXM);
}
/**--------------------------------------------------------------------------
 **		Decode Command (Phase Session Start)
 **------------------------------------------------------------------------*/
static void _fdc_decode_command(LPFDU765A self, BYTE data) {
	BYTE code = self->cmd.push0(&self->cmd, data);
	switch (code) {
	case FDC_CMD_RD_DIAG:
		_ses_start_read_diag_session(&self->_session, self);
		break;
	case FDC_CMD_SPECIFY:
		_ses_start_specify_session(&self->_session, self);
		break;
	case FDC_CMD_SNS_DRV_STS:
		_ses_start_sns_drv_sts_session(&self->_session, self);
		break;
	case FDC_CMD_READ_DAT:
		_ses_start_read_data_session(&self->_session, self);
		break;
	case FDC_CMD_RECALIB:
		_ses_start_recalibrate_session(&self->_session, self);
		break;
	case FDC_CMD_SNS_INT_STS:
		_ses_start_sns_int_sts_session(&self->_session, self);
		break;
	case FDC_CMD_READ_ID:
		_ses_start_read_id_session(&self->_session, self);
		break;
	case FDC_CMD_READ_DEL_DAT:
		_ses_start_read_del_data_session(&self->_session, self);
		break;
	case FDC_CMD_SEEK:
		_ses_start_seek_session(&self->_session, self);
		break;
	default:
#if _DEBUG
		n80device.logger("FDU: [PD765A] ### Un-Handled Command %2.2X (%2.2X)\n", self->cmd.command_code, data);
#endif
		_ses_start_invalid_session(&self->_session, self);
		break;
	}
}
/**--------------------------------------------------------------------------
**		FDC(765A) registers
**------------------------------------------------------------------------*/
static BOOL _fdc_read_data_ready(LPFDU765A self) {
	return (self->status & (FDC_ST_RQM | FDC_ST_DIO)) == (FDC_ST_RQM | FDC_ST_DIO);
}
/**--------------------------------------------------------------------------
**		FDC(765A) write data register
**------------------------------------------------------------------------*/
static void _fdc_write_data_reg(LPFDU765A self, BYTE data) {
	self->status &= ~FDC_ST_RQM;		// when write reg PC=>FDC(~WR="0") then RQM="0"
	switch (self->phase) {
	case FDC_PHASE_IDOL:
		_fdc_decode_command(self, data);
		self->status |= FDC_ST_RQM;		// when FDC recieve from reg then RQM="1"
		break;
	case FDC_PHASE_CMD:
		if (self->current != NULL) {
			self->current->command(self->current, data);
			self->status |= FDC_ST_RQM;	// when FDC recieve from reg then RQM="1"
		}
#if _DEBUG
		else
			n80device.logger("FDU: [PD765A] invalid write data access, duaring cmd-phase.\n");
#endif
		break;
	case FDC_PHASE_EXE:
#if _DEBUG
		n80device.logger("FDU: [PD765A] invalid write data access, duaring exe-phase.\n");
#endif
		break;
	case FDC_PHASE_RESULT:
#if _DEBUG
		n80device.logger("FDU: [PD765A] invalid write data access, duaring result-phase.\n");
#endif
		break;
	}
}
/**--------------------------------------------------------------------------
 **		PD765A read data register
 **------------------------------------------------------------------------*/
static BYTE _fdc_read_data_reg(LPFDU765A self) {
	BYTE v = 0xff;
	if (self->rd_ready(self)) {
		if (self->current != NULL && (
			self->phase == FDC_PHASE_RESULT || self->phase == FDC_PHASE_EXE)) {
			return self->current->read_data(self->current);
		}
	}
#if _DEBUG
	else {
		n80device.logger("FDU: [PD765A] unexpected read data.\n");
	}
#endif
	return v;
}
/**--------------------------------------------------------------------------
 **		PD765A status register
 **------------------------------------------------------------------------*/
static BYTE _fdc_read_status_reg(LPFDU765A self) {
	return self->status;
}
/**--------------------------------------------------------------------------
 **		PD765A TC pin
 **------------------------------------------------------------------------*/
static void _fdc_terminal_count_pin(LPFDU765A self) {
	if (self->current != NULL)
		self->current->clk(self->current, 0, FDC_CLK_BY_TC);
}
/**--------------------------------------------------------------------------
 **		PD765A reset
 **------------------------------------------------------------------------*/
static void _fdc_reset(LPFDU765A self) {
	self->status		= FDC_ST_RQM;
	self->phase			= FDC_PHASE_IDOL;
	self->current		= NULL;
	self->cmd.clear(&self->cmd);
	self->result.clear(&self->result);
	self->last_result.clear(&self->last_result);
}
/**--------------------------------------------------------------------------
 **		PD765A
 **------------------------------------------------------------------------*/
export void _fdu_init_fdc(LPFDU765A self, LPFDU owner) {
	self->owner				= owner;
	self->reset				= _fdc_reset;
	self->read_status_reg	= _fdc_read_status_reg;
	self->write				= _fdc_write_data_reg;
	self->read				= _fdc_read_data_reg;
	self->enter_idol_phase	= _fdc_enter_idol_phase;
	self->enter_cmd_phase	= _fdc_enter_cmd_phase;
	self->enter_result_phase = _fdc_enter_result_phase;
	self->enter_exe_phase	= _fdc_enter_exe_phase;
	self->rd_ready			= _fdc_read_data_ready;
	self->TC				= _fdc_terminal_count_pin;
	_fdc_init_command(&self->cmd);
	_fdc_init_result(&self->result);
	_fdc_init_result(&self->last_result);
	_fdc_reset(self);
}
#ifdef __cplusplus
}	// extern "C"
#endif
