/**--------------------------------------------------------------------------
 **
 **		Z80 Emulator Ver 3.00
 **		Copyright (C)1993-2010 H.Kanemoto
 **
 **		This is a free software. You may copy or distribute
 **		it without any permissions. But there are NO WARRANTY
 **		of use. Entire risk of use is with you.
 **
 **------------------------------------------------------------------------*/
#ifndef _Z80_H_
#define _Z80_H_
#ifndef _MSC_VER
#define __cdecl
#define __inline inline
#include "../include/n80types.h"
#ifdef __GNUC__
#define PACK( __Declaration__ ) __Declaration__ __attribute__((__packed__))
#endif
#else
#include <windows.h>
#define PACK( __Declaration__ ) __pragma( pack(push, 1) ) __Declaration__ __pragma( pack(pop))
#endif
#ifdef __cplusplus
extern "C"
{
#endif
/**--------------------------------------------------------------------------
 **		Z80 Flag bits
 **------------------------------------------------------------------------*/
#define _CF	(0x01)
#define _NF	(0x02)
#define _PF	(0x04)
#define _VF	(0x04)
#define _XF	(0x08)
#define _HF	(0x10)
#define _YF	(0x20)
#define _ZF	(0x40)
#define _SF	(0x80)

#define	_Z80HALT_76		(1)			// HALT by HALT instruction
#define	_Z80HALT_IL		(2)			//      by illegal instruction
#define _Z80HALT_UI		(3)			//      by unimplemented instruction
#define _Z80HALT_BUS	(4)			//		by interrupt data bus error
#define _Z80HALT_INT	(5)			//		by unexpected interrupt mode error
#define _HLTBIT			(0x01)		// HALT signal
#define _NMIBIT			(0x80)		// NMI signal
#define _INTBIT			(0x40)		// INT signal
#define _INTMASK		(0xC0)
/**--------------------------------------------------------------------------
 **		Z80 WX^CfbNX
 **------------------------------------------------------------------------*/
enum Z80REGIDX	{Z80_BC, Z80_DE, Z80_HL, Z80_AF,
				 Z80_BC2,Z80_DE2,Z80_HL2,Z80_AF2,
				 Z80_IX, Z80_IY, Z80_PC, Z80_SP};
#ifndef __BIG_ENDIAN__
enum Z80REG8IDX {Z80_C = 0,Z80_B,Z80_E,Z80_D,Z80_L,Z80_H,Z80_F,Z80_A,
				 Z80_C2,Z80_B2,Z80_E2,Z80_D2,Z80_L2,Z80_H2,Z80_F2,Z80_A2,
				 Z80_IXL,Z80_IXH,Z80_IYL,Z80_IYH,
				 Z80_PCL,Z80_PCH,Z80_SPL,Z80_SPH};
#else
enum Z80REG8IDX {Z80_B = 0, Z80_C, Z80_D, Z80_E, Z80_H, Z80_L, Z80_A, Z80_F,
				 Z80_B2, Z80_C2, Z80_D2, Z80_E2, Z80_H2, Z80_L2, Z80_A2, Z80_F2,
				 Z80_IXH, Z80_IXL, Z80_IYH, Z80_IYL,
				 Z80_PCH, Z80_PCL, Z80_SPH, Z80_SPL
};
#endif
/**--------------------------------------------------------------------------
 **		Z80 vg^Cv錾
 **------------------------------------------------------------------------*/
struct Z80CPU_tag;
typedef	struct Z80CPU_tag			Z80CPU, *LPZ80CPU;
struct DZ80CONTEXT_tag;
typedef	struct DZ80CONTEXT_tag		DZ80CONTEXT, *LPDZ80CONTEXT;
typedef	void						(*DISCALLBACK)(LPDZ80CONTEXT);
/**--------------------------------------------------------------------------
 **		Z80 WX^\
 **------------------------------------------------------------------------*/
#ifdef _MSC_VER
#pragma region Z80 registers
#else
// #region Z80 registers
#endif
#ifndef __BIG_ENDIAN__
PACK(struct  R8_tag {
	BYTE	C;			// BC register pare
	BYTE	B;
	BYTE	E;			// DE register pare
	BYTE	D;
	BYTE	L;			// HL register pare
	BYTE	H;
	BYTE	F;			// AF register pare
	BYTE	A;
	BYTE	C2;			// BC' register pare
	BYTE	B2;
	BYTE	E2;			// DE' register pare
	BYTE	D2;
	BYTE	L2;			// HL' register pare
	BYTE	H2;
	BYTE	F2;			// AF' register pare
	BYTE	A2;
	BYTE	IXL;		// IX register
	BYTE	IXH;
	BYTE	IYL;		// IY register
	BYTE	IYH;
	BYTE	PCL;		// PC register
	BYTE	PCH;
	BYTE	SPL;		// SP
	BYTE	SPH;
	BYTE	Z;			// WZ internal temporary register
	BYTE	W;
});
#else
PACK(struct  R8_tag {
	BYTE	B;			// BC register pare
	BYTE	C;
	BYTE	D;			// DE register pare
	BYTE	E;
	BYTE	H;			// HL register pare
	BYTE	L;
	BYTE	A;			// AF register pare
	BYTE	F;
	BYTE	B2;			// BC' register pare
	BYTE	C2;
	BYTE	D2;			// DE' register pare
	BYTE	E2;
	BYTE	H2;			// HL' register pare
	BYTE	L2;
	BYTE	A2;			// AF' register pare
	BYTE	F2;
	BYTE	IXH;		// IX register
	BYTE	IXL;
	BYTE	IYH;		// IY register
	BYTE	IYL;
	BYTE	PCH;		// PC register
	BYTE	PCL;
	BYTE	SPH;		// SP
	BYTE	SPL;
	BYTE	W;
	BYTE	Z;
});
#endif
typedef struct R8_tag R8;

PACK(struct R16_tag {
	USHORT	BC;			// general registers
	USHORT	DE;
	USHORT	HL;
	USHORT	AF;
	USHORT	altBC;		// alternate general registers
	USHORT	altDE;
	USHORT	altHL;
	USHORT	altAF;
	USHORT	IX;			// index register
	USHORT	IY;
	USHORT	PC;			// instruction pointer
	USHORT	SP;			// stack pointer
	USHORT	WZ;			// internal temporary register
	DWORD	T;			// state in M cycle
	DWORD	TwM1;		// M1-wait state
	DWORD	W;			// memory access wait (L)
	DWORD	Wh;			// memory access wait (H)
	DWORD	M;			// interrupt machine cycle
	BYTE	I;			// interrupt vector register
	BYTE	IM;			// interrupt mode 0..2
	BYTE	R;			// re-flesh register
	BYTE	IFF1;		// interrupt flip flop #1
	BYTE	IFF2;		//					   #2
	BYTE	HALT;		// decode "HALT" code flag
	USHORT	head;		// headder PC of current code stream
	USHORT	IPC;		// interrupt jump PC
});
typedef struct R16_tag R16;
#ifdef _MSC_VER
#pragma endregion
#else
// #endregion
#endif
typedef union  Z80REGS {	// registers of Z80 Micro Processor
	R8		h;
	R16		x;
	BYTE	br[24];
	USHORT	xr[12];
}	Z80REGS, *LPZ80REGS;

typedef union IMM_tag {
	BYTE	b;
	USHORT	w;
}	IMM, *LPIMM;

struct DZ80CONTEXT_tag {
	LPZ80CPU		cpu;
	USHORT			PC;				// debug PC
	Z80REGS			rbf;			// reg-cache before exec
	Z80REGS			raf;			// reg-cache after  exec
	SHORT			opv_count;		// operand immediate value index
	SHORT			bin_w_count;	// binary cache write index;
	SHORT			bin_r_count;	// binary cache read index
	IMM				opv[5];			// operand immediate values
	BYTE			bin[8];			// binary code/op cache
	BYTE			enable;			// debug mode enable flag
	BYTE			optype;
	const char*		mnemonic;
	char			instruction[64];
	// save function pointers
	BYTE			(*z80_mbread)(LPZ80CPU, USHORT Adr, DWORD* W);
	void			(*z80_mbwrite)(LPZ80CPU, USHORT Adr, BYTE v, DWORD* W);
	int				(*z80_int_signal)(LPZ80CPU, int vector);
	void			(*z80_nmi_signal)(LPZ80CPU);
	int				(*z80_step)(LPZ80CPU);
	void			(*z80_decoder)(LPZ80CPU);

	int				(*pre_decode)(LPZ80CPU);
	BYTE			(*fetchb)(LPZ80CPU);
	BYTE			(*fetchb_M1)(LPZ80CPU);
	void			(*debug_enable)(LPDZ80CONTEXT, BYTE flag);
	void			(*disasm)(USHORT PC, int steps, DISCALLBACK callback);
	const char*		(*format_op)(LPDZ80CONTEXT);
	int				(*step)(LPZ80CPU);
	int				(*int_signal)(LPZ80CPU, int vector);
	void			(*nmi_signal)(LPZ80CPU);
	// callback functions
	int				(*on_query_breakpt)(LPDZ80CONTEXT);
	void			(*on_replace_breakpt)(LPDZ80CONTEXT, int);
	BOOL			(*on_restore_breakpt)(LPDZ80CONTEXT, int);
	void			(*on_remove_breakpt)(LPDZ80CONTEXT, int);
	void			(*on_before_exec)(LPDZ80CONTEXT);
	void			(*on_after_exec)(LPDZ80CONTEXT);
	void			(*on_req_signal)(LPDZ80CONTEXT);
};

struct Z80CPU_tag {
	LPVOID			arg;
	Z80REGS			r;
	volatile BYTE	clk;			// z80 system clock
	volatile BYTE	signal;			// z80 control signals
	BYTE			Rmsb;			// keep msb of assigned R register
	volatile BYTE	req_mode0;		// interrupt mode 0 fetch request
	BYTE			fetch_cache[8];	// fetch code cache
	SHORT			fetch_cnt;		// fetch code cache count
	BYTE			int_dbus[4];	// interrupt mode 0 data bus buffer
	SHORT			int_dbus_cnt;	// INT0 data bus count
	volatile BYTE	halt_detect;	// HALT detect edge
	BYTE			id;				// CPU ID (user defined cpu id)
	volatile DWORD	extrn_sync_clk;	// external sync-clock
	DZ80CONTEXT		dbgctx;			// debugger context
#ifdef _MSC_VER
	volatile LONG	Ts;
#else
#ifdef _ZAURUS_
	volatile DWORD	lock;		// pxa270 does not have ldrex/strex, so use swp
#else
	volatile DWORD	Ts;
#endif
#endif

	void	(*decoder)(LPZ80CPU);
	int		(*step)(LPZ80CPU);
	void	(*reset)(LPZ80CPU);
	void	(*refresh)(LPZ80CPU);
	int		(*int_signal)(LPZ80CPU, int vector);
	void	(*nmi_signal)(LPZ80CPU);

	BYTE	(*mbread)(LPZ80CPU, USHORT Adr, DWORD *W);
	void	(*mbwrite)(LPZ80CPU, USHORT Adr, BYTE v, DWORD *W);
	BYTE	(*ioread)(LPZ80CPU, USHORT port);
	void	(*iowrite)(LPZ80CPU, USHORT port, BYTE v);
};
/**--------------------------------------------------------------------------
 **		Z80 O֐|C^`
 **------------------------------------------------------------------------*/
typedef BYTE	(__cdecl *LPZ80FNCMBR)(LPZ80CPU, USHORT Adr, DWORD *W);
typedef void	(__cdecl *LPZ80FNCMBW)(LPZ80CPU, USHORT Adr, BYTE b, DWORD *W);
typedef BYTE	(__cdecl *LPZ80FNCIOR)(LPZ80CPU, USHORT port);
typedef void	(__cdecl *LPZ80FNCIOW)(LPZ80CPU, USHORT port, BYTE b);

#ifdef __cplusplus
}	// extern "C"
#endif
#endif//_Z80_H_
