/**--------------------------------------------------------------------------
 **
 **		Z80 Emulator Ver 3.00
 **		Copyright (C)1993-2010 H.Kanemoto
 **
 **		This is a free software. You may copy or distribute
 **		it without any permissions. But there are NO WARRANTY
 **		of use. Entire risk of use is with you.
 **
 **------------------------------------------------------------------------*/
#include "z80.h"
#ifndef _Z80MAC_H_
#define _Z80MAC_H_
#ifndef _MSC_VER
#define __cdecl
#define __inline inline
#else
#include <windows.h>
#endif
#include <stdio.h>
#ifdef __cplusplus
extern "C"
{
#endif
#define DZ80_OPTYPE_UNKNOWN		(-1)
#define DZ80_OPTYPE_PLANE		(0)
#define DZ80_OPTYPE_IMM8		(1)
#define DZ80_OPTYPE_IMM16		(2)
#define DZ80_OPTYPE_IXD			(3)
#define DZ80_OPTYPE_IXD_IMM8	(4)
#define DZ80_OPTYPE_OFFSET		(5)
#define DZ80_OPTYPE_MEM16W		(6)
#define DZ80_OPTYPE_MEM8W		(7)
#define DZ80_OPTYPE_MEM16R		(9)
#define DZ80_OPTYPE_MEM8R		(19)
/**--------------------------------------------------------------------------
 **		Z80 microcode inline functions
 **------------------------------------------------------------------------*/
extern volatile BYTE z80_signal;		// z80 control signals
/**--------------------------------------------------------------------------
 **		Z80 debug
 **------------------------------------------------------------------------*/
/**--------------------------------------------------------------------------
 **		Z80 macro
 **------------------------------------------------------------------------*/
static __inline USHORT	MAKER16(BYTE h, BYTE l) { return (((USHORT) h) << 8) | ((USHORT) l); }
static __inline BYTE	R16L(USHORT x) { return (BYTE) (x & 0xff); }
static __inline BYTE	R16H(USHORT x) { return (BYTE) ((x >> 8) & 0xff); }
static __inline void	SWAP8(LPBYTE r1, LPBYTE r2) { BYTE t = *r1; *r1 = *r2; *r2 = t; }
static __inline void	SWAP16(USHORT *r1, USHORT *r2) { USHORT t = *r1; *r1 = *r2; *r2 = t; }
/**--------------------------------------------------------------------------
 **		Z80 microcode inline functions (memory access)
 **------------------------------------------------------------------------*/
static __inline BYTE __z80_mbread(LPZ80CPU cpu, USHORT adr) {
	return cpu->mbread(cpu, adr, &cpu->r.x.W);
}
static __inline void __z80_mbwrite(LPZ80CPU cpu, USHORT adr, BYTE v) {
	cpu->mbwrite(cpu, adr, v, &cpu->r.x.W);
}
static __inline USHORT __z80_mwread(LPZ80CPU cpu, USHORT adr) {
	BYTE dl = cpu->mbread(cpu, adr, &cpu->r.x.W);
	BYTE dh = cpu->mbread(cpu, ++adr, &cpu->r.x.Wh);
	return MAKER16(dh, dl);
}
static __inline void __z80_mwwrite(LPZ80CPU cpu, USHORT adr, USHORT v) {
	cpu->mbwrite(cpu, adr, R16L(v), &cpu->r.x.W);
	cpu->mbwrite(cpu, ++adr, R16H(v), &cpu->r.x.Wh);
}
static __inline BYTE __z80_fetchb(LPZ80CPU cpu) {
	BYTE b = cpu->mbread(cpu, cpu->r.x.PC++, &cpu->r.x.W);
	cpu->fetch_cache[cpu->fetch_cnt++] = b;
	return b;
}
static __inline BYTE __z80_fetchb_Mn(LPZ80CPU cpu) {
	++cpu->r.x.TwM1;
	cpu->refresh(cpu);	// Refresh R register
	return __z80_fetchb(cpu);
}
static __inline BYTE __z80_fetchb_M1(LPZ80CPU cpu) {
	++cpu->r.x.TwM1;
	cpu->refresh(cpu);	// Refresh R register
	if (cpu->signal & _HLTBIT) {
		cpu->halt_detect = 0;
		return 0x00;	// force NOP instruction duaring HALT status
	} else 	if (cpu->req_mode0 == 1) {
		// MODE 0: fetch from data bus which is set by interrupt device
		if (cpu->int_dbus_cnt > 0) {
			BYTE d = cpu->int_dbus[0];
			cpu->int_dbus_cnt = 0;
			cpu->req_mode0 = 0;
			return d;
		} else {
			cpu->signal |= _HLTBIT;	// int-data bus error
			cpu->r.x.HALT = _Z80HALT_BUS;
			cpu->halt_detect = 1;
			return 0x00;		//NOP
		}
	} else
		return __z80_fetchb(cpu);
}
static __inline USHORT __z80_fetchw(LPZ80CPU cpu) {
	BYTE bl = cpu->mbread(cpu, cpu->r.x.PC++, &cpu->r.x.W);
	BYTE bh = cpu->mbread(cpu, cpu->r.x.PC++, &cpu->r.x.Wh);
	return MAKER16(bh, bl);
}
static __inline void __z80_pushw(LPZ80CPU cpu, USHORT v) {
	cpu->mbwrite(cpu, --cpu->r.x.SP, R16H(v), &cpu->r.x.W);
	cpu->mbwrite(cpu, --cpu->r.x.SP, R16L(v), &cpu->r.x.Wh);
}
static __inline USHORT __z80_pop(LPZ80CPU cpu) {
	BYTE bl = cpu->mbread(cpu, cpu->r.x.SP++, &cpu->r.x.W);
	BYTE bh = cpu->mbread(cpu, cpu->r.x.SP++, &cpu->r.x.Wh);
	return MAKER16(bh, bl);
}
static __inline USHORT __z80_index(USHORT idx, BYTE d) {
	return idx + ((USHORT) ((signed char) d));
}
/**--------------------------------------------------------------------------
 **		Z80 debug macro (pre-decode)
 **------------------------------------------------------------------------*/
static __inline USHORT __dz80_fetchw(LPZ80CPU cpu) {
	BYTE bl = cpu->dbgctx.fetchb(cpu);
	BYTE bh = cpu->dbgctx.fetchb(cpu);
	return MAKER16(bh, bl);
}
static __inline void __dz80_op_plane(LPZ80CPU cpu, const char* mnemonic) {
	cpu->dbgctx.mnemonic = mnemonic;
	cpu->dbgctx.optype = DZ80_OPTYPE_PLANE;
}
static __inline void __dz80_op_imm16(LPZ80CPU cpu, const char* mnemonic) {
	cpu->dbgctx.mnemonic = mnemonic;
	cpu->dbgctx.optype = DZ80_OPTYPE_IMM16;
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].w = __dz80_fetchw(cpu);
}
static __inline void __dz80_op_imm8(LPZ80CPU cpu, const char* mnemonic) {
	cpu->dbgctx.mnemonic = mnemonic;
	cpu->dbgctx.optype = DZ80_OPTYPE_IMM8;
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].b = cpu->dbgctx.fetchb(cpu);
}
static __inline void __dz80_op_ixd(LPZ80CPU cpu, const char* mnemonic) {
	__dz80_op_imm8(cpu, mnemonic);
	cpu->dbgctx.optype = DZ80_OPTYPE_IXD;
}
static __inline void __dz80_op_ixd0(LPZ80CPU cpu, BYTE d, const char* mnemonic) {
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].b = d;
	cpu->dbgctx.mnemonic = mnemonic;
	cpu->dbgctx.optype = DZ80_OPTYPE_IXD;
}
static __inline void __dz80_op_ixd_imm8(LPZ80CPU cpu, const char* mnemonic) {
	cpu->dbgctx.mnemonic = mnemonic;
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].b = cpu->dbgctx.fetchb(cpu);
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].b = cpu->dbgctx.fetchb(cpu);
	cpu->dbgctx.optype = DZ80_OPTYPE_IXD_IMM8;
}
static __inline void __dz80_op_offset(LPZ80CPU cpu, const char* mnemonic) {
	cpu->dbgctx.mnemonic = mnemonic;
	BYTE e = cpu->dbgctx.fetchb(cpu);
	USHORT offset = (USHORT)((signed char)e);
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].b  = e;
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].w = cpu->dbgctx.PC + offset;
	cpu->dbgctx.optype = DZ80_OPTYPE_OFFSET;
}
static __inline void __dz80_op_mem16w(LPZ80CPU cpu, const char* mnemonic) {
	cpu->dbgctx.mnemonic = mnemonic;
	USHORT nn = __dz80_fetchw(cpu);
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].w = nn;
	cpu->dbgctx.optype = DZ80_OPTYPE_MEM16W;
}
static __inline void __dz80_op_mem8w(LPZ80CPU cpu, const char* mnemonic) {
	cpu->dbgctx.mnemonic = mnemonic;
	USHORT nn = __dz80_fetchw(cpu);
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].w = nn;
	cpu->dbgctx.optype = DZ80_OPTYPE_MEM8W;
}
static __inline void __dz80_op_mem16r(LPZ80CPU cpu, const char* mnemonic) {
	cpu->dbgctx.mnemonic = mnemonic;
	USHORT nn = __dz80_fetchw(cpu);
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].w = nn;
	//cpu->dbgctx.dr[1].x.RX = __dz80_mwread(cpu, nn);
	cpu->dbgctx.optype = DZ80_OPTYPE_MEM16R;
}
static __inline void __dz80_op_mem8r(LPZ80CPU cpu, const char* mnemonic) {
	cpu->dbgctx.mnemonic = mnemonic;
	USHORT nn = __dz80_fetchw(cpu);
	cpu->dbgctx.opv[cpu->dbgctx.opv_count++].w = nn;
	//cpu->dbgctx.dr[1].r.R  = __dz80_mbread(cpu, nn);
	cpu->dbgctx.optype = DZ80_OPTYPE_MEM8R;
}
/**--------------------------------------------------------------------------
 **		Z80 microcode inline functions (illegal instructions)
 **------------------------------------------------------------------------*/
static __inline void __z80_halt_instruction(LPZ80CPU cpu) {
	cpu->signal  |= _HLTBIT;
	cpu->r.x.HALT = _Z80HALT_76;
	cpu->halt_detect = 1;
}
static __inline void __z80_illegal_instruction(LPZ80CPU cpu) {
	//cpu->signal  |= _HLTBIT;
	//cpu->r.x.HALT = _Z80HALT_IL;
	//cpu->halt_detect = 1;
}
static __inline void __z80_unimplemented_instruction(LPZ80CPU cpu) {
	cpu->signal  |= _HLTBIT;
	cpu->r.x.HALT = _Z80HALT_UI;
	cpu->halt_detect = 1;
}
/**--------------------------------------------------------------------------
 **		Z80 microcode inline functions (Flag checker)
 **------------------------------------------------------------------------*/
static __inline BYTE __z80_microcode_sf8(BYTE r) { return (r & _SF); }
static __inline BYTE __z80_microcode_zf8(BYTE r) { return r == 0 ? _ZF : 0; }
static __inline BYTE __z80_microcode_zf16(int r) { return (r & 0xffff) == 0 ? _ZF : 0; }
static __inline BYTE __z80_microcode_yxf(BYTE r) { return (r & (_XF | _YF)); }
static __inline BYTE __z80_microcode_hf8(BYTE r1, BYTE r2, BYTE r3) { return ((r1 ^ r2) ^ r3) & _HF; }
static __inline BYTE __z80_microcode_pf8(BYTE r) {
	BYTE pf = 0;
	int i;
	for (i = 0; i < 8; ++i) {
		pf ^= r;
		r >>= 1;
	}
	return ((~pf & 0x01) << 2);
}
static __inline BYTE __z80_microcode_vf8add(BYTE r1, BYTE r2, BYTE r3) {
	return ((~(r1 ^ r2) & (r1 ^ r3)) >> 5) & _VF;
}
static __inline BYTE __z80_microcode_vf8sub(BYTE r1, BYTE r2, BYTE r3) {
	return (((r1 ^ r2) & (r1 ^ r3)) >> 5) & _VF;
}
static __inline BYTE __z80_microcode_cf8(int r)  { return (r & 0x100) ? _CF : 0; }
/**--------------------------------------------------------------------------
 **		Z80 microcode inline functions (ADD/SUB/INC/DEC)
 **------------------------------------------------------------------------*/
static __inline USHORT __z80_microcode_mem16r(LPZ80CPU cpu) {
	USHORT nn = __z80_fetchw(cpu);
	USHORT v  = __z80_mwread(cpu, nn);
	cpu->r.x.WZ = nn + 1;
	return v;
}
static __inline void __z80_microcode_mem16w(LPZ80CPU cpu, USHORT v) {
	USHORT nn = __z80_fetchw(cpu);
	__z80_mwwrite(cpu, nn, v);
	cpu->r.x.WZ = nn + 1;
}
static __inline void __z80_microcode_jp(LPZ80CPU cpu, BOOL cc) {
	USHORT nn = __z80_fetchw(cpu);
	if (cc)
		cpu->r.x.PC = nn;
	cpu->r.x.WZ = cpu->r.x.PC;
}
static __inline BOOL __z80_microcode_jr(LPZ80CPU cpu, BOOL cc) {
	USHORT e = (USHORT)((signed char)__z80_fetchb(cpu));
	if (cc)
		cpu->r.x.PC += e;
	cpu->r.x.WZ = cpu->r.x.PC;
	return cc;
}
static __inline BOOL __z80_microcode_call(LPZ80CPU cpu, BOOL cc) {
	USHORT nn = __z80_fetchw(cpu);
	if (cc) {
		__z80_pushw(cpu, cpu->r.x.PC);
		cpu->r.x.PC = nn;
	}
	cpu->r.x.WZ = cpu->r.x.PC;
	return cc;
}
static __inline BOOL __z80_microcode_ret(LPZ80CPU cpu, BOOL cc) {
	if (cc) {
		cpu->r.x.PC = __z80_pop(cpu);
		cpu->r.x.WZ = cpu->r.x.PC;
	}
	return cc;
}
static __inline void __z80_microcode_rst(LPZ80CPU cpu, USHORT pc) {
	__z80_pushw(cpu, cpu->r.x.PC);
	cpu->r.x.PC = pc;
	cpu->r.x.WZ = cpu->r.x.PC;
}
static __inline void __z80_microcode_add16(LPBYTE h1, LPBYTE l1, BYTE h2, BYTE l2, LPBYTE flag) {
	BYTE f = *flag & (_SF | _ZF | _PF);
	int lr = (int) *l1 + (int) l2, hr;
	if (__z80_microcode_cf8(lr) == _CF) {
		hr = (int) *h1 + (int) h2 + 1;
	} else {
		hr = (int) *h1 + (int) h2;
	}
	f |= __z80_microcode_yxf(hr);
	f |= __z80_microcode_hf8(*h1, h2, hr);
	f |= __z80_microcode_cf8(hr);
	*flag = f;
	*h1 = (hr & 0xff);
	*l1 = (lr & 0xff);
}
static __inline USHORT __z80_microcode_add16wz(LPBYTE h1, LPBYTE l1, BYTE h2, BYTE l2, LPBYTE flag) {
	__z80_microcode_add16(h1, l1, h2, l2, flag);
	return MAKER16(h2, l2) + 1;
}
static __inline void __z80_microcode_adc8(LPBYTE r1, BYTE r2, LPBYTE flag) {
	BYTE f = 0;
	int ir = (int)*r1 + (int)r2 + ((*flag & _CF) ? 1 : 0);
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_hf8(*r1, r2, ir);
	f |= __z80_microcode_vf8add(*r1, r2, ir);
	f |= __z80_microcode_cf8(ir);
	*flag = f;
	*r1 = (ir & 0xff);
}

static __inline void __z80_microcode_adc16(LPBYTE h1, LPBYTE l1, BYTE h2, BYTE l2, LPBYTE flag) {
	USHORT r;
	BYTE f = 0;
	// adc8 low-byte
	int lr = (int) *l1 + (int) l2 + ((*flag & _CF) ? 1 : 0);
	*l1 = (lr & 0xff);
	f = __z80_microcode_cf8(lr);
	// adc8 hi-byte
	__z80_microcode_adc8(h1, h2, &f);
	f &= ~_ZF;
	r = MAKER16(*h1, *l1);
	f |= __z80_microcode_zf16(r);
	*flag = f;
}

static __inline void __z80_microcode_sbc8(LPBYTE r1, BYTE r2, LPBYTE flag) {
	BYTE f = _NF;
	int ir = (int)*r1 - (int)r2 - ((*flag & _CF) ? 1 : 0);
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_hf8(*r1, r2, ir);
	f |= __z80_microcode_vf8sub(*r1, r2, ir);
	f |= __z80_microcode_cf8(ir);
	*flag = f;
	*r1 = (ir & 0xff);
}

static __inline void __z80_microcode_sbc16(LPBYTE h1, LPBYTE l1, BYTE h2, BYTE l2, LPBYTE flag) {
	USHORT r;
	BYTE f = 0;
	// sbc8 low-byte
	int lr = (int) *l1 - (int) l2 - ((*flag & _CF) ? 1 : 0);
	*l1 = (lr & 0xff);
	f = __z80_microcode_cf8(lr);
	// sbc8 hi-byte
	__z80_microcode_sbc8(h1, h2, &f);
	f &= ~_ZF;
	r = MAKER16(*h1, *l1);
	f |= __z80_microcode_zf16(r);
	*flag = f;
}

static __inline void __z80_microcode_add8(LPBYTE r1, BYTE r2, LPBYTE flag) {
	BYTE f = 0;
	int ir = (int) *r1 + (int) r2;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_hf8(*r1, r2, ir);
	f |= __z80_microcode_vf8add(*r1, r2, ir);
	f |= __z80_microcode_cf8(ir);
	*flag = f;
	*r1 = (ir & 0xff);
}

static __inline void __z80_microcode_sub8(LPBYTE r1, BYTE r2, LPBYTE flag) {
	BYTE f = _NF;
	int ir = (int) *r1 - (int) r2;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_hf8(*r1, r2, ir);
	f |= __z80_microcode_vf8sub(*r1, r2, ir);
	f |= __z80_microcode_cf8(ir);
	*flag = f;
	*r1 = (ir & 0xff);
}

static __inline void __z80_microcode_cp8(BYTE r1, BYTE r2, LPBYTE flag) {
	BYTE f = _NF;
	int ir = (int) r1 - (int) r2;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(r2);
	f |= __z80_microcode_hf8(r1, r2, ir);
	f |= __z80_microcode_vf8sub(r1, r2, ir);
	f |= __z80_microcode_cf8(ir);
	*flag = f;
}

static __inline void __z80_microcode_inc8(LPBYTE r, LPBYTE flag) {
	BYTE f = *flag & _CF;
	int ir = (int) *r + 1;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_hf8(*r, 1, ir);
	f |= __z80_microcode_vf8add(*r, 1, ir);
	*flag = f;
	*r = (ir & 0xff);
}

static __inline void __z80_microcode_dec8(LPBYTE r, LPBYTE flag) {
	BYTE f = (*flag & _CF) | _NF;
	int ir = (int) *r - 1;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_hf8(*r, 1, ir);
	f |= __z80_microcode_vf8sub(*r, 1, ir);
	*flag = f;
	*r = (ir & 0xff);
}

static __inline void __z80_microcode_and8(LPBYTE r1, BYTE r2, LPBYTE flag) {
	BYTE f = _HF;
	int ir = (int) *r1 & (int) r2;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r1 = (ir & 0xff);
}

static __inline void __z80_microcode_xor8(LPBYTE r1, BYTE r2, LPBYTE flag) {
	BYTE f = 0;
	int ir = (int) *r1 ^ (int) r2;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r1 = (ir & 0xff);
}

static __inline void __z80_microcode_or8(LPBYTE r1, BYTE r2, LPBYTE flag) {
	BYTE f = 0;
	int ir = (int) *r1 | (int) r2;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r1 = (ir & 0xff);
}

static __inline void __z80_microcode_daa(LPBYTE r1, BYTE r2, LPBYTE flag) {
	BYTE f = *flag & _NF;
	int ir = (int) *r1 + (int) r2;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_hf8(*r1, r2, ir);
	f |= __z80_microcode_pf8(ir);
	f |= __z80_microcode_cf8(ir);
	*flag = f;
	*r1 = (ir & 0xff);
}

static __inline void __z80_microcode_rlc(LPBYTE r, LPBYTE flag) {
	int ir = ((int) *r) << 1;
	int c = (ir >> 8) & 0x01;
	BYTE f = c;
	ir |= c;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r = (ir & 0xff);
}

static __inline void __z80_microcode_rrc(LPBYTE r, LPBYTE flag) {
	BYTE c = *r & 0x01;
	BYTE ir = *r >> 1;
	BYTE f = c;
	ir |= (c << 7);
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r = (ir & 0xff);
}

static __inline void __z80_microcode_rl(LPBYTE r, LPBYTE flag) {
	int ir = (((int) *r) << 1) | (*flag & _CF);
	int f = (ir >> 8) & 0x01;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r = (ir & 0xff);
}

static __inline void __z80_microcode_rr(LPBYTE r, LPBYTE flag) {
	BYTE f = *r & 0x01;
	BYTE ir = (*r >> 1) | ((*flag & _CF) << 7);
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r = (ir & 0xff);
}

static __inline void __z80_microcode_sla(LPBYTE r, LPBYTE flag) {
	BYTE f = (*r >> 7) & 0x01;
	BYTE ir = (*r << 1);
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r = ir;
}

static __inline void __z80_microcode_sra(LPBYTE r, LPBYTE flag) {
	BYTE m = *r & 0x80;
	BYTE f = *r & 0x01;
	BYTE ir = (*r >> 1) | m;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r = ir;
}

static __inline void __z80_microcode_sll(LPBYTE r, LPBYTE flag) {
	BYTE f = (*r >> 7) & 0x01;
	BYTE ir = (*r << 1) | 0x01;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r = ir;
}

static __inline void __z80_microcode_srl(LPBYTE r, LPBYTE flag) {
	BYTE f = *r & 0x01;
	BYTE ir = (*r >> 1);
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	*flag = f;
	*r = ir;
}

static __inline void __z80_microcode_bit(BYTE r, BYTE bit, LPBYTE flag) {
	BYTE f = (*flag & _CF) | _HF;			// CF Unchanged, HF Always set, NF Always reset.
	BYTE b = r & bit;
	f |= bit == 0x80 && b != 0 ? _SF : 0;	// SF flag Set if n = 7 and tested bit is set
	f |= __z80_microcode_yxf(r);			// XF/YF Copy from hi-byte of IXY + d 
	if (b == 0) {							// ZF/PF flag Set if the tested bit is reset
		f |= (_ZF | _PF);
	}
	*flag = f;
}

static __inline void __z80_microcode_bitx(BYTE r, BYTE bit, LPBYTE flag, USHORT ixd) {
	BYTE f = (*flag & _CF) | _HF;			// CF Unchanged, HF Always set, NF Always reset.
	BYTE b = r & bit;
	f |= bit == 0x80 && b != 0 ? _SF : 0;	// SF flag Set if n = 7 and tested bit is set
	f |= __z80_microcode_yxf(R16H(ixd));	// XF/YF Copy from hi-byte of IXY + d 
	if (b == 0) {							// ZF/PF flag Set if the tested bit is reset
		f |= (_ZF | _PF);
	}
	*flag = f;
}

static __inline void __z80_microcode_bitwz(BYTE r, BYTE bit, LPBYTE flag, BYTE w) {
	BYTE f = (*flag & _CF) | _HF;			// CF Unchanged, HF Always set, NF Always reset.
	BYTE b = r & bit;
	f |= bit == 0x80 && b != 0 ? _SF : 0;	// SF flag Set if n = 7 and tested bit is set
	f |= __z80_microcode_yxf(w);			// XF/YF Copy from hi-byte of WZ reg 
	if (b == 0) {							// ZF/PF flag Set if the tested bit is reset
		f |= (_ZF | _PF);
	}
	*flag = f;
}


static __inline void __z80_microcode_in(LPZ80CPU cpu, LPBYTE r, BYTE ph, BYTE pl) {
	BYTE f  = cpu->r.h.F & _CF;
	BYTE ir = cpu->ioread(cpu, MAKER16(ph, pl));
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_yxf(ir);
	f |= __z80_microcode_pf8(ir);
	cpu->r.h.F = f;
	*r = ir;
}

static __inline BOOL __z80_microcode_ini(LPZ80CPU cpu) {
	cpu->r.x.WZ = cpu->r.x.BC + 1;
	BYTE v = cpu->ioread(cpu, cpu->r.h.C);
	__z80_mbwrite(cpu, cpu->r.x.HL++, v);
	int b = (int)cpu->r.h.B - 1;
	int c = (int)cpu->r.h.C + 1;
	BYTE f = 0;
	f |= __z80_microcode_sf8(b);
	f |= __z80_microcode_zf8(b);
	f |= __z80_microcode_yxf(b);
	f |= __z80_microcode_pf8(((v + (c & 0xff)) & 7) ^ b);
	f |= (v + (c & 0xff) > 255) ? (_CF | _HF) : 0;
	f |= v & 0x80 ? _NF : 0;
	cpu->r.h.F = f;
	cpu->r.h.B--;
	return cpu->r.h.B == 0;
}

static __inline BOOL __z80_microcode_ind(LPZ80CPU cpu) {
	cpu->r.x.WZ = cpu->r.x.BC - 1;
	BYTE v = cpu->ioread(cpu, cpu->r.h.C);
	__z80_mbwrite(cpu, cpu->r.x.HL--, v);
	int b = (int)cpu->r.h.B - 1;
	int c = (int)cpu->r.h.C - 1;
	BYTE f = 0;
	f |= __z80_microcode_sf8(b);
	f |= __z80_microcode_zf8(b);
	f |= __z80_microcode_yxf(b);
	f |= __z80_microcode_pf8(((v + (c & 0xff)) & 7) ^ b);
	f |= (v + (c & 0xff) > 255) ? (_CF | _HF) : 0;
	f |= v & 0x80 ? _NF : 0;
	cpu->r.h.F = f;
	cpu->r.h.B--;
	return cpu->r.h.B == 0;
}

static __inline BOOL __z80_microcode_outi(LPZ80CPU cpu) {
	BYTE v = __z80_mbread(cpu, cpu->r.x.HL++);
	cpu->iowrite(cpu, cpu->r.x.BC, v);
	int b = (int)cpu->r.h.B - 1;
	int l = (int)cpu->r.h.L;
	BYTE f = 0;
	f |= __z80_microcode_sf8(b);
	f |= __z80_microcode_zf8(b);
	f |= __z80_microcode_yxf(b);
	f |= __z80_microcode_pf8(((v + l) & 7) ^ b);
	f |= (v + l) > 255 ? (_CF | _HF) : 0;
	f |= v & 0x80 ? _NF : 0;
	cpu->r.h.F = f;
	cpu->r.h.B--;
	cpu->r.x.WZ = cpu->r.x.BC + 1;
	return cpu->r.h.B == 0;
}

static __inline BOOL __z80_microcode_outd(LPZ80CPU cpu) {
	BYTE v = __z80_mbread(cpu, cpu->r.x.HL--);
	cpu->iowrite(cpu, cpu->r.x.BC, v);
	int b = (int)cpu->r.h.B - 1;
	int l = (int)cpu->r.h.L;
	BYTE f = 0;
	f |= __z80_microcode_sf8(b);
	f |= __z80_microcode_zf8(b);
	f |= __z80_microcode_yxf(b);
	f |= __z80_microcode_pf8(((v + l) & 7) ^ b);
	f |= (v + l) > 255 ? (_CF | _HF) : 0;
	f |= v & 0x80 ? _NF : 0;
	cpu->r.h.F = f;
	cpu->r.h.B--;
	cpu->r.x.WZ = cpu->r.x.BC + 1;
	return cpu->r.h.B == 0;
}

static __inline void __z80_microcode_ldair(LPZ80CPU cpu, BYTE r) {
	BYTE f = cpu->r.h.F & _CF;
	cpu->r.h.A = r;
	f |= __z80_microcode_sf8(r);
	f |= __z80_microcode_zf8(r);
	f |= __z80_microcode_yxf(r);
	if (cpu->r.x.IFF2)
		f |= _PF;
	cpu->r.h.F = f;
}

static __inline void __z80_microcode_rld(LPZ80CPU cpu) {
	BYTE f = cpu->r.h.F & _CF;
	BYTE n = __z80_mbread(cpu, cpu->r.x.HL);
	BYTE a = cpu->r.h.A;
	a &= 0xf0;
	a |= (n >> 4);
	n <<= 4;
	n |= (cpu->r.h.A & 0x0f);
	cpu->r.h.A = a;
	__z80_mbwrite(cpu, cpu->r.x.HL, n);
	f |= __z80_microcode_sf8(a);
	f |= __z80_microcode_zf8(a);
	f |= __z80_microcode_yxf(a);
	f |= __z80_microcode_pf8(a);
	cpu->r.h.F = f;
}

static __inline void __z80_microcode_rrd(LPZ80CPU cpu) {
	BYTE f = cpu->r.h.F & _CF;
	BYTE n = __z80_mbread(cpu, cpu->r.x.HL);
	BYTE a = cpu->r.h.A;
	a &= 0xf0;
	a |= (n & 0x0f);
	n >>= 4;
	n |= (cpu->r.h.A << 4);
	cpu->r.h.A = a;
	__z80_mbwrite(cpu, cpu->r.x.HL, n);
	f |= __z80_microcode_sf8(a);
	f |= __z80_microcode_zf8(a);
	f |= __z80_microcode_yxf(a);
	f |= __z80_microcode_pf8(a);
	cpu->r.h.F = f;
}

static __inline BOOL __z80_microcode_ldi(LPZ80CPU cpu) {
	BYTE f = cpu->r.h.F & (_SF | _ZF | _CF);
	BYTE n = __z80_mbread(cpu, cpu->r.x.HL);
	__z80_mbwrite(cpu, cpu->r.x.DE, n);
	++ cpu->r.x.HL;
	++ cpu->r.x.DE;
	-- cpu->r.x.BC;
	if (cpu->r.x.BC != 0)
		f |= _PF;
	n += cpu->r.h.A;
	if (n & 0x02)	f |= _YF;
	if (n & 0x08)	f |= _XF;
	cpu->r.h.F = f;
	return cpu->r.x.BC == 0;
}

static __inline BOOL __z80_microcode_ldd(LPZ80CPU cpu) {
	BYTE f = cpu->r.h.F & (_SF | _ZF | _CF);
	BYTE n = __z80_mbread(cpu, cpu->r.x.HL);
	__z80_mbwrite(cpu, cpu->r.x.DE, n);
	-- cpu->r.x.HL;
	-- cpu->r.x.DE;
	-- cpu->r.x.BC;
	if (cpu->r.x.BC != 0)
		f |= _PF;
	n += cpu->r.h.A;
	if (n & 0x02)	f |= _YF;
	if (n & 0x08)	f |= _XF;
	cpu->r.h.F = f;
	return cpu->r.x.BC == 0;
}

static __inline BOOL __z80_microcode_cpi(LPZ80CPU cpu) {
	BYTE f = (cpu->r.h.F & _CF) | _NF;
	BYTE n = __z80_mbread(cpu, cpu->r.x.HL);
	BYTE ir = (int) cpu->r.h.A - (int) n;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_hf8(cpu->r.h.A, n, ir);
	if (f & _HF)	--ir;
	if (ir & 0x02)	f |= _YF;
	if (ir & 0x08)	f |= _XF;
	++cpu->r.x.HL;
	--cpu->r.x.BC;
	if (cpu->r.x.BC != 0)
		f |= _PF;
	cpu->r.h.F = f;
	cpu->r.x.WZ += 1;
	return cpu->r.x.BC == 0;
}

static __inline BOOL __z80_microcode_cpd(LPZ80CPU cpu) {
	BYTE f = (cpu->r.h.F & _CF) | _NF;
	BYTE n = __z80_mbread(cpu, cpu->r.x.HL);
	int ir = (int) cpu->r.h.A - (int) n;
	f |= __z80_microcode_sf8(ir);
	f |= __z80_microcode_zf8(ir);
	f |= __z80_microcode_hf8(cpu->r.h.A, n, ir);
	if (f & _HF)	--ir;
	if (ir & 0x02)	f |= _YF;
	if (ir & 0x08)	f |= _XF;
	--cpu->r.x.HL;
	--cpu->r.x.BC;
	if (cpu->r.x.BC != 0)
		f |= _PF;
	cpu->r.h.F = f;
	cpu->r.x.WZ -= 1;
	return cpu->r.x.BC == 0;
}

#ifdef __cplusplus
}	// extern "C"
#endif
#endif//_Z80MAC_H_
